// anim.c: Animation related functions for Bapple2.  Created on April 12th, 1998

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>

#include <allegro.h>
#include "bapple2.h"
#include "bap2dat.h"
#include "globals.h"
#include "map.h"
#include "msg.h"

#define MAX_ANIMS	50

typedef struct anim_t
{
	char active;
	int tilenum;
	char numframes;
	unsigned long rate;
	unsigned long last_switch;
	DATAFILE *datfile;
} anim_t;

anim_t anim[MAX_ANIMS];
int active_anims = 0;

void read_anim_data(char *filename, DATAFILE *datfile);
void process_animations(void);

/* read_anim_data(): ------------------------------------------------------
 *   Reads animation data (definitions) from the given filename.
 */
void read_anim_data(char *filename, DATAFILE *datfile)
{
	FILE *fp;
	char buf[256];
	int i, j = 0;

	// If we're loading animations for a normal terrain set, (non-frill layer),
	// get rid of any animations from other terrain sets (non-frill layer)
	if (datfile != frillgfx)
		for (i = 0; i < active_anims; i++)
			if ((anim[i].active) && (anim[i].datfile != frillgfx))
				anim[i].active = 0;

	// Try to open the specified animation definition file
	fp = fopen(filename, "r");
	if (!fp)
	{
		add_msg(MSG_COMPUTER, "Warning: No animation definitions for %s", filename);
		return;
	}

	// Read entries from the animation definition file
	while (!feof(fp))
	{
		if (!fgets(buf, 256, fp))
			break;
		if ((buf[0] == '/') || (buf[0] == '#') || (buf[0] == ';'))
			continue;
		if (strlen(buf) < 3)
			continue;
		// Find a free animation entry
		for (i = 0; i < active_anims; i++)
			if (!anim[i].active)
				break;
		if (i == active_anims)
			active_anims++;
		if (active_anims >= MAX_ANIMS)
		{
			add_msg(MSG_COMPUTER, "Warning: Max # of animations (%d) exceeded in %s", MAX_ANIMS, filename);
			fclose(fp);
			return;
		}
		anim[i].active = 1;
		anim[i].datfile = datfile;
		sscanf(buf, "%d %d %d\n", &anim[i].tilenum, &anim[i].numframes, &anim[i].rate);
		anim[i].last_switch = ms_time;
		j++;
	}
	fclose(fp);

	if (debug)
		add_msg(MSG_COMPUTER, "%d animations read from %s", j, filename);
}

/* process_animations(): --------------------------------------------------
 *   Processes animations - changes frames at the proper times.
 */
void process_animations(void)
{
	static first_time = 1;
	int i, f;
	void *bp, *p;

	// Load the frill animation data if this is the 1st time processing animations
	if (first_time)
	{
		first_time = 0;
		read_anim_data("frill.ani", frillgfx);
		return;
	}

	// Process animations
	for (i = 0; i < active_anims; i++)
	{
		// If it's time to switch frames..
		if (ms_time - anim[i].last_switch > anim[i].rate)
		{
			if (anim[i].numframes > 1)
			{
				// Rotate the datafile pointers to provide the animation effect
				bp = (void *) anim[i].datfile[anim[i].tilenum].dat;
				for (f = anim[i].tilenum; f < anim[i].tilenum + anim[i].numframes - 1; f++)
					anim[i].datfile[f].dat = anim[i].datfile[f + 1].dat;
				anim[i].datfile[f].dat = bp;
			}
			anim[i].last_switch = ms_time;
		}
	}
}
