// BAPPLE2  Copyright 1997, 1998 by Tom Grandgent
// Started January of 1997

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <allegro.h>

// Macros to make DJGPP version ignore WinAllegro calls
#ifndef WIN32

#include <crt0.h>

#define acquire_screen() ;
#define release_screen() ;
#endif

#include "libnet.h"			// LIBNET communications library
#include "comm.h"			// Communications stuff specific to Bapple2
#include "msg.h"			// Messaging stuff specific to Bapple2

#include "bap2dat.h"        // The Allegro grabber datafile definitions

#include "pathext.h"        // Used for the directional path finding
#include "bapple2.h"        // Miscellaneous structures and constants

#include "globals.h"		// Global variables

//int _crt0_startup_flags = _CRT0_FLAG_LOCK_MEMORY;

// Timer functions
void inc_simclock(void);
void frame_counter(void);
void main_timer(void);
void move_timer(void);
void toggle_timers(void);
void zero_timers(void);

// Sound functions
int play_sound_vol(SAMPLE *spl, int x, int y, int vol, int pan, int freq, int loop);

// System functions
int init_system(void);
void do_game(void);
void do_game_processing(void);
void do_game_drawing(void);
void deallocate_memory(void);
void shutdown_system(void);
extern void options_dialog(void);

// Communications functions (see comm.c)
extern void init_comm(void);
extern void init_conn(void);
extern void do_communications(void);
extern int Send(void *Buffer, int BufferLength, char *Address);
extern void send_unit_state(unit_t *u);
void send_signal(char packet_type);
void relay_packet(char *inbuf, int packetlen, int e);

// Messaging functions
extern void init_msg_system(void);
extern void add_msg(char type, char *text, ...);
extern void process_messages(void);
extern void draw_messages(void);
void process_talking(void);

// Miscellaneous drawing/graphics functions
void page_flip(void);
void rotate_palette(int from, int to);
void replace_color(BITMAP *bmp, int oldcolor, int newcolor);
void draw_locator(void);
void DrawPath(int x1, int y1, int x2, int y2, int tilenum);
void display_info(void);
extern void display_help(void);
void screen_shot(void);
void render_world(void);
void execution_help(void);
void ingame_help(void);
int melt_in(BITMAP *source, BITMAP *dest, int slices);
COMPILED_SPRITE *recompile_bitmap(COMPILED_SPRITE *cbitmap);
void convert_to_linear(DATAFILE *datafile);
void display_sprite(BITMAP *bmp, void *sprite, int x, int y);
extern void remap_colors(BITMAP *bmp, PALETTE from, PALETTE to);	// in msg.c

// From radar.c: Radar related functions
radar_t *create_radar(char type, int interval, int speed, int range, int x, int y, unit_t *u);

// Player functions
void init_player(void);
void process_player(void);
void process_enemy(void);
extern void process_enemy_player_input(int e, signed int *ldir);
void draw_player(BITMAP *page);
void player_shoot(void);            // make the player shoot -if- it's time

// Input functions
void process_input(void);
void process_classic_player_input(signed int *ldir);
void process_modern_player_input(signed int *ldir);

// Explosion functions
void create_explosion(char type, int x, int y, int xoff, int yoff);
void process_explosions(void);
void draw_explosions(void);

// Animation functions
extern void process_animations(void);

// Map related functions
fixed distance(signed int x1, signed int y1, signed int x2, signed int y2);
void create_empty_map(void);
void tiled_background(void);
void fix_coords(int *x, int *y, signed int *xoff, signed int *yoff);
void find_closest_free_tile(int *dx, int *dy);
int FreeTile(int x, int y);
int TileNum(int x, int y);
// From map.c:
extern int save_map(char *filename);
extern int load_map(char *filename);
extern void load_frill_gfx(char *filename);
extern void init_gui(void);
extern void init_editor_gui(void);
extern void process_gui(void);
extern wnd_t *create_wnd(int x, int y, int w, int h, char style, char *title, void (*draw)(wnd_t *w), int (*input)(wnd_t *w), void *data);
extern void bring_to_front(wnd_t *w);
extern void destroy_window(wnd_t *wnd);
extern void draw_tilesrc(wnd_t *w);
extern void draw_tilewnd(wnd_t *w);
extern void draw_orderwnd(wnd_t *w);
extern int input_tilesrc(wnd_t *w);
extern int input_tilewnd(wnd_t *w);
extern int input_orderwnd(wnd_t *w);
extern int input_easymove(wnd_t *w);
extern char scatter_windows;
extern int tileperiodicity;
extern int tileareaw, tileareah;
// From bigmap.c:
extern void init_map_view(void);
extern void display_big_map(void);

// From base.c: (Base related functions)
extern void init_bases(void);
extern void place_bases(void);
extern void create_base(short x, short y, signed char owner);
extern void create_mbase(short x, short y, signed char owner);
extern void destroy_base(struct BASE_NODE *b);
extern void destroy_mbase(struct MBASE_NODE *b);
extern void draw_base(struct BASE_NODE *base);
extern void draw_mbase(struct MBASE_NODE *base);
extern void process_bases(void);

// Projectile related functions
void initialize_projectiles(void);
int create_projectile(float speed, signed char type, signed char team, player_t *ptarget, unit_t *utarget, int x, int y, int xoff, int yoff, int direction);
void create_burst(float speed, signed char type, signed char team, int x, int y, int xoff, int yoff);
void process_projectiles(void);

// Unit related functions
void initialize_units(void);
void draw_sprites(void);
void process_units(void);
void modify_direction(signed int *source, signed int target, int step);
// From orders.c:
extern void head_towards(int x1, int y1, int xoff1, int yoff1, int x2, int y2, int xoff2, int yoff2, int *direction);
extern void process_orders(unit_t *u);
extern void issue_order(unit_t *unit, char order);
extern unit_t *create_unit(signed char type, int x, int y, int direction, unsigned char owner);
extern void destroy_all_units(void);
extern void destroy_unit(unit_t *u);
extern void free_path_nodes(unit_t *u);
extern int unit_exists_at(int x, int y);


// Special debug functions
void Log(char *text);

// ******* COMMUNICATIONS STUFF *******
//void do_communications(void);

// ********* WSOCK STUFF **********
extern char server_addr[80];
extern int WSOCK_PORT;
//extern Socket *s;
//extern InetAddress Connect;
extern char Connect[80];

int firstpage = 1;		// had to make this global...

int depth = 8;			// color depth (should move this to globals)

/* Log(): -----------------------------------------------------------------
 *   Saves some text out to disk - primitive logging function for
 *   debugging game startup problems.
 */
void Log(char *text)
{
	FILE *fp;

	fp = fopen("init.log", "a");
	fprintf(fp, "%s\n", text);
	fclose(fp);
}

/* create_empty_map(): ----------------------------------------------------
 *   Creates a new map, loading islands.map by default, and creating
 *   randomly placed bases to go onto that map.
 * Notes:
 *   This sets up both the displayed tile map and the terrain/blocking map.
 *   It also sets the unit_at map to all zeroes, though it is updated every
 *   time the units are processed.
 */
void create_empty_map(void)     // Create a blank map
{
    int x, y;

    for (y = 0; y < MAP_HEIGHT; y++)
        for (x = 0; x < MAP_WIDTH; x++)
        {
            mpos(terrain, y, x) = TERRAIN_WATER;
			mpos(frill, y, x) = 0;
            mpos(unit_at, y, x) = 0;
        }

// OLD CODE: Randomly generates a map from scratch,
// or loads a terrain map if one is specified

// Random mix of tiles 0 and 1 (for a light ocean)
/*    for (y = 0; y < MAP_HEIGHT; y++)
        for (x = 0; x < MAP_WIDTH; x++)
            if (rand() % 40)
                mpos(map, y, x) = rand() % 2;
            else
                switch (rand() % 3)
                {
                    case 0: mpos(map, y, x) = 11; break;
                    case 1: mpos(map, y, x) = 27; break;
                    case 2: mpos(map, y, x) = 28; break;
                }

// Starting line of tile 2 (dark ocean)
    for (x = 0; x < MAP_WIDTH; x++)
    {
		mpos(map, 0, x) = 2;
		mpos(map, 1, x) = 2;
		mpos(map, 2, x) = 2;
		mpos(map, (MAP_HEIGHT - 1), x) = 2;
		mpos(map, (MAP_HEIGHT - 2), x) = 2;
		mpos(map, (MAP_HEIGHT - 3), x) = 2;
    }

    for (y = 0; y < MAP_HEIGHT; y++)
    {
		mpos(map, y, 0) = 2;
		mpos(map, y, 1) = 2;
		mpos(map, y, 2) = 2;
		mpos(map, y, (MAP_WIDTH - 1)) = 2;
		mpos(map, y, (MAP_WIDTH - 2)) = 2;
		mpos(map, y, (MAP_WIDTH - 3)) = 2;
    }

    for (x = 0; x < MAP_WIDTH; x++)
        mpos(map, x, x) = 2;

    if (!terrainfile[0] && (!(fp = fopen("map.bap", "rb"))))
    {
        for (i = 0; i < 40; i++)
        {
            px = rand() % MAP_WIDTH;
            py = rand() % MAP_HEIGHT;
            good = 1;
            for (y = py; ((y < py + 6) && good); y++)
                for (x = px; ((x < px + 7) && good); x++)
                    if (px > MAP_WIDTH - 1)
                        good = 0;
                    else if (py > MAP_HEIGHT - 1)
                        good = 0;
                    else if (mpos(terrain, y, x) != TERRAIN_WATER)
                        good = 0;
            if (!good)
            {
                i--;
                continue;
            }
            for (y = 0; y < 6; y++)
                for (x = 0; x < 7; x++)
                {
                    if (chunk[y][x] >= 0)
                    {
                        mpos(map, (py + y), (px + x)) = chunk[y][x];
                        mpos(terrain, (py + y), (px + x)) = chunk_terrain[y][x];
                    }
                }
        }
    }
    else    // load a terrain file
    {
		if (terrainfile[0])
	        fp = fopen(terrainfile, "rb");
        for (y = 0; y < 100; y++)
        {
            for (x = 0; x < 100; x++)
            {
                mpos(map, y, x) = fgetc(fp);
				if ((mpos(map, y, x) == 1) && (rand() % 2))
					mpos(map, y, x) = 0;
                if (mpos(map, y, x) == 43)
                    mpos(map, y, x) = 13;
                if (mpos(map, y, x) == 26)
                    mpos(terrain, y, x) = TERRAIN_LAND;
                else if ((mpos(map, y, x) == 1) || (mpos(map, y, x) == 0))
				{
                    mpos(terrain, y, x) = TERRAIN_WATER;
					// Random placement of islands in the water
					if (rand() % 120 == 0)
					{
						mpos(terrain, y, x) = TERRAIN_ISLAND;
						switch (rand() % 4)
						{
                            case 0: mpos(map, y, x) = 11; break;
                            case 1: mpos(map, y, x) = 27; break;
							default: mpos(map, y, x) = 28;
						}
					}
				}
                else
                    mpos(terrain, y, x) = TERRAIN_COAST;
            }
            fgetc(fp);  // get rid of the CR's
            fgetc(fp);  // get rid of the LF's
        }
        fclose(fp);
    }
*/
	// Load the initial map
	if (terrainfile[0])
		load_map(terrainfile);
	else
		load_map("islands.map");

	// Place the bases
	place_bases();

    for (y = 0; y < MAP_HEIGHT; y++)
    {
		mpos(terrain, y, 0) = TERRAIN_BLOCKED;
        mpos(terrain, y, (MAP_WIDTH - 1)) = TERRAIN_BLOCKED;
    }
    for (x = 0; x < MAP_WIDTH; x++)
    {
		mpos(terrain, 0, x) = TERRAIN_BLOCKED;
        mpos(terrain, (MAP_WIDTH - 1), x) = TERRAIN_BLOCKED;
    }

// Initialize the player's main base to NULL if it's a client
	if (comm_mode == COMM_CLIENT)
		player.mbase = 0;
}

/* tiled_background(): ----------------------------------------------------
 *   This is the core of the rendering engine.  It paints the terrain
 *   based on the tile references in the terrain map, at the position
 *   based on the player's current location.
 * Notes:
 *   This function also calculates the screen coordinates for where
 *   the player should be drawn (usually the center of the screen, but
 *   not always, as in when the player approaches map edges).
 *   Also, this function will make sure that the display is kept within
 *   a valid area, and it will also adjust the player's map coordinates
 *   if they go out of bounds.
 */

void tiled_background(void)
{
    int x, y, mx, my, xoffset, yoffset;
	int tilesize = 32;
    char no_xbg_sync = 0, no_ybg_sync = 0;
	unsigned short tilenum;
	static int oldbgx = 0, oldbgy = 0, oldbgxoff = 0, oldbgyoff = 0;

	// Calculate the background upper-left coordinates by subtracting from the player
	// coordinates.  bgx and bgy are the world coordinates (in tiles) of the first
	// tile that gets rendered (upper-left tile).

// - 5		// REAL: -4	// No, -4 // -5!!
    bgx = player.x - (PhysicalWidth / 64);
// - 5      // REAL: -5	// No, -4 // -4!!
    bgy = player.y - (PhysicalHeight / 60);

    bgxoff = player.xoff + 8;       // - 16//-24// changes hoffset on desync
    bgyoff = player.yoff + 16;      // + 27//+16// changes voffset on desync

	// Modify the pixel offsets to account for window sliding (if enabled)
	bgxoff -= bgxslide;
	bgyoff -= bgyslide;

	// Get bgxoff and bgyoff in the range of 0-31 by modifying bgx and bgy accordingly
	fix_coords(&bgx, &bgy, &bgxoff, &bgyoff);

	// If we're rendering the edge of the world, don't pass beyond certain boundaries..
	// Also turn off vertical and/or horizontal synchronization between player location
	// and the background, so that the player can move up to the edge of the world
	// (and the edge of the screen)  This is a bit hackish but it more or less works..
    if (bgx < 2)
    {
        bgx = 2;
        bgxoff = 0;
        no_xbg_sync = 1;
    }
    if (bgy < 1)
    {
        bgy = 1;
        bgyoff = 0;
        no_ybg_sync = 1;
    }
    if (bgx > MAP_WIDTH - (((SCREEN_W + 64) / 32) + 1))   // 14    // 13 for 320x240
    {
        bgx = MAP_WIDTH - ((SCREEN_W + 64) / 32);   // 13    // 12 for 320x240
        bgxoff = 0;
        no_xbg_sync = 1;
    }
    y = (SCREEN_H + 80) / 32;
    if ((bgy > MAP_HEIGHT - (y)) || ((bgy > MAP_HEIGHT - (y + 1)) && (bgyoff > 16)))
    {
        bgy = MAP_HEIGHT - y;
        bgyoff = 16;
        no_ybg_sync = 1;
    }

	// Calculate the player's screen coordinates (dependant on resolution)
    centerx = (LogicalWidth / 2) - 48 + bgxslide;
    centery = (LogicalHeight) / 2 - 16 + bgyslide;

	// Do some fine-tuning alignment adjustments to get the player to line up
	// with the game world..  ack, this method SUCKS!
	if ((SCREEN_W == 400) && (SCREEN_H == 300))
	{
		centerx -= 12;
		centery -= 9;
	}

	if ((SCREEN_W == 640) && (SCREEN_H == 480))
	{
		centerx -= 3;
		centery -= 3;
	}

	if ((SCREEN_W == 800) && (SCREEN_H == 600))
	{
	}

	// If we don't have synchronization between the player and the background,
	// adjust the player's screen coordinates appropriately.
    if (no_xbg_sync)
        centerx -= (centerx - (((player.x - bgx) * 32) + player.xoff - 8)); // 32

    if (no_ybg_sync)                                                 //67
	{
        if (bgy == MAP_HEIGHT - y)
            centery -= (centery - (((player.y - bgy) * 32) + player.yoff + 10));
        else
            centery -= (centery - (((player.y - bgy) * 32) + player.yoff + 24));
		if (SCREEN_H > 240)
			centery -= 12;
	}

	// Initialize the world coordinates that get varied over the main rendering loop
    xoffset = bgx;
    yoffset = bgy;

	// Unless we're in Mode-X, align drawing of the background to the nearest DWORD
	if (gfx_mode != GFX_MODEX)
		while (bgxoff % 4)
			bgxoff++;

	// Calculate the tile-under-mouse world coordinates
	mtilex = (mouse_x / 32) + bgx;
	mtiley = (mouse_y / 32) + bgy;

	// EXPERIMENTAL: Don't redraw the background if it hasn't moved
//	if (gfx_method == DOUBLE_BUFFER)
//		if ((bgx == oldbgx) && (bgy == oldbgy) && (bgxoff == oldbgxoff) && (bgyoff == oldbgyoff))
//			return;

#define SR	4	// player sight range - used if darkness is active for the main screen

	// Render the background!
    for (y = tilesize; y < LogicalHeight; y += tilesize)
    {
		my = yoffset % MAP_HEIGHT;
		for (x = 0; x <= LogicalWidth - tilesize; x += tilesize)
        {
				mx = xoffset % MAP_WIDTH;
/*				if (mpos(vis, my, mx) == VIS_DARK)
					if (!(BETWEEN(mx, player.x - SR, player.x + SR) &&
						  BETWEEN(my, player.y - SR, player.y + SR)))
					{
		                rectfill(HiddenPage, 32 + (x - bgxoff), (y - bgyoff), 64 + (x - bgxoff), 64 + (y - bgyoff), 0);
						xoffset++;
						continue;
					} */
// This next bit is debugging info, illuminating tiles occupied by units
//            if ((edit_mode) && (mpos(unit_at, (yoffset % MAP_HEIGHT), (xoffset % MAP_WIDTH))))
//                display_sprite(HiddenPage, datafile[TILE22_PCX].dat, 32 + (x /*- bgxoff*/), (y - bgyoff));
//            else
				// Get the tile index # from the tile map, making sure that it's in range
				tilenum = mpos(map, (yoffset % MAP_HEIGHT), (xoffset % MAP_WIDTH));
				tilenum %= NUM_TILES;

				// Draw the tile
				if (terraingfx)
				{
					if (motion_blur)
		                draw_trans_sprite(HiddenPage, terraingfx[tilenum].dat, 32 + (x - bgxoff), (y - bgyoff));
					else
						blit(terraingfx[tilenum].dat, HiddenPage, 0, 0, 32 + (x - bgxoff), (y - bgyoff), 32, 32);
				}

				// Now get the frill tile index #
				tilenum = mpos(frill, (yoffset % MAP_HEIGHT), (xoffset % MAP_WIDTH));
				tilenum %= NUM_FRILL_TILES;

				// Draw the frill tile
				if (frillgfx && tilenum)	// do a little masking here - don't show 0's
					draw_sprite(HiddenPage, frillgfx[tilenum].dat, 32 + (x - bgxoff), (y - bgyoff));

				// Debugging info, displays numbers over each tile, toggled with /tt
				if (display_terrain_type)
					textprintf(HiddenPage, font, 32 + (x - bgxoff) + 13, (y - bgyoff) + 13, TEXT_COLOR, "%p", mpos(unit_at, (yoffset % MAP_HEIGHT), (xoffset % MAP_WIDTH)));
//					textprintf(HiddenPage, font, 32 + (x - bgxoff) + 13, (y - bgyoff) + 13, TEXT_COLOR, "%d", mpos(terrain, (yoffset % MAP_HEIGHT), (xoffset % MAP_WIDTH)));

				// Displays a grid by drawing a rectangle around each tile
				if (display_grid)
	                rect(HiddenPage, 32 + (x - bgxoff), (y - bgyoff), 64 + (x - bgxoff), 64 + (y - bgyoff), TEXT_COLOR);

            xoffset++;
        }
        xoffset = bgx;
        yoffset++;
    }

	// Save the positioning information for further comparison (EXPERIMENTAL)
	// This is only used if we're trying not to do full screen redraws every frame
	oldbgx = bgx;
	oldbgy = bgy;
	oldbgxoff = bgxoff;
	oldbgyoff = bgyoff;
}

/* modify_direction(): ----------------------------------------------------
 *   This handy function takes a pointer to a direction, a desired direction
 *   value, and the maximum step size, and turns whatever accordingly.  For
 *   example, you could pass it &player.direction, 128, and 16 to turn the
 *   player 16 units towards pointing westward. Remember, 256 deg in a circle.
 * Notes:
 *   This function is pretty safe.  If the given direction is closer to the
 *   desired direction than "step" degrees, it will be set to the desired.
 */
void modify_direction(signed int *source, signed int target, int step)
{
    signed int i, j, c = 0;

    if (target == *source)
        return;

    i = j = *source;
    while ((i != target) && (j != target) && (c < 400))
    {
        i++;
        j--;
        if (i > (NUM_PLAYER_ROTATIONS - 1))
            i = 0;
        if (j < 0)
            j = (NUM_PLAYER_ROTATIONS - 1);
        c++;
    }

    if (j == target)
    {
        *source -= step;
    }
    else
        *source += step;

    if ((*source > (target - step)) && (*source < (target + step)))
        *source = target;

    if (*source < 0)
        *source += NUM_PLAYER_ROTATIONS;
    *source %= NUM_PLAYER_ROTATIONS;
}

/* process_player(): ------------------------------------------------------
 *   This function processes the life, death, movement, turning, shooting,
 *   and input for the player.  It also keeps the player's map coordinates
 *   within the confines of the map.
 */
void process_player(void)
{
    int i, j;
	int dx, dy;
	float fdx, fdy;
    static signed int ldir = -1;
	static int scroll_back = 0;
	static int slide_pass = 0;
	static unsigned long wait_time = 0;
	struct BASE_NODE *node;
	char regen;
	static unsigned long last_shot_time = 0;
	static unsigned long last_income_time = 0;
	char buf[80];
	PALETTE curpal;

	// Handle the increasing of the player's money every now and then
	if (last_income_time == 0)
		last_income_time = ms_time;
	else
	{
		if (ms_time - last_income_time > 5000)
		{
			player.money += player.income;
			last_income_time = ms_time;
		}
	}

	// Don't handle the player while in help (FIXME?)
	if (game_state == GAME_STATE_HELP)
		return;

	// Handle burst fire
    if ((player.health > 0) && (!player.unit_carried) && ((key_shifts & KB_ALT_FLAG) || ((joystick_on && joy_b3))))
	{
		if (last_shot_time == 0)
		{
			last_shot_time = ms_time;
			create_burst(fixtof(player.speed), PROJECTILE_TYPE_YELLOW, player.id, (int) player.x, (int) player.y, (int) player.xoff, (int) player.yoff);
		}
		else
		{
			if (ms_time - last_shot_time > 1000)
			{
				last_shot_time = ms_time;
				create_burst(fixtof(player.speed), PROJECTILE_TYPE_YELLOW, player.id, (int) player.x, (int) player.y, (int) player.xoff, (int) player.yoff);
			}
		}
	}

	// Handle regeneration
	while (ms_regen > 100)
	{
        ms_regen -= 100;

		// See if the player is over a center of regeneration
        regen = 0;	// Check the main base (starting location)
		if ((player.mbase) && (player.x == player.mbase->x + 2) && (player.y == player.mbase->y + 2))
			regen = 1;
		else		// Search our linked list of bases
			for (node = base; ((regen) && (node)); node = node->next)
				if ((player.x == (node->x + 1)) && (player.y == (node->y + 1)))
					regen = 1;

		// If the player is dead, they're not allowed to regenerate. :)
		if (player.health <= 0)
			regen = 0;

		// Perform a tick of regeneration
		if (regen)
		{
			player.health++;
			player.fuel += 100;
			if (player.health > 100)
				player.health = 100;
			if (player.fuel > 10000)
				player.fuel = 10000;
		}
	}

	// Process the player's movement input
    while (ms_fly > 20)
	{
	ms_fly -= 20;

// Process the enemy (for client-side prediction) if in multiplayer mode
	if (comm_mode != COMM_NONE)
		process_enemy();

// Initialize the movement direction to the player direction
    if (ldir == -1)
       ldir = player.direction;

// If the player runs out of fuel, they're toast!
	if (player.fuel <= 0)
		player.health = 0;

// If the player's health is below 0, they're obviously dead so cap it
    if (player.health < 0)
        player.health = 0;

// If they're holding down shift, turn on the afterburner flag
    if (key_shifts & KB_SHIFT_FLAG)
	{
        afterburner = 1;
		player.input |= INPUT_AFTERBURNER;
	}
    else
	{
        afterburner = 0;
		player.input &= ~INPUT_AFTERBURNER;
	}

// If they aren't dying but should be, initiate the Death Sequence(tm)
    if (player.health <= 0 && death_start_time == 0)
	{
        death_start_time = simclock;
        last_explosion_time = simclock;
    }
    else
    if (player.health <= 0)     // If player is dead or more than dead..
    {
        if ((simclock - death_start_time > 18*3) && !scroll_back) // and they've been in flames for long enough...
		{
			// Play a nice long explosion sound
			play_sample((SAMPLE *) datafile[SND_LONGEXPLOSION].dat, 255, 128, 1000 + (rand() % 400) - 200, FALSE);

			j = 255;

			// Match victor ID #
			for (i = 0; i < active_enemies; i++)
				if (enemy[i].comm_state == CS_CONNECTED)
					if (enemy[i].id == player.lastattacker)
					{
						j = 0;
						break;
					}

			// Tell the victim who killed them
			if (!j)
			{
				sprintf(buf, "You have been killed by %s!", enemy[i].name);
				add_msg(MSG_DEATH, buf);
				// Give the enemy a kill
				enemy[i].kills++;
			}
			else
			{
				if (player.fuel > 0)
					add_msg(MSG_DEATH, "You have been killed by a unit!");
				else
					add_msg(MSG_DEATH, "You have run out of fuel!");
			}

			// Add another death for the victim
			player.deaths++;

			// Set up a death announcement packet, if in multiplayer
			if (comm_mode != COMM_NONE)
			{
				buf[0] = PACKET_DEATH;
				buf[1] = player.id;				// victim
				buf[2] = player.lastattacker;	// victor
	
				// Send out the death packet
				if (comm_mode == COMM_CLIENT)
					Send(&buf[0], 3, enemy[0].addr);
				else if (comm_mode == COMM_SERVER)
					for (i = 0; i < active_enemies; i++)
						if (enemy[i].comm_state == CS_CONNECTED)
		                    Send(&buf[0], 3, enemy[i].addr);
			}

			scroll_back = 1;	// Initiate the "scrolling back to base" sequence
		}

		// If we're scrolling back to base...
		if (scroll_back)
        {
			if (!player.mbase)	// Don't go anywhere if their main base is dead (muha)
				return;
			// If we're not back to base yet, move towards it..
            if ((player.x != player.mbase->x + 2) || (player.y != player.mbase->y + 2))
            {
				// Figure out which direction we need to move in
                head_towards(player.x, player.y, player.xoff, player.yoff, player.mbase->x + 2, player.mbase->y + 2, plsxoff, plsyoff, &player.direction);
                player.speed = itofix(12);  // twice the normal player speed

                // Actually go and change the player coordinates
#ifdef FLOAT_MOVE
                player.fxoff += (float) fixtof(fmul(fcos(itofix(player.direction)), player.speed));
                player.fyoff += (float) fixtof(fmul(fsin(itofix(player.direction)), player.speed));
				player.xoff = (int) player.fxoff;
				player.yoff = (int) player.fyoff;
#else
                player.xoff += fixtoi(fmul(fcos(itofix(player.direction)), player.speed));
                player.yoff += fixtoi(fmul(fsin(itofix(player.direction)), player.speed));
#endif
            }
            else
            {
				// We have reached the respawn destination (main base..)
                // Have we reached the right sub-tile (pixel) coordinates?
                if ((player.xoff == plsxoff) && (player.yoff == plsyoff))
                {
					// If we haven't been waiting, set the respawn time
					if (wait_time == 0)
						wait_time = ms_time + 2000;	// 2 seconds to respawn
					else
					{
						// If it's past the respawn time, respawn
                        if (ms_time > wait_time)
						{
							wait_time = 0;
		                    // Play the respawn sound
							play_sample((SAMPLE *) datafile[SND_SPAWN].dat, 255, 128, 1000, FALSE);
							// Tell other clients to play the respawn sound
                            send_signal(PACKET_REGEN);
							// Fade the screen to white
							get_palette(curpal);
		                    fade_from((RGB *) curpal, whitepal, 16);
							// Draw the player to the screen
		                    draw_player(VisiblePage);
							// Fade the screen back to the current palette
		                    fade_from(whitepal, curpal, 16);
							// Reset player parameters (bring em back to live)
		                    death_start_time = 0;
		                    scroll_back = 0;
		                    init_player();
		                    player.speed = 0;
						}
					}
                }
                else    // home in on the correct sub-pixel coordinates
                {
#ifdef FLOAT_MOVE
                    if (player.xoff < plsxoff)
					{
                        player.fxoff += 1.0;
						player.xoff = (int) player.fxoff;
					}
                    else
                        if (player.xoff > plsxoff)
						{
                            player.fxoff -= 1.0;
							player.xoff = (int) player.fxoff;
						}

                    if (player.yoff < plsyoff)
					{
                        player.fyoff += 1.0;
						player.yoff = (int) player.fyoff;
					}
                    else
                        if (player.yoff > plsyoff)
						{
                            player.fyoff -= 1.0;
							player.yoff = (int) player.fyoff;
						}
#else
                    if (player.xoff < plsxoff)
                        player.xoff++;
                    else
                        if (player.xoff > plsxoff)
                            player.xoff--;

                    if (player.yoff < plsyoff)
                        player.yoff++;
                    else
                        if (player.yoff > plsyoff)
                            player.yoff--;
#endif
                }
            }
        }
        else    // They're dying, but not dead yet.. so make some explosions
        {
			// If it's been long enough since our last explosion, make a new one
            if (simclock - last_explosion_time > explosion_rate)
            {
                last_explosion_time = simclock;
                create_explosion(BIG_EXPLOSION, player.x, player.y, player.xoff + ((rand() % 16) - 8), player.yoff + ((rand() % 16) - 8));
				play_sample((SAMPLE *) datafile[SND_EXPLOSION].dat, 100, 128, 1000 - (rand() % 200), FALSE);
            }
            // Also make the player spaz a bit
            if (rand() % 2)
               player.direction += rand() % 40;
            else
               player.direction -= rand() % 40;
            player.xoff += (rand() % 16) - 8;
            player.yoff += (rand() % 16) - 8;
        }

    // Make sure the player direction is in a valid state (0 <= x <= 255)
        if (player.direction < 0)
            player.direction = 255;
        if (player.direction > 255)
            player.direction = 0;
    }

        // Call the correct player movement input routine
        if (alt_input)
            process_classic_player_input(&ldir);
        else
            process_modern_player_input(&ldir);

        // [Move the player and let them shoot] if they're alive
        if (player.health > 0)
        {
            // If they press control, create a projectile
            if (key_shifts & KB_CTRL_FLAG)
                player_shoot();
			// Also allow for joystick button 1 firing..
			if (joystick_on && joy_b1)
				player_shoot();

#ifndef FLOAT_MOVE
			dx = fixtoi(fmul(fcos(itofix(ldir)), player.speed));
			dy = fixtoi(fmul(fsin(itofix(ldir)), player.speed));
#else
			fdx = (float) fixtof(fmul(fcos(itofix(ldir)), player.speed));
			fdy = (float) fixtof(fmul(fsin(itofix(ldir)), player.speed));
			dx = (int) fdx;
			dy = (int) fdy;
#endif

//			dx = (int) (cos(((float) ldir) * 3.1415265359 / 180.0) * fixtof(player.speed));
//			dy = (int) (sin(((float) ldir) * 3.1415265359 / 180.0) * fixtof(player.speed));
//			add_msg(MSG_COMPUTER, "dx = %d, dy = %d", dx, dy);

			// Decrease fuel supply
			if (player.speed > 0)
			{
//				if (comm_mode != COMM_NONE)
					player.fuel -= fixtoi(player.speed / 4);
//				else
//					player.fuel -= fixtoi(player.speed);
			}

			// Move the player's world coordinates
#ifndef FLOAT_MOVE
            player.xoff += dx;
            player.yoff += dy;
#else
            player.fxoff += fdx;
            player.fyoff += fdy;
			player.xoff = (int) player.fxoff;
			player.yoff = (int) player.fyoff;
#endif

			// Make the player slide off-center on the screen (a la Herzog Zwei)
			// These next two lines should eventually be a user option
			if (slide_enabled)
			{
			slide_pass++;
//			if (slide_pass % 2 == 0)
			{
				bgxslide -= (dx * 1);
				bgyslide -= (dy * 1);
			}

			// Keep the sliding in-bounds so the player remains on the screen
			if (bgxslide < 0 - (PhysicalWidth / 4))
				bgxslide = 0 - (PhysicalWidth / 4);
			if (bgxslide > PhysicalWidth / 4)
				bgxslide = PhysicalWidth / 4;
			if (bgyslide < 0 - (PhysicalHeight / 4))
				bgyslide = 0 - (PhysicalHeight / 4);
			if (bgyslide > PhysicalHeight / 4)
				bgyslide = PhysicalHeight / 4;
			}
        } // END PLAYER ALIVE SECTION
			// Make the sliding offsets slide back towards 0 if not moving in that way..
			if ((!dx && dy) || (player.health <= 0))
				if (bgxslide < 0)
				{
					bgxslide += ((PhysicalWidth < 640) ? 2 : 4);
					if (bgxslide > 0)
						bgxslide = 0;
				}
				else
                if (bgxslide > 0)
				{
					bgxslide -= ((PhysicalWidth < 640) ? 2 : 4);
					if (bgxslide < 0)
						bgxslide = 0;
				}
			if ((!dy && dx) || (player.health <= 0))
				if (bgyslide < 0)
				{
					bgyslide += ((PhysicalWidth < 640) ? 2 : 4);
					if (bgyslide > 0)
						bgyslide = 0;
				}
				else
                if (bgyslide > 0)
				{
					bgyslide -= ((PhysicalWidth < 640) ? 2 : 4);
					if (bgyslide < 0)
						bgyslide = 0;
				}
	} // End process player loop (ms_fly.. ya know?  Look at the top!)

    // Translate cross-sub-tile offset movement to change tile # coords.
#ifndef FLOAT_MOVE
    fix_coords(&player.x, &player.y, &player.xoff, &player.yoff);
#else
	remove_thing(THING_PLAYER, &player, player.x, player.y);
    while (player.fxoff < 0.0)
    {
        player.fxoff += 32.0;
		player.xoff += 32;
        (player.x)--;
    }
    while (player.fxoff > 31.0)
    {
        player.fxoff -= 32.0;
		player.xoff -= 32;
        (player.x)++;
    }
    while (player.fyoff < 0.0)
    {
        player.fyoff += 32.0;
		player.yoff += 32;
        (player.y)--;
    }
    while (player.fyoff > 31.0)
    {
        player.fyoff -= 32.0;
		player.yoff -= 32;
        (player.y)++;
    }
	player.xoff = (int) player.fxoff;
	player.yoff = (int) player.fyoff;
	add_thing(THING_PLAYER, &player, player.x, player.y);
#endif

// Restrict the player (not background scrolling) to bounds within the map
        if (((player.x * 32) + player.xoff) < (2 * 32 + 7))
        {
            player.x = 2;
            player.xoff = 7;
#ifdef FLOAT_MOVE
			player.fxoff = 7.0;
#endif
        }
        if (((player.y * 32) + player.yoff) < (32 + 10))
        {
            player.y = 1;
            player.yoff = 10;
#ifdef FLOAT_MOVE
			player.fyoff = 10.0;
#endif
        }
        if (((player.x * 32) + player.xoff) > (((MAP_WIDTH - 3) * 32) + 8))
        {
            player.xoff = 8;
#ifdef FLOAT_MOVE
			player.fxoff = 8.0;
#endif
            player.x = MAP_WIDTH - 3;
        }
        if (((player.y * 32) + player.yoff) > (((MAP_HEIGHT - 3) * 32) + 9))
        {
            player.yoff = 9;
#ifdef FLOAT_MOVE
			player.fyoff = 9.0;
#endif
            player.y = MAP_HEIGHT - 3;
        }
}

/* process_enemy(): ------------------------------------------------------
 *   This function basically performs client-side prediction on where the
 *   enemy is going.  It does this based on the original player movement
 *   code, just using the enemy's last known position and velocity.
 */
void process_enemy(void)
{
	int dx, dy, i, ldir;

	for (i = 0; i < active_enemies; i++)
	{
		if (enemy[i].comm_state != CS_CONNECTED)
			continue;

	// Initialize the movement direction to the enemy direction
		ldir = player.direction;

		process_enemy_player_input(i, &ldir);

    // Make sure the enemy direction is in a valid state (0 <= x <= 255)
		if (enemy[i].direction < 0)
			enemy[i].direction = 255;
		if (enemy[i].direction > 255)
			enemy[i].direction = 0;

		// Actually go and change the coordinates to move, if in bounds
		if (enemy[i].health > 0)  // This case is only if they're alive
		{
			dx = fixtoi(fmul(fcos(itofix(ldir)), enemy[i].speed));
			dy = fixtoi(fmul(fsin(itofix(ldir)), enemy[i].speed));
			
			// Move the enemy's world coordinates
			enemy[i].xoff += dx;
			enemy[i].yoff += dy;
		}

		// Translate cross-sub-tile offset movement to change tile # coords.
		fix_coords(&enemy[i].x, &enemy[i].y, &enemy[i].xoff, &enemy[i].yoff);
	
		// Restrict the enemy (not background scrolling) to bounds within the map
	    if (((enemy[i].x * 32) + enemy[i].xoff) < (2 * 32 + 7))
	    {
	        enemy[i].x = 2;
	        enemy[i].xoff = 7;
	    }
	    if (((enemy[i].y * 32) + enemy[i].yoff) < (32 + 10))
	    {
	        enemy[i].y = 1;
	        enemy[i].yoff = 10;
	    }
	    if (((enemy[i].x * 32) + enemy[i].xoff) > (((MAP_WIDTH - 3) * 32) + 8))
	    {
	        enemy[i].xoff = 8;
	        enemy[i].x = MAP_WIDTH - 3;
	    }
	    if (((enemy[i].y * 32) + enemy[i].yoff) > (((MAP_HEIGHT - 3) * 32) + 9))
	    {
	        enemy[i].yoff = 9;
	        enemy[i].y = MAP_HEIGHT - 3;
	    }
	}
}

/* init_player(): ---------------------------------------------------------
 *   Initialize the player variables so that the player is in a usable
 *   and living state.  This will be done whenever the player is spawned.
 */
void init_player(void)
{
    int g;
	static first = 1;

	if (first)	// Initialize the unit_carried pointer to NULL on 1st init
	{
		first = 0;
		player.unit_carried = NULL;
		player.income = 100;
		player.money = 0;
	}
	else
	if (player.unit_carried)
	{
		destroy_unit(player.unit_carried);
		player.unit_carried = NULL;
	}

	// Initialize the helicopter animation stuff
	player.rotorframe = 0;
	player.rotortime = ms_time;

	// Initialize the last-fired-time for all of the player's weapons
    for (g = 0; g < NUM_GUNS; g++)
        last_gun_time[g] = simclock;
    player.direction = 192;
    player.speed = 0;
    player.maxspeed = itofix(5);	// Formerly 6...
    player.shiptype = 0;
	if (player.mbase)
	{
	    player.x = player.mbase->x + 2;
	    player.y = player.mbase->y + 2;
	}
    player.xoff = plsxoff;
    player.yoff = plsyoff;
#ifdef FLOAT_MOVE
	player.fxoff = (float) plsxoff;
	player.fyoff = (float) plsyoff;
#endif
    player.health = 100;
	player.fuel = 10000;
}

extern msg_buffer_t msg_buffer;		// from msg.c

/* process_talking(): -------------------------------------------------------
 *   Process all player input for communications purposes.
 */
void process_talking(void)
{
	unsigned int key, i, j;
	char buf[1000];

	// See if the player has pressed a key for talking
	if (keypressed())
	{
		key = readkey();
		switch (key & 0xff)
		{
			case 27: talking = 0; break;	// Esc aborts talking
			case 8: if (cursor > 0) talk_buf[--cursor] = '\0'; break;	// Backspace
			case '~':
			case 13:	// Enter submits the line of text that was typed
			{
				talk_buf[cursor] = '\0';
				if (cursor > 0)
				{
					// Change any occurances of '%' to '%%' to stop evil users :)
					for (i = j = 0; i <= strlen(talk_buf); i++)
					{
						buf[j++] = talk_buf[i];
						if (talk_buf[i] == '%')
							buf[j++] = talk_buf[i];
					}
					strcpy(talk_buf, buf);

					// Send the text out to the other players
					if (comm_mode != COMM_NONE)
						if (conn_type == CONN_WSOCK)
							if (active_enemies)
							{
								if (comm_mode == COMM_SERVER)
								{
									for (i = 0; i < active_enemies; i++)
									{
										if (enemy[i].comm_state != CS_CONNECTED)
											continue;
										sprintf(buf, "%c<%s> %s", PACKET_MSG, player.name, talk_buf);
					                    Send(&buf, strlen(buf) + strlen(player.name) + 3, enemy[i].addr);
									}
								}
								else if (comm_mode == COMM_CLIENT)
								{
									sprintf(buf, "%c<%s> %s", PACKET_MSG, player.name, talk_buf);
				                    Send(&buf, strlen(buf) + strlen(player.name) + 3, enemy[0].addr);
								}
							}
					// Echo the talking locally
	                sprintf(buf, "> %s", talk_buf);
					add_msg(MSG_NORMAL, buf);
				}
				talking = cursor = 0;
				talk_buf[0] = '\0';
            	break;
			}
			default:	// Add a character to the talk buffer and move the cursor
			{
				if ((isprint(key & 0xff)) && (cursor < msg_buffer.max_msg_width))
					talk_buf[cursor++] = key & 0xff;
/*				else
				{
					sprintf(buf, "Not allowed to type key #%d", key & 0xff);
	                add_msg(MSG_COMPUTER, buf);
				} */
			}
		}
	}
}

/* process_input(): -------------------------------------------------------
 *   Process all user input except for movement and shooting.
 *   Also handles picking up and putting down units.
 */
void process_input(void)
{
    int i;
    int x, y;
    static space_locked = 0;
	char linear;
    PALETTE oldpal;
	unit_t *u;
	int ch, cmd_issued;
	wnd_t *w;
	char txtbuf[100];
	char buf[1000];
	int old_gm, old_pw, old_ph;
	int mw, mh;
	thing_t *tmpthing;

    if ((key[KEY_SPACE]) && (edit_mode))
    {
        mpos(map, player.y, player.x) = 2;
        mpos(terrain, player.y, player.x) = TERRAIN_BLOCKED;
    }
    else
    if (!edit_mode)
    if ((key[KEY_SPACE]) || (joystick_on && joy_b2)) // Space: Pick up or put down a unit
    {
        if (!space_locked)
        {
            space_locked = 1;
            if (!player.unit_carried)   // Pick up unit
            {
                ms_turn = 0;	// reset the turning rate timer cause it'll change

				// If the player has a unit ready to be picked up, see if they're on a base
				if ((unit_ready) && (player.mbase) && (player.x == player.mbase->x + 2) && (player.y == player.mbase->y + 2))
				{
					player.unit_carried = create_unit(order_state, player.x, player.y, player.direction, player.id);
					player.unit_carried->carrier = player.id;
					if (debug)
						add_msg(MSG_COMPUTER, "You have created a %s with ID #%d", unit_name[player.unit_carried->type], player.unit_carried->id);
					if (comm_mode != COMM_NONE)
					{
						buf[0] = PACKET_NEWUNIT;
						memcpy(&buf[1], player.unit_carried, sizeof(unit_t));	// unit data
						for (i = 0; i < ((comm_mode == COMM_SERVER) ? active_enemies : 1); i++)
							Send(&buf[0], 1 + sizeof(unit_t), enemy[i].addr);
					}
					order_state = -1;
					unit_ready = 0;
					for (w = wnd_list; w; w = w->next)
						if (w->input == input_orderwnd)
						{
							if (w->title)
								free(w->title);
							sprintf(txtbuf, "Place Unit Orders");
							w->title = (char *) malloc(strlen(txtbuf) + 1);
							strcpy(w->title, txtbuf);
						}
				}
				else
				// Look around for units under the player to get picked up
				for (u = units; u; u = u->next)
					if ((abs(((u->x * 32) + u->xoff) - ((player.x * 32) + player.xoff)) < 16))
						if ((abs(((u->y * 32) + u->yoff) - ((player.y * 32) + player.yoff)) < 16))
						{
							if (u->owner == player.id)
							{
								player.unit_carried = u;
								u->carrier = player.id;
								send_unit_state(u);
							}
							break;
						}
            }
            else                                    // Put down unit
            {
				desired_terrain = unit_terrain[player.unit_carried->type];
                if (FreeTile(player.x, player.y))
                {
                    ms_turn = 0;// reset the turning rate timer cause it'll change
					u = player.unit_carried;
					player.unit_carried = NULL;
					u->carrier = -1;
                    u->direction = u->targetdir = player.direction;
                    u->x = player.x;
                    u->y = player.y;
                    u->xoff = player.xoff;
                    u->yoff = player.yoff;
					// Gotta recalculate their path...
					// Figure out what their final destination was
					// If they aren't heading anywhere, issue a guard order
					if ((!u->path) && (u->dx == -1) && (u->dy == -1))
					{
//	                    issue_order(u, ORDER_GUARD);
					}
					else	// Otherwise, recalculate their path
					{
						x = u->dx;
						y = u->dy;
						free_path_nodes(u);
	
						// Now calculate a new path for em
						desired_terrain = unit_terrain[u->type];
						find_closest_free_tile(&x, &y);
						tmpthing = mpos(unit_at, u->y, u->x);
						mpos(unit_at, u->y, u->x) = 0;
						u->path = (struct NODE *) FindPath((long)x,(long)y,(long)u->x,(long)u->y);
						if (u->path)
						{
							u->OPEN = OPEN;
							u->CLOSED = CLOSED;
							u->dx = x;
							u->dy = y;
						}
						mpos(unit_at, u->y, u->x) = tmpthing;
						if (u->path)
		                    issue_order(u, ORDER_GOTO);
					}
					send_unit_state(u);
                }
            }
        }
    }
    else
       space_locked = 0;

    if (keypressed())
	{
		ch = readkey();

		cmd_issued = 0;
		if (player.unit_carried)
		{
			cmd_issued = 1;
			switch (ch)
			{
				case KEY_G << 8:	// Alt-G: Guard order
				{
					if (player.unit_carried)
					{
						add_msg(MSG_COMPUTER, "Ordering %s to guard", unit_name[player.unit_carried->type]);
						issue_order(player.unit_carried, ORDER_GUARD);
					}
					break;
				}
				case KEY_A << 8:	// Alt-A: Attack order
				{
					if (player.unit_carried)
					{
						add_msg(MSG_COMPUTER, "Ordering %s to attack nearest main-base", unit_name[player.unit_carried->type]);
						issue_order(player.unit_carried, ORDER_GOTO_CLOSEST_MBASE);
					}
					break;
				}
				case KEY_S << 8:	// Alt-S: Seizure order
				{
					if (player.unit_carried)
					{
						add_msg(MSG_COMPUTER, "Ordering %s to go into a seizure", unit_name[player.unit_carried->type]);
						issue_order(player.unit_carried, ORDER_RANDOM);
					}
					break;
				}
				case KEY_T << 8:	// Alt-T: Travel order
				{
					if (player.unit_carried)
					{
						add_msg(MSG_COMPUTER, "Ordering %s to travel somewhere", unit_name[player.unit_carried->type]);
						issue_order(player.unit_carried, ORDER_GOTO);
					}
					break;
				}
				default: cmd_issued = 0;
			}
		}

		ch >>= 8;

		if (!cmd_issued)
        switch (ch)
        {
            case 1: {	// ESC has been pressed - QUIT!!!
				add_msg(MSG_COMPUTER, "Quitting...");
				game_state = GAME_STATE_EXIT;
				// Tell the other clients that we're quitting
				if (comm_mode != COMM_NONE)
				{
					for (i = 0; i < ((comm_mode == COMM_SERVER) ? active_enemies : 1); i++)
					{
						if (enemy[i].comm_state == CS_NONEXISTENT)
							continue;
						buf[0] = PACKET_KILLCL;
						buf[1] = player.id;
				        Send(&buf[0], 2, enemy[i].addr);
						enemy[i].comm_state = CS_NONEXISTENT;
					}
				}
				// Everyone's favorite part of Bapple 2 :)
				play_sample((SAMPLE *) datafile[SND_WOOHOO].dat, 255, 128, 1000, FALSE);
				fade_out(5);
				break;
			}
#ifdef VCSUX	// THE FOLLOWING CODE IS COMMENTED OUT.. No more land drawing!
// Hit the diagonals (for making terrain)
   /* NW */ case 22: if (mpos(map, player.y, player.x) == 14) mpos(map, player.y, player.x) = 15; else mpos(map, player.y, player.x) = 14; break;
   /* NE */ case 24: if (mpos(map, player.y, player.x) == 19) mpos(map, player.y, player.x) = 18; else mpos(map, player.y, player.x) = 19; break;
   /* SW */ case 50: if (mpos(map, player.y, player.x) == 12) mpos(map, player.y, player.x) = 13; else mpos(map, player.y, player.x) = 12; break;
   /* SE */ case 52: if (mpos(map, player.y, player.x) == 16) mpos(map, player.y, player.x) = 17; else mpos(map, player.y, player.x) = 16; break;
// Now the regular directions
   /* N */  case 23: mpos(map, player.y, player.x) = 25; break;
   /* S */  case 51: mpos(map, player.y, player.x) = 24; break;
   /* E */  case 38: mpos(map, player.y, player.x) = 22; break;
   /* W */  case 36: mpos(map, player.y, player.x) = 23; break;
// Now the center (land)
            case 37: {
                if (mpos(map, player.y, player.x) == 26)
                    mpos(map, player.y, player.x) = 20;
                else
                    if (mpos(map, player.y, player.x) == 20)
                        mpos(map, player.y, player.x) = 21;
                    else
                        mpos(map, player.y, player.x) = 26;
                break;
            }
#endif
            case 20: {      // T for talk!
                talking = 1;
				cursor = 0;
                break;
            }
			case 50: {		// M for minimize all windows (in the map editor)
				hide_all_windows = !hide_all_windows;
				break;
			}
            case 53: {      // / for invert terrain.. woo!!!
				talking = 1;
				cursor = 1;
				talk_buf[0] = '/';
                break;
            }
			case 17: {		// W - creates a new window (for demonstration purposes)
				if (map_editor)
					create_wnd(rand() % SCREEN_W, rand() % SCREEN_H, 32, 32, WND_PLAIN, "Tile", draw_tilewnd, input_tilewnd, NULL);
				break;
			}
			case 31: {		// S - scatter windows (make em shake)
                if (!scatter_windows)
                    scatter_windows = 1;
                else
                    scatter_windows = 0;
				break;
			}
            case 15: {      // TAB - Change ship types
                player.shiptype++;
				if (player.shiptype >= num_ship_types)
					player.shiptype = 0;
                break;
            }
            case 55: {      // * or Print Screen.. for a screen shot
				add_msg(MSG_COMPUTER, "Screen shot taken.");
                screen_shot();
                break;
            }
            case 44: {      // Z ... Creates new units to replace dead ones
// FIXME - this is not working for some reason...?
/*                for (i = 0; i < active_units; i++)
                    if (unit[i].type == UNIT_NONEXISTENT)
                        create_unit(rand() % NUM_UNITS, ((rand() % (MAP_WIDTH - 8)) + 4), ((rand() % (MAP_HEIGHT - 8)) + 4), rand() % 256, player.id); */
                break;
            }
            case 46: {      // C for C'mere!  Tells the units to get over here
                // Freeze game time
                toggle_timers();
				for (u = units; u; u = u->next)
					if ((u->owner == player.id) && (u->orders == ORDER_GUARD))
                    {
                        // If they're guarding, calculate a path for em
                        // First their tile must not be marked in unit_at though!
                        tmpthing = mpos(unit_at, u->y, u->x);
						mpos(unit_at, u->y, u->x) = 0;
                        // Calculate the path and store it with the unit
						desired_terrain = unit_terrain[u->type];
                        u->path = (struct NODE *) FindPath((long)player.x,(long)player.y,(long)u->x,(long)u->y);
                        // Now the unit tile can be remarked as being occupied
                        mpos(unit_at, u->y, u->x) = tmpthing;
                        if (u->path)
                        {
                            u->OPEN = OPEN;
                            u->CLOSED = CLOSED;
                            issue_order(u, ORDER_GOTO);
							u->dx = player.x;
							u->dy = player.y;
							send_unit_state(u);
                        }
                    }
                // Restore game time
                toggle_timers();
                break;
            }
			case 38: {		// L for lucky number (MSG_EXPENDABLE test)
				add_msg(MSG_EXPENDABLE, "Your lucky number for today is: %d", rand() % 1000);
				break;
			}
            case 39: {      // ; for MARK the beginning of a line
                penx = player.x;
                peny = player.y;
                break;
            }
            case 40: {      // ' for drawing a line, testing path finding
                DrawPath(penx, peny, player.x, player.y, 2);
                penx = player.x;
                peny = player.y;
                break;
            }
            case 74: {      // Grey - for modifying tiles
                if (mpos(map, player.y, player.x) > 0)
	                mpos(map, player.y, player.x)--;
				else
                    mpos(map, player.y, player.x) = NUM_TILES - 1;
                break;
            }
            case 78: {      // Grey + for modifying tiles
                mpos(map, player.y, player.x)++;
                mpos(map, player.y, player.x) %= NUM_TILES;
                break;
            }
            case 35: ingame_help(); break;  // H for help
            case 41: { // ` now, not -->  // This is the System Request key!! :)
                if (!invincible)
                    invincible = 1;
                else
                    invincible = 0;
				player.money += 100;
				blit(terrain_bmp, vismap_bmp, 0, 0, 0, 0, vismap_bmp->w, vismap_bmp->h);
                break;
            }
			case 32: {		// D for toggling debug info
                if (!debug)
                    debug = 1;
                else
                    debug = 0;
                if (edit_mode)
                    edit_mode = 0;
                else
                    edit_mode = 1;
				break;
			}
            case 34: {      // G for GUNS!  (enables ALL GUNS)
                for (i = 0; i < NUM_GUNS; i++)
                    gun_status[i] = ON;
                break;
            }
			case 19: {		// R for run long messaging test :)
				add_msg(MSG_COMPUTER, "It is usually a bad idea to use buffered I/O in interactive programs; you should instead use screen-oriented functions like cprintf and cputs.  If you must use buffered I/O, you should be sure that both stdout and stderr are line-buffered or unbuffered (you can change the buffering by calling the setvbuf library function); another solution would be to fflush the output stream before calling any input function, which will ensure all pending output is written to the operating system.");
				break;
			}
/*          case 19: {      // R for randomize unit locations
                for (i = 0; i < active_units; i++)
                    if (unit[i].type != UNIT_NONEXISTENT)
                    {
                        unit[i].x = rand() % MAP_WIDTH;
                        unit[i].y = rand() % MAP_HEIGHT;
                    }
                break;
            } */
//          case 13:                // +
			case 28: {				// Enter for bringing up the "order unit(s)" window
				// Search for existing order windows
				i = 0;
				for (w = wnd_list; w; w = w->next)
					if (w->input == input_orderwnd)
					{
						i = 1;
						destroy_window(w);
					}

				// Only allow one
				if (!i)
				{
					switch (order_state)
					{
						case -1: strcpy(buf, "Place Unit Orders"); break;
						case -2:
						default: {
							if (unit_ready)
								sprintf(buf, "%s Ready", unit_name[order_state]);
							else
								sprintf(buf, "Ordering a %s", unit_name[order_state]);
						}
					}
					// Find the width and height of the widest and tallest units
					mw = mh = 0;
					for (i = 0; i < NUM_UNITS; i++)
					{
						if ((((BITMAP *) datafile[UNIT1_PCX + i].dat)->w) > mw)
							mw = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->w;
						if ((((BITMAP *) datafile[UNIT1_PCX + i].dat)->h) > mh)
							mh = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->h;
					}

//					create_wnd(((SCREEN_W / 2) - ((36 * (NUM_UNITS + 1) + 8) / 2)), 32, 36 * NUM_UNITS + 8, 40, WND_PLAIN, buf, draw_orderwnd, input_orderwnd, NULL);
					bring_to_front(create_wnd(((SCREEN_W / 2) - ((36 * (NUM_UNITS + 1) + 8) / 2)), 32, mw * NUM_UNITS + 28, mh + 16, WND_PLAIN, buf, draw_orderwnd, input_orderwnd, NULL));
				}
				break;
			}
            case 18: {              // E for toggling edit mode
				if (comm_mode == COMM_NONE)
				{
	                if (map_editor)
	                    map_editor = 0;
	                else
					{
	                    map_editor = 1;
						init_editor_gui();
					}
				}
                break;
            }
            case 12:                                // -
//                      case 74: if (speed > 0) speed--; break; // - (keypad)
//                      case 78: speed++; break;                // + (keypad)
            case 26: if (player.maxspeed > itofix(0)) player.maxspeed -= itofix(1); break;  // [
            case 27: player.maxspeed += itofix(1); break;           // ]
            case 25: {          // P for pause
				toggle_timers();
				// Wait for a key to be pressed
                while (!keypressed());
				toggle_timers();
                break;
            }
            case 2: {
                if (gun_status[PROJECTILE_TYPE_YELLOW] == ON)
                    gun_status[PROJECTILE_TYPE_YELLOW] = OFF;
                else {
                    gun_status[PROJECTILE_TYPE_YELLOW] = ON;
                    gun_status[PROJECTILE_TYPE_GREEN] = OFF;
                }
                break;
            }
            case 3: {
/*                if (gun_status[PROJECTILE_TYPE_GREEN] == ON)
                    gun_status[PROJECTILE_TYPE_GREEN] = OFF;
                else {
                    gun_status[PROJECTILE_TYPE_GREEN] = ON;
                    gun_status[PROJECTILE_TYPE_YELLOW] = OFF;
                } */
                if (gun_status[PROJECTILE_TYPE_MISSILE] == ON)
                    gun_status[PROJECTILE_TYPE_MISSILE] = OFF;
                else {
                    gun_status[PROJECTILE_TYPE_MISSILE] = ON;
                    gun_status[PROJECTILE_TYPE_YELLOW] = OFF;
                }
                break;
            }
            case 4: {
                if (gun_status[PROJECTILE_TYPE_RED] == ON)
                    gun_status[PROJECTILE_TYPE_RED] = OFF;
                else
                    gun_status[PROJECTILE_TYPE_RED] = ON;
                break;
            }
            case 5: {
                if (gun_status[PROJECTILE_TYPE_MISSILE] == ON)
                    gun_status[PROJECTILE_TYPE_MISSILE] = OFF;
                else {
                    gun_status[PROJECTILE_TYPE_MISSILE] = ON;
                    gun_status[PROJECTILE_TYPE_YELLOW] = OFF;
                }
                break;
            }
            case 6: {
                if (gun_status[PROJECTILE_TYPE_LASER] == ON)
                    gun_status[PROJECTILE_TYPE_LASER] = OFF;
                else {
                    gun_status[PROJECTILE_TYPE_LASER] = ON;
                    gun_status[PROJECTILE_TYPE_MISSILE] = OFF;
                }
                break;
            }
            case 7: {
                if (gun_status[PROJECTILE_TYPE_SPREAD] == ON)
                    gun_status[PROJECTILE_TYPE_SPREAD] = OFF;
                else
                    gun_status[PROJECTILE_TYPE_SPREAD] = ON;
                break;
            }
			// F1: help screen
			case 59:
			{
                if (game_state == GAME_STATE_PLAY)
					game_state = GAME_STATE_HELP;
				else
					game_state = GAME_STATE_PLAY;
				break;
			}
			// F2: toggles between the game and map views
			case 60:
			{
                if (game_state == GAME_STATE_PLAY)
				{
					game_state = GAME_STATE_MAP;
					init_map_view();
				}
				else
					game_state = GAME_STATE_PLAY;
				break;
			}
			// F4: Change resolution (experimental)
			case 62:
			{
				if (comm_mode == COMM_NONE)		// FIXME: This is NOT nice.
					toggle_timers();
				linear = is_linear_bitmap(screen);
                get_palette(oldpal);
				set_palette(desktop_pallete);
				if (set_gfx_mode(GFX_AUTODETECT, 640, 480, 640, 480))
				{
//					set_gfx_mode(GFX_VGA, 320, 200, 0, 0);
					set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0);
					set_palette(desktop_pallete);
				}

				old_gm = gfx_mode;
				old_pw = PhysicalWidth;
				old_ph = PhysicalHeight;

			    if (!gfx_mode_select(&gfx_mode, &PhysicalWidth, &PhysicalHeight))
					exit(1);

				// No switching from linear to planar allowed!!!!
				if ((linear) && (gfx_mode == GFX_MODEX))
				{
				    set_palette((RGB *) datafile[MAIN_PAL].dat);
					create_rgb_table(&rgb_table, (RGB *) datafile[MAIN_PAL].dat, NULL);
					rgb_map = &rgb_table;
				    set_gfx_mode(old_gm, old_pw, old_ph, LogicalWidth, LogicalHeight * 2);
					break;
				}

				LogicalWidth = PhysicalWidth + 64;
				LogicalHeight = PhysicalHeight + 64;
			
/*				bglagx = bglagy = 0;
				bglagxoff = PhysicalWidth / 2 - 8;
				bglagyoff = PhysicalHeight / 2 - 8;
				fix_coords(&bglagx, &bglagy, &bglagxoff, &bglagyoff); */
			
/*				if (PhysicalWidth == 320 && (PhysicalHeight = 200 || PhysicalHeight == 240))
				{
					PhysicalHeight = 240;
					if (gfx_mode == GFX_AUTODETECT)
						gfx_mode = GFX_MODEX;
				} */
			
				if (gfx_mode != GFX_MODEX)	// Convert compiled sprites to linear bitmaps
				{
					set_gfx_mode(GFX_MODEX, 320, 240, 320, 240);
					set_palette(black_palette);
					if (!linear)
						convert_to_linear(datafile);
				}
			
				if (gfx_method == PAGE_FLIP)	// double the fun for 2 pages
				{
				    if (set_gfx_mode(gfx_mode, PhysicalWidth, PhysicalHeight, LogicalWidth, LogicalHeight * 2) < 0)
					{
						set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
						printf("Ack!  Unable to set the graphics mode you requested (%dx%d) because:\n\n", PhysicalWidth, PhysicalHeight);
						printf("%s\n", allegro_error);
						exit(1);
					}
				}
				else
				{
				    if (set_gfx_mode(gfx_mode, PhysicalWidth, PhysicalHeight, PhysicalWidth, PhysicalHeight) < 0)
					{
						set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
						printf("Ack!  Unable to set the graphics mode you requested (%dx%d) because:\n\n", PhysicalWidth, PhysicalHeight);
						printf("%s\n", allegro_error);
						exit(1);
					}
				}
				strcpy(gfx_name, gfx_driver->name);
				strcpy(gfx_desc, gfx_driver->desc);
			
			    acquire_screen();

				VisiblePage = create_sub_bitmap(screen, 0, 0, LogicalWidth, LogicalHeight + 32);
				if (gfx_method == PAGE_FLIP)
				    HiddenPage = create_sub_bitmap(screen, 0, LogicalHeight, LogicalWidth, LogicalHeight + 32);
				else
				    HiddenPage = create_bitmap(LogicalWidth, LogicalHeight + 32);
			
			    set_palette((RGB *) datafile[MAIN_PAL].dat);
			
//				set_clip(screen, 0, 0, 0, 0);
				set_clip(HiddenPage, 0, 0, LogicalWidth - 1, LogicalHeight - 1);
				set_clip(VisiblePage, 0, 0, LogicalWidth - 1, LogicalHeight - 1);
				clear(screen);
			    text_mode(-1);

				release_screen();

				if (comm_mode == COMM_NONE)		// FIXME: This is NOT nice.
					toggle_timers();

				set_palette(oldpal);
				create_rgb_table(&rgb_table, oldpal, NULL);
				rgb_map = &rgb_table;

				firstpage = 1;
				break;
			}
			// F3 through F6, modify the bgxslide and bgyslide variables (experimental)
/*			case 61: bgxslide -= 4; break;
			case 62: bgxslide += 4; break;
			case 63: bgyslide -= 4; break;
			case 64: bgyslide += 4; break; */
			case 65: 			// F7: Toggle the sound
			{
				if (digi_vol)
					digi_vol = 0;
				else
					digi_vol = 255;
				set_volume(digi_vol, midi_vol);
				break;
			}
			case 66: 			// F8: Toggle the music
			{
				if (music_on)
				{
					stop_midi();
					music_on = 0;
				}
				else
				{
					play_midi(game_music, TRUE);
					music_on = 1;
				}
				break;
			}
			case 67: 			// F9: Test find_closest_free_tile()
			{
				x = player.x;
				y = player.y;
				for (i = 0; i < 4; i++)
				{
					find_closest_free_tile(&x, &y);
					mpos(map, y, x) = 2;
					mpos(terrain, y, x) = TERRAIN_BLOCKED;
					redraw_terrain_bmp = 1;
				}
				break;
			}
			case 68:			// F10: Create a friendly unit where the player is
			{
				if (comm_mode == COMM_NONE)
				{
					u = NULL;
					// Find a unit that can go on this terrain
					for (i = NUM_UNITS - 1; i >= 0; i--)
						if (unit_terrain[i] == mpos(terrain, player.y, player.x))
							if (!unit_exists_at(player.x, player.y))
							{
								u = create_unit(i, player.x, player.y, player.direction, player.id);
								create_radar(RADAR_NORMAL, 2000 + (rand() % 4000), 30 + rand() % 70, 8 + rand() % 30, 0, 0, u);
								break;
							}
					if ((comm_mode != COMM_NONE) && (u))
					{
						buf[0] = PACKET_NEWUNIT;
						memcpy(&buf[1], u, sizeof(unit_t));		// unit data
						for (i = 0; i < ((comm_mode == COMM_SERVER) ? active_enemies : 1); i++)
							Send(&buf[0], 1 + sizeof(unit_t), enemy[i].addr);
					}
				}
				break;
			}
			case 87:			// F11: Create an enemy unit where the player is
			{
				// Find a unit that can go on this terrain
				if (comm_mode == COMM_NONE)
					for (i = NUM_UNITS - 1; i >= 0; i--)
						if (unit_terrain[i] == mpos(terrain, player.y, player.x))
							if (!unit_exists_at(player.x, player.y))
							{
								create_unit(i, player.x, player.y, player.direction, player.id + 1);
								break;
							}
				break;
			}
			default:
			{
//				if (debug)
//					add_msg(MSG_COMPUTER, "Unknown keypress: %d", ch);
			}
        }
	}
}

/* create_projectile(): ---------------------------------------------------
 *   Creates a projectile with the given speed, type, team, coordinates,
 *   and direction.  Unbalanced tile vs pixel coordinates are ok.
 *   Note that you should not create projectiles on top of existing
 *   entities in the game, or collisions will occur immediately..
 */
int create_projectile(float speed, signed char type, signed char team, player_t *ptarget, unit_t *utarget, int x, int y, int xoff, int yoff, int direction)
{
    int p, i;
    char flag = NO;
	PACKET_PROJ_T send;

	// Stop enemies from firing missiles at a dead player
//	if ((type == PROJECTILE_TYPE_MISSILE) && (player.health <= 0))
//		return;

    fix_coords(&x, &y, &xoff, &yoff);

	// Find a free projectile slot
    for (p = 0; p < active_projectiles; p++)
	{
        if (proj[p].type == PROJ_NONEXISTENT)
		{
            flag = YES;
            break;
        }
    }
    if (flag == NO)
	{
        if (active_projectiles > MAX_PROJECTILES - 1)
            return 0;
        p = active_projectiles++;
    }
    proj[p].direction = direction;
    proj[p].x = x;
    proj[p].y = y;
    proj[p].xoff = xoff;
    proj[p].yoff = yoff;
    proj[p].type = type;
    proj[p].team = team;
    proj[p].speed = ftofix(speed);
	proj[p].utarget = utarget;
	proj[p].ptarget = ptarget;
	if (proj[p].type == PROJECTILE_TYPE_MISSILE)
	    proj[p].dietime = ms_time + 8000;
	else
	    proj[p].dietime = ms_time + 1200;

	// Tell the remote player(s) to create a projectile
	if ((comm_mode != COMM_NONE) && (team == player.id))
	{
		for (i = 0; i < ((comm_mode == COMM_SERVER) ? active_enemies : 1); i++)
		{
			if (enemy[i].comm_state != CS_CONNECTED)
				continue;
			send.type = PACKET_PROJ;
			send.ptype = proj[p].type;
			send.x = proj[p].x;
			send.y = proj[p].y;
			send.xoff = proj[p].xoff;
			send.yoff = proj[p].yoff;
			send.speed = proj[p].speed;
			send.direction = proj[p].direction;
			send.id = player.id;
			if (proj[p].ptarget)
				send.ptarget = proj[p].ptarget->id;
			else
				send.ptarget = -1;
			if (proj[p].utarget)
				send.utarget = (short) proj[p].utarget->id;
			else
				send.utarget = -1;
	        Send(&send, sizeof(PACKET_PROJ_T), enemy[i].addr);
		}
	}
	else
	if (comm_mode != COMM_NONE)
	{
		// Make noise for enemy shooting
		play_sound_vol((SAMPLE *) datafile[SND_SHOOT1].dat, x, y, 128, 128, 1000 + (rand() % 100) - 50, FALSE);
	}

	return p;
}

/* inc_simclock():   (timer interrupt handler) ----------------------------
 *   This increments "simclock" 18.2 times a second, to maintain
 *   compatibility with some old functions that relied on clock() in
 *   Borland C/C++ 3.1 (ew!)
 */
void inc_simclock(void)
{
   simclock++;
}

END_OF_FUNCTION(inc_simclock);

/* frame_counter():   (timer interrupt handler) ---------------------------
 *   This is used to measure the number of frames per second.
 *   It is called once every second, and it sets "fps" to whatever
 *   "frame_count" has been accumulating.  Then it resets "frame_count".
 *   See, "frame_count" is incremented on every frame.  Neat, eh?
 */
void frame_counter(void)
{
   fps = frame_count;
   frame_count = 0;
}

END_OF_FUNCTION(frame_counter);

/* main_timer():   (timer interrupt handler) ------------------------------
 *   This is the main timer function, called 1000 times a second, which
 *   increments most of the main timing counters.
 */
void main_timer(void)
{
    ms_time++;
    ms_unit++;
    ms_proj++;
    ms_fly++;
    ms_accel++;
    ms_turn++;
	ms_regen++;
	ms_send++;
}

END_OF_FUNCTION(main_timer);

/* move_timer():   (timer interrupt handler) ---------------------------
 *   This moves the player by changing the coordinates.
 */
void move_timer(void)
{
	player.xoff += fixtoi(fmul(fcos(player_update_direction), player_update_speed));
	player.yoff += fixtoi(fmul(fsin(player_update_direction), player_update_speed));
}

END_OF_FUNCTION(move_timer);

/* init_system(): ---------------------------------------------------------
 *   This does all sorts of things to get the game ready to go into
 *   the main loop.  Any initializing things should be called from here.
 */
int init_system(void)
{
    int i;
	PALETTE curpal;
	FILE *fp;

// Make sure we free all allocated memory upon program termination!!!
	atexit(deallocate_memory);

	Log("Registered memory deallocation function");

// Get rid of DJGPP's stdout buffering
#ifdef DJGPP
	setvbuf(stdout, NULL, _IONBF, 0);
#endif

// Create the "white palette" for use with fading
    for (i = 0; i < 256; i++)
    {
        whitepal[i].r = 63;
        whitepal[i].g = 63;
        whitepal[i].b = 63;
    }

	Log("Created white palette");

// We want a 4-button joystick type
	joy_type = JOY_TYPE_4BUTTON;

    printf("BAPPLE2 System Initialization (DJGPP and Allegro version!)\n\n");

	if (conn_type != CONN_NONE)
	{
		printf("Initializing communications system...\n");
		init_comm();
	}

	printf("Initializing..  ");
    allegro_init();

	Log("Initialized Allegro");

#ifdef DJGPP
	// George Foot's coredump thing
	coredump_hooksignals(1);
#endif

    printf("%s initialized.\n", allegro_id);
    install_timer();

	Log("Installed timer");

    srand((unsigned)time(NULL));    // randomize()

    printf("--- Using a virtual screen of %dx%d ---\n\n", LogicalWidth, LogicalHeight);

	if (sound_on)
	{
		printf("Initializing sound/music system...\n");
//		if (install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, __crt0_argv[0]) != 0)
		if (install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL) != 0)
		{
			printf("Error initializing sound system!\n%s\n", allegro_error);
			exit(1);
		}
		else
			printf("  Sound driver: %s   MIDI driver: %s\n", digi_driver->name, midi_driver->name);
		printf("Loading game music...\n");
		game_music = load_midi("bapple2.mid");
		if (!game_music)
		{
			printf("Oops, couldn't load bapple2.mid.  Did Tom forget to send you it?!\n");
			exit(1);
		}
		if (music_on)
		{
			printf("Starting game music...\n");
			play_midi(game_music, TRUE);
		}
		
		Log("Initialized sound system");
	}

	if (joystick_on)
	{
		printf("Initializing the joystick...");
		rest(500);
		if (initialise_joystick())
		{
			joystick_on = 0;
			printf("  No joystick found!\n");
		}
		else
			printf("\n");
		
		Log("Initialized joystick");
	}

    printf("Initializing path-finding system...\n");
    Stack = (struct STACK *) calloc(1, sizeof(struct STACK));

	Log("Initialized path-finding system");

	printf("Initializing linked lists for bases...\n");
	init_bases();

	Log("Initialized bases");

	printf("Allocating memory for a %dx%d map...\n", MAP_WIDTH, MAP_HEIGHT);
    terrain_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
	mapobj_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
	vismap_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
//	map = (char *) malloc(MAP_WIDTH * MAP_HEIGHT);
	map = (short *) calloc(1, MAP_WIDTH * MAP_HEIGHT * 2);
	frill = (short *) calloc(1, MAP_WIDTH * MAP_HEIGHT * 2);
	terrain = (signed char *) calloc(1, MAP_WIDTH * MAP_HEIGHT);
	vis = (char *) calloc(1, MAP_WIDTH * MAP_HEIGHT);
	unit_at = (thing_t **) calloc(1, MAP_WIDTH * MAP_HEIGHT * sizeof(thing_t *));
	if (!map || !terrain || !unit_at || !frill)
	{
		printf("Allocation failed!!\n");
		exit(1);
	}

	Log("Allocated memory for maps");

	// Clear the newly allocated bitmaps
	clear(terrain_bmp);
	clear(mapobj_bmp);
	clear(vismap_bmp);

	Log("Cleared maps");

/*	printf("Loading terrain graphics data file...\n");
    terraingfx = load_datafile("terrain.dat");
	if (!terraingfx)
	{
       allegro_exit();
       printf("Error loading TERRAIN.DAT!\n\n");
       exit(1);
	} */

// I moved this to after the graphics mode has been set,
// so that screen dimensions would be available
//	printf("Initializing windowing system...\n");
//	init_gui();

	if (conn_type != CONN_NONE)
	{
		printf("Initializing communications connection...\n");
		init_conn();
	}

// Load the datafile
    printf("Loading data file...\n");
    datafile = load_datafile("bapple2.dat");
    if (!datafile)
	{
       allegro_exit();
       printf("Error loading BAPPLE2.DAT!\n\n");
       exit(1);
    }

	Log("Loaded bapple2.dat");

    printf("Installing keyboard handler...\n");
    install_keyboard();

	Log("Installed keyboard handler");

	printf("Initializing mouse driver...\n");
	install_mouse();

	Log("Installed mouse handler");

    printf("\nEntering the world of BAPPLE2...\n");
//    if (quick_start)	// Um, yah, like this makes sense :)
//        rest(2000);

    // We're using a virtual screen of 384x304 with a 320x240 viewport
    // However, due to the page flipping, we need 2 pages, hence 384x608
    // Now I'm trying a virtual screen size of 384x336...

	if (!quick_start)
	{
		if (set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0))
		{
			Log("Failed to set 640x480, trying 320x200");
//			set_gfx_mode(GFX_VGA, 320, 200, 0, 0);
			if (set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0))
				return 1;
			Log("Set 320x200");
			set_palette(desktop_pallete);
			Log("Set desktop_palette");
		}
		else
		{
			acquire_screen();
            clear_to_color(screen, 255);
			set_palette(black_pallete);
            blit((BITMAP *) datafile[INTRO_PIC].dat, screen, 0, 0, 12, 4, 616, 116);
			release_screen();
			Log("Cleared the screen and blitted intro pic");
            fade_from(whitepal, (RGB *) datafile[INTRO_PAL].dat, 4);
			Log("Faded in");
/*			set_palette((RGB *) datafile[INTRO_PAL].dat);
			BITMAP *tmpbuf;
			BITMAP *tmpbuf2;
			tmpbuf = create_bitmap(640, 480);
			tmpbuf2 = create_bitmap(640, 480);
            clear_to_color(tmpbuf, 255);
            clear_to_color(tmpbuf2, 255);
			blit((BITMAP *) datafile[INTRO_PIC].dat, tmpbuf2, 0, 0, 12, 4, 616, 116);
			blit((BITMAP *) datafile[INTRO_PIC].dat, tmpbuf2, 0, 0, 12, 140, 616, 116);
			blit((BITMAP *) datafile[INTRO_PIC].dat, tmpbuf2, 0, 0, 12, 280, 616, 116);
			blit((BITMAP *) datafile[INTRO_PIC].dat, tmpbuf2, 0, 0, 12, 420, 616, 116);
			melt_in(tmpbuf2, tmpbuf, 640);
			destroy_bitmap(tmpbuf);
			destroy_bitmap(tmpbuf2); */
		}
        options_dialog();
		Log("Returned from options_dialog");

		fp = fopen("display.cfg", "r");
		if (!fp)
		{
			fp = fopen("display.cfg", "w");
			gfx_mode = GFX_AUTODETECT;
			PhysicalWidth = 640;
			PhysicalHeight = 480;
			depth = 8;
			fprintf(fp, "%d %d %d %d\n", gfx_mode, PhysicalWidth, PhysicalHeight, depth);
		}
		else
			fscanf(fp, "%d %d %d %d\n", &gfx_mode, &PhysicalWidth, &PhysicalHeight, &depth);
		fclose(fp);

	    if (!gfx_mode_select_ex(&gfx_mode, &PhysicalWidth, &PhysicalHeight, &depth))
			return 1;

		fp = fopen("display.cfg", "w");
		fprintf(fp, "%d %d %d %d\n", gfx_mode, PhysicalWidth, PhysicalHeight, depth);
		fclose(fp);

		Log("Returned from gfx_mode_select_ex");
	}
	else	// The -quick parameter was used, make some assumptions
	{
#ifdef WIN32
		gfx_mode = GFX_DIRECTX;
#else
		gfx_mode = GFX_AUTODETECT;
#endif
		PhysicalWidth = 640;
		PhysicalHeight = 480;
		Log("Assuming 640x480x256 fullscreen, quick-start");
	}
	LogicalWidth = PhysicalWidth + 64;
	LogicalHeight = PhysicalHeight + 64;

/*	bglagx = bglagy = 0;
	bglagxoff = PhysicalWidth / 2 - 8;
	bglagyoff = PhysicalHeight / 2 - 8;
	fix_coords(&bglagx, &bglagy, &bglagxoff, &bglagyoff); */

	// Prevent non-mode-x 320x240 and all 320x200
/*	if (PhysicalWidth == 320 && (PhysicalHeight = 200 || PhysicalHeight == 240))
	{
		PhysicalHeight = 240;
		if (gfx_mode == GFX_AUTODETECT)
			gfx_mode = GFX_MODEX;
	} */

// Convert compiled sprites to linear bitmaps
/*	if ((gfx_mode != GFX_MODEX) || (gfx_method == DOUBLE_BUFFER))
	{
		set_color_depth(8);
		set_gfx_mode(GFX_MODEX, 320, 240, 320, 240);
	    set_palette((RGB *) datafile[MAIN_PAL].dat);
		convert_to_linear(datafile);
		set_color_depth(depth);
	    set_gfx_mode(gfx_mode, PhysicalWidth, PhysicalHeight, PhysicalWidth, PhysicalHeight);
	} */

	set_color_depth(depth);

	Log("Set color depth");

DblBufferSwitch:			// OH NO!  A LABEL!!!

	if (gfx_method == PAGE_FLIP)	// double the fun for 2 pages
	{
		Log("Attempting to set graphics mode with virtual screen for page flipping");
		if (set_gfx_mode(gfx_mode, PhysicalWidth, PhysicalHeight, LogicalWidth, LogicalHeight * 2) < 0)
		{
			Log("Failed setting graphics mode for page flipping, trying double buffering");
			gfx_method = DOUBLE_BUFFER;
			goto DblBufferSwitch;	// OH NO!  A GOTO!!!  How COULD you, Tom?!?!
/*			set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
			printf("Ack!  Unable to set the graphics mode you requested (%dx%d) because:\n\n", PhysicalWidth, PhysicalHeight);
			printf("%s\n", allegro_error);
			exit(1); */
		}
	}
	else  // double buffering
	{
//	    if (set_gfx_mode(gfx_mode, PhysicalWidth, PhysicalHeight, PhysicalWidth, PhysicalHeight) < 0)
	    Log("Setting double-buffered graphics mode");
		if (set_gfx_mode(gfx_mode, PhysicalWidth, PhysicalHeight, 0, 0) < 0)
		{
			Log("Failed to set double-buffered graphics mode");
			set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
			allegro_exit();
			printf("Ack!  Unable to set the graphics mode you requested (%dx%d) because:\n\n", PhysicalWidth, PhysicalHeight);
			printf("%s\n", allegro_error);
			return 1;
		}
	}
	strcpy(gfx_name, gfx_driver->name);
	strcpy(gfx_desc, gfx_driver->desc);

// Convert compiled sprites to linear bitmaps	// FIXME!!!
/*#ifndef WIN32
	if ((is_linear_bitmap(screen)) || (gfx_method == DOUBLE_BUFFER))
	{
		set_color_depth(8);
		set_gfx_mode(GFX_MODEX, 320, 240, 320, 240);
	    set_palette((RGB *) datafile[MAIN_PAL].dat);
		convert_to_linear(datafile);
		convert_to_linear(shipgfx);
		set_color_depth(depth);
	    set_gfx_mode(gfx_mode, PhysicalWidth, PhysicalHeight, LogicalWidth, LogicalHeight);
	} 
#endif */
    set_palette((RGB *) datafile[MAIN_PAL].dat);

	Log("Set main game palette");

	acquire_screen();
    VisiblePage = create_sub_bitmap(screen, 0, 0, LogicalWidth, LogicalHeight + 32);
	release_screen();

	Log("Created sub-bitmap of screen for VisiblePage");

	if (gfx_method == PAGE_FLIP)
	    HiddenPage = create_sub_bitmap(screen, 0, LogicalHeight, LogicalWidth, LogicalHeight + 32);
	else
	    HiddenPage = create_bitmap(LogicalWidth, LogicalHeight + 32);

	Log("Created double buffer");

	// GRAPHICS MODE HAS BEEN ESTABLISHED

    printf("Loading ship data file...\n");
    shipgfx = load_datafile("ships.dat");
    if (!datafile)
	{
       allegro_exit();
       printf("Error loading SHIPS.DAT!\n\n");
       exit(1);
    }

    printf("Loading chopper data file...\n");
    choppergfx = load_datafile("chop.dat");
    if (!datafile)
	{
       allegro_exit();
       printf("Error loading CHOP.DAT!\n\n");
       exit(1);
    }

	if (shipgfx[0].type != DAT_BITMAP)
	{
		allegro_exit();
		printf("Error in SHIPS.DAT: The first entry in the file is not a BITMAP!\n\n");
		exit(1);
	}
	for (num_ship_types = 0; shipgfx[num_ship_types].type == DAT_BITMAP; num_ship_types++);

	Log("Loaded player ship graphics");

	// Load up our map and terrain files
    create_empty_map();

	Log("Loaded up map and terrain files");

	// Load the frill graphics (main and mini-base gfx at this point)
	load_frill_gfx("frill.dat");

	Log("Loaded up frill graphics");

	// Initialize the player info
    init_player();

	Log("Initialized player info");

	// Initialize the units
    add_msg(MSG_COMPUTER, "Initializing units and projectiles...");
    initialize_units();
    initialize_projectiles();

	Log("Initialized units and projectiles");

//    PhysicalHeight = SCREEN_H;
//    scroll_screen(32, 32);
//	if (!terraingfx)
	{
		get_palette(curpal);
//	    set_palette((RGB *) datafile[MAIN_PAL].dat);
		create_rgb_table(&rgb_table, (RGB *) curpal, NULL);
		create_trans_table(&trans_table, (RGB *) curpal, 64, 64, 64, NULL);

		rgb_map = &rgb_table;
		color_map = &trans_table;
	}
/*	else	// We have a terrain file...
	{
		// Find the palette and set it
		for (i = 0; terraingfx[i].type != DAT_END; i++)
			if (terraingfx[i].type == DAT_PALETTE)
			{
				set_palette((RGB *) terraingfx[i].dat);
				create_rgb_table(&rgb_table, (RGB *) terraingfx[i].dat, NULL);
				rgb_map = &rgb_table;
				j = i;
				break;
			}

		// Remap colors in the "old" graphics datafile
		for (i = 0; datafile[i].type != DAT_END; i++)
			if (datafile[i].type == DAT_BITMAP)
				remap_colors((BITMAP *) datafile[i].dat, (RGB *) datafile[MAIN_PAL].dat, (RGB *) terraingfx[j].dat);

		// Set some variables for the map editor, making assumptions about the tilefile
		tileperiodicity = 16;
		tileareaw = 16;
		tileareah = 28;
	} */

/*	set_clip(screen, 0, 0, LogicalWidth, LogicalHeight * 2);
	set_clip(HiddenPage, 0, 0, LogicalWidth, LogicalHeight);
	set_clip(VisiblePage, 0, 0, LogicalWidth, LogicalHeight); */


	Log("Setting clipping");
	acquire_screen();
//	set_clip(screen, 0, 0, 0, 0);
	set_clip(HiddenPage, 0, 0, LogicalWidth - 1, LogicalHeight - 1);
	set_clip(VisiblePage, 0, 0, LogicalWidth - 1, LogicalHeight - 1);
	clear(screen);
    text_mode(-1);
	release_screen();

	Log("Set clipping and cleared the screen");

	// Initialize the message display system
	init_msg_system();

	Log("Initialized the messaging system");

//  Initialize the GUI / windowing system
	init_gui();

	Log("Initialized the GUI");

// Move the mouse out of the way to the bottom center of the screen
	position_mouse(SCREEN_W / 2, SCREEN_H);

	Log("Positioned the mouse");

//  Register timing interrupt handlers...
    LOCK_VARIABLE(simclock);
    LOCK_FUNCTION(inc_simclock);
    install_int_ex(inc_simclock, BPM_TO_TIMER(1092));

    LOCK_VARIABLE(fps);
    LOCK_VARIABLE(frame_count);
    LOCK_FUNCTION(frame_counter);
    install_int_ex(frame_counter, BPS_TO_TIMER(1));

	LOCK_VARIABLE(ms_time);
	LOCK_VARIABLE(ms_unit);
	LOCK_VARIABLE(ms_proj);
	LOCK_VARIABLE(ms_fly);
	LOCK_VARIABLE(ms_accel);
	LOCK_VARIABLE(ms_turn);
	LOCK_VARIABLE(ms_regen);
	LOCK_VARIABLE(ms_send);
    LOCK_FUNCTION(main_timer);
    install_int_ex(main_timer, BPS_TO_TIMER(1000));

	Log("Locked all functions");

    return 0;
}

/* page_flip(): -----------------------------------------------------------
 *   Flips the visible page and the working page.  Or, copies the double
 *   buffer to the screen if we're using double buffering.
 */

void page_flip(void)
{
//    static int firstpage = 1;  // (had to make this global)
    BITMAP *tmppage;

	if (gfx_method == PAGE_FLIP)
	{
	    tmppage = VisiblePage;
	    VisiblePage = HiddenPage;
	    HiddenPage = tmppage;

	    if (firstpage)
	    {
	        scroll_screen(32, LogicalHeight + 32); // 32, 336  // REAL: 32, 368
	        firstpage = 0;
	    }
	    else
	    {
	        scroll_screen(32, 32);  // 32, 32   // REAL: 32, 64
	        firstpage = 1;
	    }
	}
	else
	{
		acquire_screen();
		blit(HiddenPage, screen, 32, 32, 0, 0, LogicalWidth, LogicalHeight);
		release_screen();
	}
}

/* deallocate_memory(): ---------------------------------------------------
 *   This function should be called via atexit() to deallocate all memory.
 */
void deallocate_memory(void)
{
	struct BASE_NODE *tmpbase;
	struct MBASE_NODE *tmpmbase;

// Deallocate all allocated memory!!!!!!
//	printf("Deallocating memory..");
	if (debug)
		add_msg(MSG_COMPUTER, "Freeing helptext..");
	if (helptext)
		free(helptext);
	if (debug)
		add_msg(MSG_COMPUTER, "Freeing map, terrain, and unit_at..");
	if (map)
		free(map);
	if (frill)
		free(frill);
	if (terrain)
		free(terrain);
	if (terrain)
		free(vis);
	if (unit_at)
		free(unit_at);
	if (debug)
		add_msg(MSG_COMPUTER, "Freeing Stack..");
	if (Stack)
		free(Stack);
	if (debug)
		add_msg(MSG_COMPUTER, "Freeing bases..");
	while (base)
	{
		tmpbase = base->next;
		free(base);
		base = tmpbase;
	}
	if (debug)
		add_msg(MSG_COMPUTER, "Freeing mbases..");
	while (mbase)
	{
		tmpmbase = mbase->next;
		free(mbase);
		mbase = tmpmbase;
	}

	if (debug)
		add_msg(MSG_COMPUTER, "Freeing datafiles..");
	if (datafile)
		unload_datafile(datafile);
	if (halfdatafile)
		unload_datafile(halfdatafile);
	if (terraingfx)
		unload_datafile(terraingfx);
	if (shipgfx)
		unload_datafile(shipgfx);
	if (choppergfx)
		unload_datafile(choppergfx);
	if (terrain_bmp)
		destroy_bitmap(terrain_bmp);
	if (mapobj_bmp)
		destroy_bitmap(mapobj_bmp);
	if (vismap_bmp)
		destroy_bitmap(vismap_bmp);
	if (VisiblePage)
		destroy_bitmap(VisiblePage);
	if (HiddenPage)
		destroy_bitmap(HiddenPage);

	if (debug)
		add_msg(MSG_COMPUTER, "Freeing units..");
	destroy_all_units();

/*	if (s)
	{
		if (debug)
			add_msg(MSG_COMPUTER, "Freeing socket..");
		delete s;
	} */
//	if (comm_mode != COMM_NONE)
//		net_shutdown();
//	printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
}

/* shutdown_system(): -----------------------------------------------------
 *   Thanks to atexit(), this function is like the Maytag repair man..
 */
void shutdown_system(void)
{
	FILE *fp;
	int i;
#ifndef WIN32	
	struct date d;
	struct time t;
#endif

// If we were logging, log a closing message
	if (msg_buffer.log)
	{
		fp = fopen("msg.log", "a");
#ifndef WIN32
		getdate(&d);
		gettime(&t);
		fprintf(fp, "\n---Message log closed on %d/%d/%d at %d:%02d%cM---\n", d.da_mon, d.da_day, d.da_year, (t.ti_hour > 12) ? t.ti_hour - 12 : t.ti_hour, t.ti_min, (t.ti_hour > 12) ? 'P' : 'A');
#else
		fprintf(fp, "\n---Message log closed---\n");
#endif
		fclose(fp);
	}

// Get us back in text mode
    set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);

// Print bye-bye message
    printf("Congratulations.  You probably survived running this program!\n");
    printf("\nThis fairly early version of Bapple 2 was brought to you by Tom Grandgent.\n");
    printf("Please contact me at tgrand@canvaslink.com and tell me what you think!\n");
	printf("Or check out the web site at:  http://www.canvaslink.com/bapple2/\n\n");
#ifndef WIN32
	printf("Video mode:     %dx%dx%d\n", PhysicalWidth, PhysicalHeight, (int) pow2((double) depth));
#else	
	printf("Video mode:     %dx%dx%d-bit color\n", PhysicalWidth, PhysicalHeight, depth);
#endif	
	printf("Video driver:   %s - %s\n", gfx_name, gfx_desc);
	printf("Drawing method: %s\n\n", (gfx_method == PAGE_FLIP) ? "Page flipping" : "Double buffering");
	printf("%d frames / %.2f seconds = %.2f frames per second\n", total_frames, (float) ms_time / 1000.0, (float) (((float) total_frames) / (((float) ms_time) / 1000.0)));
	printf("\nThis is a -minimum- of %.0f pixels drawn per second!\n", (float) PhysicalWidth * (float) PhysicalHeight * (float) (((float) total_frames) / (((float) ms_time) / 1000.0)));

/*	for (i = 0; i < 256; i++)
	{
		if (__djgpp_memory_handle_list[i].handle == 0)
			break;
		printf("#%d: %08x\n", i, __djgpp_memory_handle_list[i].address);
		if (i % 23 == 0 && i > 0)
		{
			printf("More...");
			readkey();
		}
	} */

}

/* execution_help(): ------------------------------------------------------
 *   Displays help on the command line parameters.. like for "bapple2 -?"
 *   FIXME: This is fairly out of date :(
 */
void execution_help(void)
{
	printf("Please see bapple2.txt, and/or http://www.canvaslink.com/bapple2/\n");
	printf("Here are some examples of command line parameters supported:\n\n");
	printf("bapple2 -nosound                // Disables the sound-system\n");
	printf("bapple2 -depth 16               // Uses a color depth of 16 bits\n");
	printf("bapple2 -listen                 // Starts up a server\n");
	printf("bapple2 -connect 207.180.64.37  // Connects to a server\n");
	printf("bapple2 -port 1234              // Uses port 1234 for communications\n");
	printf("bapple2 -map ice                // Uses the map \"ice.map\"\n");
	printf("bapple2 -units 200              // Uses 200 randomly-placed units\n");
	printf("bapple2 -bases 8                // Uses 8 randomly-placed main-bases\n");
	printf("bapple2 -minibases 4            // Uses 4 randomly-placed mini-bases\n");

}

/* create_explosion(): ----------------------------------------------------
 *   Creates an explosion of the given type (large or small) at the
 *   given coordinates.  Imbalanced coordinates are ok.
 */
void create_explosion(char type, int x, int y, int xoff, int yoff)
{
    int e;
    char flag = NO;

    fix_coords(&x, &y, &xoff, &yoff);

    for (e = 0; e < active_explosions; e++)
    {
        if (explosion[e].type == EXPLOSION_NONEXISTENT)
        {
            flag = YES;
            break;
        }
    }
    if (flag == NO)
    {
        e = active_explosions++;
        if (active_explosions > MAX_EXPLOSIONS)
        {
            active_explosions--;
            return;
        }
    }
    explosion[e].x = x;
    explosion[e].y = y;
    explosion[e].xoff = xoff;
    explosion[e].yoff = yoff;
    explosion[e].frame = 0;
    explosion[e].last_explosion_time = simclock;
    explosion[e].type = type;
    switch (type)
    {
        case SMALL_EXPLOSION: explosion[e].numframes = 7; break;
        case BIG_EXPLOSION: explosion[e].numframes = 9; break;
    }
}

/* process_explosions(): --------------------------------------------------
 *   This will advance the frame of each explosion every now and then..
 *   It will also terminate explosions that have ended.
 */
void process_explosions(void)
{
    int e;

    for (e = 0; e < active_explosions; e++)
    {
        if (explosion[e].type == EXPLOSION_NONEXISTENT)
            continue;
        if (simclock - explosion[e].last_explosion_time >= EXPLOSION_RATE)
        {
            explosion[e].frame++;                                       
            if (explosion[e].frame >= explosion[e].numframes)
            {
                explosion[e].type = EXPLOSION_NONEXISTENT;
                continue;
            }
            explosion[e].last_explosion_time = simclock;
        }
    }
}

/* create_burst():
 *   Fires a circular burst of projectiles out from around the player.
 */

void create_burst(float speed, signed char type, signed char team, int x, int y, int xoff, int yoff)
{
    int i, p;
    signed int lxoff, lyoff, dir;
    static int angle = 0, inc = 1;
	float spd;
	PACKET send;

    i = 26;
    if (speed > 0.0)
        i += 6;
    else
        if (speed < 0.0)
            i -= 6;
	for (dir = 0; dir < 256; dir += 8)
	{
	    lxoff = fixtoi(fmul(fcos(itofix(dir)), itofix(i)));
	    lyoff = fixtoi(fmul(fsin(itofix(dir)), itofix(i)));
		if (SCREEN_H > 240)
		{
	    	lxoff -= 2;		// FIXME!  FIX ME!!!!!!!!
			lyoff -= 10;
		}
// If the player is stopped, don't add any speed to their projectiles
	    if (speed == 0)
			spd = (float) fixtof(player.maxspeed);
		else
			spd = (float) speed + 4.0f;
//    	create_projectile(spd, PROJECTILE_TYPE_YELLOW, team, x, y, (xoff + 5 + lxoff), (yoff + 2 + lyoff), dir);
////////////// CREATE PROJECTILE START

	    fix_coords(&x, &y, &xoff, &yoff);
	
	    for (p = 0; p < active_projectiles; p++)
	        if (proj[p].type == PROJ_NONEXISTENT)
	            break;
	    if (p == active_projectiles)
		{
	        p = active_projectiles++;
	        if (active_projectiles > MAX_PROJECTILES)
			{
	            active_projectiles--;
	            continue;
	        }
	    }
	    proj[p].direction = dir;
	    proj[p].x = x;
	    proj[p].y = y;
	    proj[p].xoff = (xoff + 5 + lxoff);
	    proj[p].yoff = (yoff + 2 + lyoff);
	    proj[p].type = type;
	    proj[p].team = team;
	    proj[p].speed = ftofix(spd);
		if (proj[p].type == PROJECTILE_TYPE_MISSILE)
		    proj[p].dietime = ms_time + 8000;
		else
		    proj[p].dietime = ms_time + 1200;

////////////// CREATE PROJECTILE END
	}
	// Tell the remote player(s) to create a burst
	if ((comm_mode != COMM_NONE) && (team == player.id))
	{
		for (i = 0; i < ((comm_mode == COMM_SERVER) ? active_enemies : 1); i++)
		{
			if (enemy[i].comm_state != CS_CONNECTED)
				continue;
			send.type = PACKET_BURST;
			send.x = x;
			send.y = y;
			send.xoff = xoff;
			send.yoff = yoff;
			send.speed = ftofix(speed);
			send.id = team;
	        Send(&send, sizeof(PACKET), enemy[i].addr);
		}
	}
	play_sound_vol((SAMPLE *) datafile[SND_SHOOT1].dat, x, y, 128, 128, 1000 + (rand() % 100) - 50, FALSE);
}

/* player_shoot(): --------------------------------------------------------
 *   This will create projectiles from the player at a set interval.
 */
void player_shoot(void)
{
    int i, g, e, closest;
    signed int xoff, yoff;
    static int angle = 0, inc = 1;
	fixed dist;
	unit_t *u, *uclosest;
	player_t *pclosest;

    for (g = 0; g < NUM_GUNS; g++)
    {
        if (gun_status[g] == ON)
            if (simclock - last_gun_time[g] >= gun_rate[g])
            {
                switch (g)
                {
                    case PROJECTILE_TYPE_YELLOW:
                    {
// These offsets are to create the projectile in -front- of the player
                        i = 15;
                        if (player.speed > 0)
                            i += 6;
                        else
                        if (player.speed < 0)
                            i -= 6;

// Fire LEFT projectile
                        xoff = fixtoi(fmul(fcos(itofix(player.direction - 25)), itofix(i)));
                        yoff = fixtoi(fmul(fsin(itofix(player.direction - 25)), itofix(i)));
						if (SCREEN_H > 240)
						{
                            xoff -= 2;		// FIXME!  FIX ME!!!!!!!!
							yoff -= 10;
						}
// If the player is stopped, don't add any speed to their projectiles
                        if (player.speed == 0)
                            create_projectile(fixtof(player.maxspeed), PROJECTILE_TYPE_YELLOW, player.id, 0, 0, player.x, player.y, (player.xoff + 5 + xoff), (player.yoff + 2 + yoff), player.direction);
                        else
                            create_projectile(4.0 + fixtof(player.speed), PROJECTILE_TYPE_YELLOW, player.id, 0, 0, player.x, player.y, (player.xoff + 5 + xoff), (player.yoff + 2 + yoff), player.direction);

// Fire RIGHT projectile
                        xoff = fixtoi(fmul(fcos(itofix(player.direction + 25)), itofix(i)));
                        yoff = fixtoi(fmul(fsin(itofix(player.direction + 25)), itofix(i)));
						if (SCREEN_H > 240)
						{
                            xoff -= 2;		// FIXME!  FIX ME!!!!!!!!
							yoff -= 10;
						}
// If the player is stopped, don't add any speed to their projectiles
                        if (player.speed == 0)
                            create_projectile(fixtof(player.maxspeed), PROJECTILE_TYPE_YELLOW, player.id, 0, 0, player.x, player.y, (player.xoff + 5 + xoff), (player.yoff + 2 + yoff), player.direction);
                        else
                            create_projectile(4.0 + fixtof(player.speed), PROJECTILE_TYPE_YELLOW, player.id, 0, 0, player.x, player.y, (player.xoff + 5 + xoff), (player.yoff + 2 + yoff), player.direction);

						play_sample((SAMPLE *) datafile[SND_SHOOT1].dat, 100, 128, 1000 + (rand() % 100) - 50, FALSE);
                        break;
                    }
                    case PROJECTILE_TYPE_MISSILE:
                    {
// These offsets are to create the projectile in -front- of the player
                        i = 15;
                        if (player.speed > 0)
                            i += 6;
                        else
                        if (player.speed < 0)
                            i -= 6;

// Fire forward projectile
                        xoff = fixtoi(fmul(fcos(itofix(player.direction)), itofix(i)));
                        yoff = fixtoi(fmul(fsin(itofix(player.direction)), itofix(i)));
						if (SCREEN_H > 240)
						{
                            xoff -= 2;		// FIXME!  FIX ME!!!!!!!!
							yoff -= 10;
						}

						// Find a target
						dist = itofix(9999);

						// Look around for enemies in multiplayer to seek after
						pclosest = NULL;
						if (comm_mode != COMM_NONE)
						{
							closest = 0;
							for (e = 0; e < active_enemies; e++)
								if (distance(player.x, player.y, enemy[e].x, enemy[e].y) < dist)
								{
                                    dist = distance(player.x, player.y, enemy[e].x, enemy[e].y);
									closest = e;
								}
							if (dist < itofix(10))
								pclosest = &enemy[closest];
						}
						// Look around for units to seek after
						if (dist >= itofix(10))
						{
							dist = itofix(9999);
							uclosest = NULL;
							for (u = units; u; u = u->next)
								if (distance(player.x, player.y, u->x, u->y) < dist)
								{
                                    dist = distance(player.x, player.y, u->x, u->y);
									uclosest = u;
								}
							if (dist >= itofix(10))
								uclosest = NULL;
						}

						// If the player is stopped, don't add any speed to their projectiles
                        if (player.speed == 0)
                            i = create_projectile(fixtof(player.maxspeed), PROJECTILE_TYPE_MISSILE, player.id, pclosest, uclosest, player.x, player.y, (player.xoff + 5 + xoff), (player.yoff + 2 + yoff), player.direction);
                        else
                            i = create_projectile(4.0 + fixtof(player.speed), PROJECTILE_TYPE_MISSILE, player.id, pclosest, uclosest, player.x, player.y, (player.xoff + 5 + xoff), (player.yoff + 2 + yoff), player.direction);

						play_sample((SAMPLE *) datafile[SND_SHOOT3].dat, 100, 128, 1000 + (rand() % 100) - 50, FALSE);
                        break;
                    }
                }
                last_gun_time[g] = simclock;
            }
    }
}

/* rotate_palette(): ------------------------------------------------------
 *   This cycles colors in a given range to provide an animation effect.
 */
void rotate_palette(int from, int to)
{
// FIXME: Change this function to use Allegro palette routines, so
// that it'll work with WinAllegro
#ifndef WIN32
    int a;
    char tr,tg,tb;
    char r[256];
    char g[256];
    char b[256];

    for (a = from; a < to + 1; a++)
    {
        outp(0x3C7, a);
        r[a] = inp(0x3C9);
        g[a] = inp(0x3C9);
        b[a] = inp(0x3C9);
    }

    tr = r[from];
    tg = g[from];
    tb = b[from];
    for (a = from; a < to; a++)
    {
        r[a] = r[a + 1];
        g[a] = g[a + 1];
        b[a] = b[a + 1];
    }
    r[to] = tr;
    g[to] = tg;
    b[to] = tb;

    for (a = from; a < to + 1; a++)
    {
        outp(0x03C8, a);
        outp(0x03C9, r[a]);
        outp(0x03C9, g[a]);
        outp(0x03C9, b[a]);
    }

#endif
}

/* draw_locator(): --------------------------------------------------------
 *   Draws a box showing the player position, as well as unit positions.
 *   Added support for mini-bases..  I should also do main-bases eventually.
 *
 *   Note: This function is not used anymore - the larger mini-map is used.
 */
void draw_locator(void)
{
    static int x1 = 0, y1, x2, y2;
    static unsigned int count = 0;
    int i;
    int basex, basey;
    int fx, fy;
    float bx1, by1, bx2, by2, px, py;
	struct BASE_NODE *node;
	unit_t *u;

    if (!x1)    // Only do these calculations the first time.
    {
        x1 = LogicalWidth - 83;
        y1 = LogicalHeight - 51 - 32;
        x2 = LogicalWidth - 33;
        y2 = LogicalHeight - 1 - 32;
    }
    rect(HiddenPage, x1, y1, x2, y2, TEXT_COLOR);

    // This plots points to show where the units are
	for (u = units; u; u = u->next)
    {
        bx1 = x1;
        bx2 = x2;
        by1 = y1;
        by2 = y2;
        px = u->x;
        py = u->y;

        fx = (int) (1 + (bx1 + (((px / (MAP_WIDTH + 3)) * (bx2 - bx1)))));
        fy = (int) (1 + (by1 + (((py / (MAP_HEIGHT + 3)) * (by2 - by1)))));
        putpixel(HiddenPage, fx, fy, TEXT_COLOR + 4);
    }

	for (node = base; node != NULL; node = node->next)
		for (basey = 0; basey < 3; basey++)
			for (basex = 0; basex < 3; basex++)
	        {
	            bx1 = x1;
	            bx2 = x2;
	            by1 = y1;
	            by2 = y2;
	            px = node->x + basex;
	            py = node->y + basey;

	            fx = (int) (1 + (bx1 + (((px / (MAP_WIDTH + 3)) * (bx2 - bx1)))));
	            fy = (int) (1 + (by1 + (((py / (MAP_HEIGHT + 3)) * (by2 - by1)))));
	            putpixel(HiddenPage, fx, fy, 36);
	        }

    // This is for the player location pinpoint
    // Save integer variables as floats to do this math coming up...
    bx1 = x1;
    bx2 = x2;
    by1 = y1;
    by2 = y2;
    px = player.x;
    py = player.y;

    fx = (int) (1 + (bx1 + (((px / (MAP_WIDTH + 3)) * (bx2 - bx1)))));
    fy = (int) (1 + (by1 + (((py / (MAP_HEIGHT + 3)) * (by2 - by1)))));
    putpixel(HiddenPage, fx, fy, TEXT_COLOR);

    if (total_frames % 4 == 0)
        count++;
    line(HiddenPage, fx, y1+1, fx, y2-1, 32 + (count % 15));
    line(HiddenPage, x1+1, fy, x2-1, fy, 32 + (count % 15));

	if (comm_mode != COMM_NONE)
	{
		for (i = 0; i < active_enemies; i++)
		{
			if (enemy[i].comm_state != CS_CONNECTED)
				continue;
		    // This is for the player location pinpoint
		    // Save integer variables as floats to do this math coming up...
		    bx1 = x1;
		    bx2 = x2;
		    by1 = y1;
		    by2 = y2;
		    px = enemy[i].x;
		    py = enemy[i].y;
		
		    fx = (int) (1 + (bx1 + (((px / (MAP_WIDTH + 3)) * (bx2 - bx1)))));
		    fy = (int) (1 + (by1 + (((py / (MAP_HEIGHT + 3)) * (by2 - by1)))));
		    putpixel(HiddenPage, fx, fy, TEXT_COLOR);
		
		    if (total_frames % 4 == 0)
		        count++;
		    line(HiddenPage, fx, y1+1, fx, y2-1, 96 + (count % 15));
		    line(HiddenPage, x1+1, fy, x2-1, fy, 96 + (count % 15));
		}
	}
}

/* initialize_projectiles(): ----------------------------------------------
 *   Initializes all of the projectiles...
 */
void initialize_projectiles(void)
{
    int i;

    active_projectiles = 0;

// Clear out all possible projectiles to PROJ_NONEXISTENT
    for (i = 0; i < MAX_PROJECTILES; i++)
        proj[i].type = PROJ_NONEXISTENT;
}

/* initialize_units(): ----------------------------------------------------
 *   Initialize all the units, creating some around the level in random spots
 */
void initialize_units(void)
{
    int i, x, y, good, type;

// Stop clients from creating initial units for now...
	if (comm_mode == COMM_CLIENT)
		return;

// Create some units in random places around the map
    for (i = 0; i < num_initial_units; i++)
    {
        good = 0;
        while (!good)
        {
			// Pick a random unit type
			type = rand() % NUM_UNITS;
			// Pick some random coordinates for it
            x = (rand() % ((MAP_WIDTH - 8) + 4));
            y = (rand() % ((MAP_HEIGHT - 8) + 4));
			// Check to see if it can exist at those coordinates
            if ((mpos(terrain, y, x) == unit_terrain[type]) && (!unit_exists_at(x, y)))
                good = 1;
        }
		// Make them have an owner ID of 1 for now... (enemy)
        create_unit(type, x, y, rand() % 256, (i > num_initial_units / 2) ? 1 : 0);
    }
}

/* fix_coords(): ----------------------------------------------------------
 * Fixes all coordinates if sub-tile (pixel) coordinates are > 31 or < 0
 * This allows boundless modification of sub-tile coordinates on the
 * application level, which eases the pain of the dual coord system.
 */
void fix_coords(int *x, int *y, signed int *xoff, signed int *yoff)
{
    while (*xoff < 0)
    {
        *xoff += 32;
        (*x)--;
    }
    while (*xoff > 31)
    {
        *xoff -= 32;
        (*x)++;
    }
    while (*yoff < 0)
    {
        *yoff += 32;
        (*y)--;
    }
    while (*yoff > 31)
    {
        *yoff -= 32;
        (*y)++;
    }
}

/* draw_sprites(): --------------------------------------------------------
 *   Draws all of the units, projectiles, and enemy players in the
 *   current view.
 */
void draw_sprites(void)
{
    int x, y, xoffset, yoffset, i, sx, sy, w, h, rx, ry, xo, yo, c, lx, ly;
	fixed dir;
	char drawn = 0;
	unit_t *u;
	int white = makecol(255, 255, 255);
	BITMAP *b;
	BITMAP *pbmp, *ubmp;
	static BITMAP *shadow = 0;

	pbmp = (BITMAP *) datafile[PROJECT1_PCX].dat;

	// Do shadow bitmap creation
	if (!shadow)
	{
		shadow = create_bitmap(pbmp->w, pbmp->h);
		blit(pbmp, shadow, 0, 0, 0, 0, pbmp->w, pbmp->h);
		for (y = 0; y < pbmp->h; y++)
			for (x = 0; x < pbmp->w; x++)
				if (getpixel(pbmp, x, y))
					putpixel(shadow, x, y, makecol(0, 0, 0));
	}

	xoffset = bgx;
 	yoffset = bgy;

	// Loop through each tile displayed on the screen and see if anyone is there
	// (I know this is really cheesy.. in fact, FIXME: make this more efficient)
    for (y = 32; y < LogicalHeight; y += 32)
    {
        for (x = 0; x <= LogicalWidth - 32; x += 32)
        {
/*			if (mpos(vis, yoffset, xoffset) == VIS_DARK)
			{
				xoffset++;
				continue;
			} */
// Draw units
			for (u = units; u; u = u->next)
				if ((u->x == xoffset) && (u->y == yoffset))
					if (u->carrier == -1)
					{
						sx = 32 + (x - bgxoff) + u->xoff;
						sy = (y - bgyoff) + u->yoff;
						b = (BITMAP *)datafile[UNIT1_PCX + u->type].dat;
						w = b->w;
						h = b->h;
						dir = itofix(u->direction);

						xo = sx + (w / 2);
						yo = sy + (h / 2);

						rx = xo - 15;
						ry = yo - abs(fixtoi(fsin(dir) * (h / 2))) - 8 - (int) ((float) w / (float) h * 3.0);

						rotate_sprite(HiddenPage, b, sx, sy, dir);

						// Draw unit health bar
						if (u->health > 0)
						{
							if ((float) u->health / (float) u->maxhealth < 0.25)
								c = makecol(255, 0, 0);
							else
							if ((float) u->health / (float) u->maxhealth < 0.66)
								c = makecol(255, 255, 0);
							else
								c = makecol(72, 117, 60);
	
							rectfill(HiddenPage, rx + 1, ry + 1, rx + (int) ((float) u->health / (float) u->maxhealth * 30.0), ry + 3, c);
							rect(HiddenPage, rx, ry, rx + 30, ry + 4, u->owner == player.id ? makecol(255, 255, 255) : makecol(255, 127, 127));
						}
					}

// Draw projectiles
            for (i = 0; i < active_projectiles; i++)
                if ((proj[i].x == xoffset) && (proj[i].y == yoffset))
                    if (proj[i].type != PROJ_NONEXISTENT)
//						if (datafile[PROJECT1_PCX + proj[i].type].type == DAT_XC_SPRITE)
						if (proj[i].type != PROJECTILE_TYPE_MISSILE)
						{
	                        draw_trans_sprite(HiddenPage, shadow, 32 + (x - bgxoff) + proj[i].xoff + 3, (y - bgyoff) + proj[i].yoff + 3);
	                        display_sprite(HiddenPage, datafile[PROJECT1_PCX + proj[i].type].dat, 32 + (x - bgxoff) + proj[i].xoff, (y - bgyoff) + proj[i].yoff);
						}
						else
                            rotate_sprite(HiddenPage, (BITMAP *)datafile[PROJECT1_PCX + proj[i].type].dat, 32 + (x - bgxoff) + proj[i].xoff, (y - bgyoff) + proj[i].yoff, itofix(proj[i].direction));
// Draw the enemy player(s)
//			if (comm_mode != COMM_NONE)
			{
				for (i = 0; i < active_enemies; i++)
				{
					if (enemy[i].comm_state != CS_CONNECTED)
						continue;
	                if ((enemy[i].x == xoffset) && (enemy[i].y == yoffset))
					{
						drawn = 1;
						if (!enemy[i].hidden)
						{
							sx = 32 + (x - bgxoff) + enemy[i].xoff;
							sy = (y - bgyoff) + enemy[i].yoff;
							b = (BITMAP *) shipgfx[enemy[i].shiptype].dat;
							w = b->w;
							h = b->h;
							dir = itofix(enemy[i].direction);
	
							xo = sx + (w / 2);
							yo = sy + (h / 2);

							// Draw the unit that the enemy is carrying (if any)
							if (enemy[i].unit_carried)
							{
								pbmp = (BITMAP *) shipgfx[enemy[i].shiptype].dat;
								ubmp = (BITMAP *) datafile[UNIT1_PCX + enemy[i].unit_carried->type].dat;
								lx = sx;
								ly = sy;
								lx += (pbmp->w - ubmp->w) / 2;
								ly += (pbmp->h - ubmp->h) / 2;
								rotate_sprite(HiddenPage, (BITMAP *) datafile[UNIT1_PCX + enemy[i].unit_carried->type].dat, lx, ly, itofix(enemy[i].direction));
							}
	
//							rx = xo - text_length(font, enemy[i].name) / 2;
							ry = yo - abs(fixtoi(fsin(dir) * (h / 2))) - 8 - (int) ((float) w / (float) h * 3.0);
							rotate_sprite(HiddenPage, b, sx, sy, dir);
							textout_centre(HiddenPage, font, enemy[i].name, xo, ry, TEXT_COLOR);
//							rotate_sprite(HiddenPage, (BITMAP *)shipgfx[enemy[i].shiptype].dat, 32 + (x - bgxoff) + enemy[i].xoff - 9, (y - bgyoff) + enemy[i].yoff - 8, dir);
						}
					}
				}
			}
// Draw the player (!)  I guess this got moved elsewhere..
//			if ((player.x == xoffset) && (player.y == yoffset))
//				rotate_sprite(HiddenPage, (BITMAP *)datafile[PLAYER_PCX].dat, 32 + (x - bgxoff) + player.xoff, (y - bgyoff) + player.yoff, itofix(player.direction));

            xoffset++;
        }
        xoffset = bgx;
        yoffset++;
    }

	// If the enemy was drawn on the screen, request a high
    // update packet frequency from them.
	// FIXME: This hasn't worked right since I rewrote the comm code for > 2 players!
	if (comm_mode != COMM_NONE)
	{
		// If the enemy update stream is not high frequency and it should be..
		if ((drawn) && (enemy_send_freq != HIGH_FREQ))
		{
			// Request a high frequency stream
			send_signal(PACKET_HIGH);
			enemy_send_freq = HIGH_FREQ;
		}
		else
		if ((!drawn) && (enemy_send_freq != LOW_FREQ))
		{
			// Request a low frequency stream
            send_signal(PACKET_LOW);
			enemy_send_freq = LOW_FREQ;
		}
	}
}

/* draw_explosions(): -----------------------------------------------------
 *   Draws all of the explosions in the current view.
 */
void draw_explosions(void)
{
    int x, y, xoffset, yoffset, i;
//	static BITMAP *circ = 0;

//	if (!circ)
//		circ = create_bitmap(64, 64);

    xoffset = bgx;
    yoffset = bgy;

//    for (y = 32; y <= LogicalHeight; y += 32)
    for (y = 32; y < LogicalHeight - 32; y += 32)
    {
        for (x = 0; x <= LogicalWidth - 32; x += 32)
        {
// Draw explosions
/*          if (edit_mode)  // Draw the numbers all over the place!
            {
                sprintf(buf, "%d", map[((yoffset + 1) % MAP_HEIGHT)][(xoffset % MAP_WIDTH)]);
                textout(HiddenPage, font, buf, x + 16, y + 16, TEXT_COLOR);
            } */
            for (i = 0; i < active_explosions; i++)
                if ((explosion[i].x == xoffset) && (explosion[i].y == yoffset))
                    switch (explosion[i].type)
                    {
                        case EXPLOSION_NONEXISTENT: break;
                        case SMALL_EXPLOSION:
						{
//							clear(circ);
//							circlefill(circ, circ->w / 2, circ->h / 2, explosion[i].frame, makecol(255, 255, 255));
							display_sprite(HiddenPage, datafile[SMALEXP1_PCX + explosion[i].frame].dat, 32 + (x - bgxoff) + explosion[i].xoff + small_exp_xoff[explosion[i].frame], (y - bgyoff) + explosion[i].yoff + small_exp_yoff[explosion[i].frame]);
//							draw_trans_sprite(HiddenPage, circ, 32 + (x - bgxoff) + explosion[i].xoff - circ->w / 2 + 4, (y - bgyoff) + explosion[i].yoff - circ->h / 2 + 4);
							break;
						}
                        case BIG_EXPLOSION:
						{
							display_sprite(HiddenPage, datafile[BIGEXP1_PCX + explosion[i].frame].dat, 32 + (x - bgxoff) + explosion[i].xoff + big_exp_xoff[explosion[i].frame], (y - bgyoff) + explosion[i].yoff + big_exp_yoff[explosion[i].frame]);
							break;
						}
                    }
            xoffset++;
        }
        xoffset = bgx;
        yoffset++;
    }
}

/* distance(): ------------------------------------------------------------
 *   Returns the distance (as a fixed point #) between two given points.
 */
fixed distance(signed int x1, signed int y1, signed int x2, signed int y2)
{
    fixed xsq, ysq, x, y;

    x = itofix(x2 - x1);
    y = itofix(y2 - y1);
//    x /= 8;
//    y /= 8;
    xsq = fmul(x, x);
    ysq = fmul(y, y);
    return (fsqrt(xsq + ysq));
}

/* process_units(): -------------------------------------------------------
 *   This updates the unit_at map (containing 1 if a unit occupies that point
 *   on the map).  It then processes the orders of all of the units, but only
 *   every now and then (at a set interval).  This includes moving the units.
 */
void process_units(void)
{
    int i, j;
    char buf[100];
	unit_t *u;
	wnd_t *w;
	char txtbuf[100];

    i = 0;

	// Only run the unit AI once every 50 milliseconds
    while (ms_unit > 50)  // (5) If it's been more than 50 milliseconds, update units
    {
        ms_unit -= 50;

// First thing to do is populate the unit_at array which tells what
// spaces are blocked by units...
/*
// Start by clearing it out:
		memset(unit_at, 0, (MAP_WIDTH * MAP_HEIGHT));

// Now populate it by looping through the active units:
		for (u = units; u; u = u->next)
	        if (u != player.unit_carried)
			{
                j = u->y * MAP_WIDTH + u->x;
//				unit_at[j] = i;

                if (u->path)
                {
//                	if (u->path->x == u->x)
//                    	if (u->path->y == u->y)
                        {
							unit_at[j] = 1;
//                            map[j]++;
//                            map[j] %= NUM_TILES;
                        }
                }
                else
                {
					unit_at[j] = 1;
//                    map[j]++;
//                    map[j] %= NUM_TILES;
				}
			}
*/

		// Check for units that should be dead
		for (u = units; u; u = u->next)
        {
            if (u == player.unit_carried)
                continue;
// Kill the unit if it deserves to die... :)
            if (u->health <= 0)
            {
                create_explosion(BIG_EXPLOSION, u->x, u->y, u->xoff - 8, u->yoff - 8);
				play_sound_vol((SAMPLE *) datafile[SND_EXPLOSION].dat, u->x, u->y, 128, 128, 1000 + (rand() % 400) - 200, FALSE);
				destroy_unit(u);

				if (debug)
				{
					sprintf(buf, "Unit %d with ID %d destroyed", i, u->id);
					add_msg(MSG_COMPUTER, buf);
				}
                continue;
            }
// Process unit orders
			process_orders(u);

// Make unit turn towards the direction that they want to go in
            modify_direction(&u->direction, u->targetdir, 8);

            fix_coords(&(u->x), &(u->y), &(u->xoff), &(u->yoff));

/*
// Make unit wrap if they go off the edge of the map
            if (u->x < 2)
               u->x = MAP_WIDTH - 2;
            if (u->y < 2)
               u->y = MAP_HEIGHT - 2;
            if (u->x > MAP_WIDTH - 2)
               u->x = 2;
            if (u->y > MAP_HEIGHT - 2)
               u->y = 2; */
        }
    }

	// Process unit ordering if we're ordering something
	if (order_state >= 0)
	{
		if (ms_time - order_time > unit_order_time[order_state])
		{
			if (!unit_ready)	// Notify the player that a unit is ready
			{
				add_msg(MSG_COMPUTER, "A %s is ready and waiting to be picked up", unit_name[order_state]);
				unit_ready = 1;
				for (w = wnd_list; w; w = w->next)
					if (w->input == input_orderwnd)
					{
						if (w->title)
							free(w->title);
						sprintf(txtbuf, "%s Ready", unit_name[order_state]);
						w->title = (char *) malloc(strlen(txtbuf) + 1);
						strcpy(w->title, txtbuf);
					}
			}
		}
	}
	else
		unit_ready = 0;
}

void destroy_projectile(projectile_struct *p)
{
	p->type = PROJ_NONEXISTENT;
	remove_thing(THING_PROJECTILE, p, p->x, p->y);
}

/* process_projectiles(): -------------------------------------------------
 *   Move the projectiles around, destroy them if they've been around for
 *   too long, and perform collision detection (unit <-> projectile and
 *   player <-> projectile).  Teams are not yet taken into account...
 */
void process_projectiles(void)
{
    int e, i, j, missile_dir, x, y;
	int lastgood;
	struct BASE_NODE *node;
	struct MBASE_NODE *mnode;
	char buf[80];
	PACKET_HURT_T hurtpacket;
	unit_t *u;
	BITMAP *ubmp;
	thing_t *thing;
	projectile_struct *tp;

	// Only process projectiles once every 20 milliseconds
    while (ms_proj > 20)  // If it's been more than 20 milliseconds, update projectiles
    {
        ms_proj -= 20;
        i = non = 0;

		if (edit_mode && player.speed)
	        create_explosion(SMALL_EXPLOSION, player.x, player.y, player.xoff + 4, player.yoff);

// Check for collision between enemy projectiles and the player
			for (y = player.y - 1; y < player.y + 1; y++)
				for (x = player.x - 1; x < player.x + 1; x++)
					for (thing = mpos(unit_at, y, x); thing; thing = thing->next)
						if (thing->type == THING_PROJECTILE)
						{
							tp = (projectile_struct *) thing->ptr;
				            if ((abs(((tp->x * 32) + tp->xoff) - ((player.x * 32) + player.xoff + 4)) < 12))
				                if ((abs(((tp->y * 32) + tp->yoff) - ((player.y * 32) + player.yoff)) < 12))
								{
									if (tp->team != player.id)
					                {
					                    create_explosion(SMALL_EXPLOSION, tp->x, tp->y, tp->xoff + ((rand() % 4) - 2), tp->yoff + ((rand() % 4) - 2));
										if ((!invincible) && (player.health > 0))
										{
					        	            player.health -= 5;
											if (debug)
											{
												sprintf(buf, "You have been hit by ID %d", tp->team);
												add_msg(MSG_COMPUTER, buf);
											}
											player.lastattacker = tp->team;
										}
										play_sample((SAMPLE *) datafile[SND_DAMAGE].dat, 128, 128, 1000 - (rand() % 200), FALSE);
										destroy_projectile(tp);
										thing = mpos(unit_at, y, x);
										if (!thing)
											break;
									}
								}
						}

		// Old fashioned collision detection
        for (i = 0; i < active_projectiles; i++)
        {
            if (proj[i].type == PROJ_NONEXISTENT)
            {
                non++;
                continue;
            }
 
			// If a projectile has expired, destroy it
            if (ms_time > proj[i].dietime)
            {
				destroy_projectile(&proj[i]);
				create_explosion(SMALL_EXPLOSION, proj[i].x, proj[i].y, proj[i].xoff, proj[i].yoff);
                continue;
            }
			lastgood = i;

// If in multiplayer, check for collision between player projectiles and enemy
			if ((proj[i].team == player.id) && (comm_mode != COMM_NONE))
				for (e = 0; e < active_enemies; e++)
				{
					if (enemy[e].comm_state != CS_CONNECTED)
						continue;
		            if ((abs(((proj[i].x * 32) + proj[i].xoff) - ((enemy[e].x * 32) + enemy[e].xoff + 4)) < 12))
	    	            if ((abs(((proj[i].y * 32) + proj[i].yoff) - ((enemy[e].y * 32) + enemy[e].yoff)) < 12))
		                {
		                    create_explosion(SMALL_EXPLOSION, proj[i].x, proj[i].y, proj[i].xoff + ((rand() % 4) - 2), proj[i].yoff + ((rand() % 4) - 2));
							play_sample((SAMPLE *) datafile[SND_DAMAGE].dat, 128, 128, 1000 - (rand() % 200), FALSE);
							destroy_projectile(&proj[i]);

							// Send a hurt packet to whoever we hurt
							hurtpacket.type = PACKET_HURT;
							hurtpacket.victim = enemy[e].id;
							hurtpacket.victor = player.id;
							hurtpacket.health = 5;
							if (comm_mode == COMM_CLIENT)
			                    Send(&hurtpacket, sizeof(PACKET_HURT_T), enemy[0].addr);
							else if (comm_mode == COMM_SERVER)
								for (j = 0; j < active_enemies; j++)
								{
									if (enemy[j].comm_state != CS_CONNECTED)
										continue;
									if (hurtpacket.victim == enemy[j].id)
									{
					                    Send(&hurtpacket, sizeof(PACKET_HURT_T), enemy[j].addr);
										break;
									}
								}
						}
				}

// If in multiplayer, check for collision between enemy projectiles and other enemies
            if ((comm_mode != COMM_NONE) && (proj[i].type != PROJ_NONEXISTENT))
			{
				for (e = 0; e < active_enemies; e++)
				{
					if (enemy[e].comm_state != CS_CONNECTED)
						continue;
					if (enemy[e].id == proj[i].team)
						continue;
		            if ((abs(((proj[i].x * 32) + proj[i].xoff) - ((enemy[e].x * 32) + enemy[e].xoff + 4)) < 12))
	    	            if ((abs(((proj[i].y * 32) + proj[i].yoff) - ((enemy[e].y * 32) + enemy[e].yoff)) < 12))
		                {
		                    create_explosion(SMALL_EXPLOSION, proj[i].x, proj[i].y, proj[i].xoff + ((rand() % 4) - 2), proj[i].yoff + ((rand() % 4) - 2));
							play_sample((SAMPLE *) datafile[SND_DAMAGE].dat, 128, 128, 1000 - (rand() % 200), FALSE);
							destroy_projectile(&proj[i]);
						}
				}
			}

// Check for collision between projectiles and other units

			for (y = proj[i].y - 1; y < proj[i].y + 1; y++)
				for (x = proj[i].x - 1; x < proj[i].x + 1; x++)
					for (thing = mpos(unit_at, y, x); thing; thing = thing->next)
						if (thing->type == THING_UNIT)
						{
							u = (unit_t *) thing->ptr;
							ubmp = (BITMAP *)datafile[UNIT1_PCX + u->type].dat;
			
							if ((abs(((proj[i].x * 32) + proj[i].xoff) - ((u->x * 32) + u->xoff + ubmp->w / 2)) < ubmp->w / 2))
								if ((abs(((proj[i].y * 32) + proj[i].yoff) - ((u->y * 32) + u->yoff + ubmp->h / 2)) < ubmp->h / 2))
									if ((u != player.unit_carried)) // && ((proj[i].type != PROJECTILE_TYPE_MISSILE) || (proj[i].team == PROJECTILE_TEAM_UNIT)))
									{
										create_explosion(SMALL_EXPLOSION, proj[i].x, proj[i].y, proj[i].xoff + ((rand() % 8) - 4), proj[i].yoff + ((rand() % 8) - 4));
										u->health -= 40;
										play_sound_vol((SAMPLE *) datafile[SND_DAMAGE].dat, u->x, u->y, 100, 128, 1000 + (rand() % 200), FALSE);
										destroy_projectile(&proj[i]);
										continue;
									}
						}

// Check for collision between projectiles and mini bases
			for (node = base; node != NULL; node = node->next)
				if ((proj[i].x >= node->x) && (proj[i].x < (node->x + 3)))
					if ((proj[i].y >= node->y) && (proj[i].y < (node->y + 3)))
	                {
	                    create_explosion(SMALL_EXPLOSION, proj[i].x, proj[i].y, proj[i].xoff, proj[i].yoff);
//						play_sample((SAMPLE *) datafile[SND_DAMAGE].dat, 128, 128, 1000 - (rand() % 200), FALSE);
						destroy_projectile(&proj[i]);
	                    continue;
					}
// Check for collision between projectiles and main bases
			if (proj[i].team == PROJECTILE_TEAM_UNIT)
			for (mnode = mbase; mnode != NULL; mnode = mnode->next)
				if ((proj[i].x >= mnode->x) && (proj[i].x < (mnode->x + 5)))
					if ((proj[i].y >= mnode->y) && (proj[i].y < (mnode->y + 5)))
	                {
	                    create_explosion(SMALL_EXPLOSION, proj[i].x, proj[i].y, proj[i].xoff, proj[i].yoff);
                        mnode->health -= 50;
						if (mnode->health < 0)
							mnode->health = 0;
//						play_sample((SAMPLE *) datafile[SND_DAMAGE].dat, 128, 128, 1000 - (rand() % 200), FALSE);
						destroy_projectile(&proj[i]);
	                    continue;
					}

			// Handle missile seeking
            if (proj[i].type == PROJECTILE_TYPE_MISSILE)
			{
				if (proj[i].ptarget)	// target is a player
					head_towards(proj[i].x, proj[i].y, proj[i].xoff + 11, proj[i].yoff + 4, proj[i].ptarget->x, proj[i].ptarget->y, proj[i].ptarget->xoff + 15, proj[i].ptarget->yoff + 15, &missile_dir);
				else
				if (proj[i].utarget)	// target is a unit
					head_towards(proj[i].x, proj[i].y, proj[i].xoff + 11, proj[i].yoff + 4, proj[i].utarget->x, proj[i].utarget->y, proj[i].utarget->xoff + 15, proj[i].utarget->yoff + 15, &missile_dir);
				modify_direction(&proj[i].direction, missile_dir, 4);
			}

			// Move the projectile
			remove_thing(THING_PROJECTILE, &proj[i], proj[i].x, proj[i].y);
			proj[i].xoff += fixtoi(fmul(fcos(itofix(proj[i].direction)), proj[i].speed));
			proj[i].yoff += fixtoi(fmul(fsin(itofix(proj[i].direction)), proj[i].speed));
			fix_coords(&proj[i].x, &proj[i].y, &proj[i].xoff, &proj[i].yoff);
			add_thing(THING_PROJECTILE, &proj[i], proj[i].x, proj[i].y);

			// Terminate projectiles that go off the edge of the world
            if (proj[i].x < 2)
            	proj[i].type = PROJ_NONEXISTENT;
            if (proj[i].y < 1)
            	proj[i].type = PROJ_NONEXISTENT;
            if (proj[i].x > MAP_WIDTH - 2)
            	proj[i].type = PROJ_NONEXISTENT;
            if (proj[i].y > MAP_HEIGHT - 2)
            	proj[i].type = PROJ_NONEXISTENT;

			if (proj[i].type == PROJ_NONEXISTENT)
				destroy_projectile(&proj[i]);
        }
    }
}

/* ingame_help(): ---------------------------------------------------------
 *   Display a crappy and hopelessly out of date help screen in the game.
 *   I should just nuke this.  Let em use F1 instead.
 */
void ingame_help(void)
{

    textout(VisiblePage, (FONT *) datafile[SMALL_FONT].dat, "Bapple2 Quick Help", 90, 64, TEXT_COLOR);
    textout(VisiblePage, (FONT *) datafile[SMALL_FONT].dat, "Use arrow keys for movement.", 32, 100, TEXT_COLOR);
    textout(VisiblePage, (FONT *) datafile[SMALL_FONT].dat, "Space bar retrieves/places units.", 32, 120, TEXT_COLOR);
    textout(VisiblePage, (FONT *) datafile[SMALL_FONT].dat, "Hold down shift for afterburners.", 32, 140, TEXT_COLOR);
    textout(VisiblePage, (FONT *) datafile[SMALL_FONT].dat, "Hold down S or B for visual effects.", 32, 160, TEXT_COLOR);
    textout(VisiblePage, (FONT *) datafile[SMALL_FONT].dat, "Hit ALT to nuke everyone.", 32, 180, TEXT_COLOR);
    textout(VisiblePage, (FONT *) datafile[SMALL_FONT].dat, "Hit E to display player coordinates.", 32, 200, TEXT_COLOR);
    textout(VisiblePage, (FONT *) datafile[SMALL_FONT].dat, "Hit escape to quit.", 32, 220, TEXT_COLOR);

    readkey();
}

//////////////// B E G I N   P A T H   C O D E ///////////////

/* DrawPath(): ------------------------------------------------------------
 *   Draws a path from (x1,y1) to (x2,y2) on the map, creating walls
 *   and setting the displayed tiles to "tilenum".
 *   Only for testing purposes.
 */
void DrawPath(int x1, int y1, int x2, int y2, int tilenum)
{
  struct NODE *path, *Node;

  desired_terrain = mpos(terrain, player.y, player.x);
  path=(struct NODE *)FindPath((long)x1,(long)y1,(long)x2,(long)y2);
  if (path == NULL)
  {
    return;
  }

  mpos(map, y1, x1) = tilenum;
  mpos(map, y2, x2) = tilenum;

  while (path->Parent != NULL)
  {
    mpos(map, path->y, path->x) = tilenum;
    mpos(terrain, path->y, path->x) = TERRAIN_BLOCKED;
    path=path->Parent;
  }
  mpos(map, path->y, path->x) = tilenum;

  // Free Nodes up
  Node=OPEN->NextNode;
  while (Node != NULL)
  {
    free(Node);
    Node=Node->NextNode;
  }
  Node=CLOSED->NextNode;
  while (Node != NULL)
  {
    free(Node);
    Node=Node->NextNode;
  }

}

/* FreeTile(): ------------------------------------------------------------
 *   This important function identifies to A* whether or not a tile is free
     for units to move onto.  I suspect that it could return values other
     than 0 and 1 for use in a cost-based transit system, though the
     heuristic would probably then become grossly inefficient if not changed.
 */
int FreeTile(int x, int y)
{
    if ((mpos(terrain, y, x) != desired_terrain) || (unit_exists_at(x, y)))
        return 0;
    else
        return 1;
}

/* TileNum(): -------------------------------------------------------------
 *   This strange little function is used to uniquely identify each location
 *   on the map as a single int, because A* needs that form of information.
 */
int TileNum(int x, int y)   // This translates from two dimensions to one
{
  return ((y * MAP_WIDTH) + x);
}

////////////////// E N D   P A T H   C O D E /////////////////

/* find_closest_free_tile(): ----------------------------------------------
 *   Given pointers to some coordinates, this function will try to change
 *   those coordinates to the closest tile that is not currently occupied.
 */
void find_closest_free_tile(int *dx, int *dy)
{
    signed int x, y, tx, ty, outward;

    tx = *dx;
    ty = *dy;

	if ((tx < 0) || (tx >= MAP_WIDTH) || (ty < 0) || (ty >= MAP_HEIGHT))
		return;

    if (FreeTile(tx, ty))
        return;

    outward = 1;
    do {
        for (y = ty - outward; y < ty + outward; y++)
		{
			if ((y >= 0) && (y < MAP_HEIGHT))
	            for (x = tx - outward; x < tx + outward; x++)
				{
					if ((x >= 0) && (x < MAP_WIDTH))
		                if (FreeTile(x, y))
		                {
		                    *dx = x;
		                    *dy = y;
		                    return;
		                }
				}
		}
    } while (outward++ < MAP_WIDTH);   // go at most MAP_WIDTH tiles away
}

/* screen_shot(): ---------------------------------------------------------
 *   Takes a screen shot of everything on the visible page.  I'm not
 *   sure if this mechanism is entirely reliable though...
 */
void screen_shot(void)
{
    BITMAP *bmp;
    PALETTE pal;
	int x, y;

	if (gfx_method == DOUBLE_BUFFER)
		x = y = 0;
	else
		x = y = 32;

    get_palette(pal);
	bmp = create_sub_bitmap(VisiblePage, x, y, PhysicalWidth, PhysicalHeight);
    save_bitmap("dump.pcx", bmp, pal);
    destroy_bitmap(bmp);
}

extern msg_buffer_t msg_buffer;

/* display_info(): --------------------------------------------------------
 *   This will display various bits of information on the screen like health.
 */
void display_info(void)
{
    char buf[255];
	int i;
	unsigned long latency;
	static unsigned long peak_latency = 0;
	struct MBASE_NODE *mnode;
	static int initial_base_health = -1;

	// First time this function is called, count up the collective health of the main-bases
	if (initial_base_health == -1)
	{
		initial_base_health = 0;
		for (mnode = mbase; mnode != NULL; mnode = mnode->next)
			initial_base_health += mnode->health;
	}

    if (player.health < 0)
        player.health = 0;

    if (edit_mode)
    {
        sprintf(buf, "%d,%d  %d,%d  %d %d", player.x, player.y, player.xoff, player.yoff, mpos(map, player.y, player.x), mpos(terrain, player.y, player.x));
        textout(HiddenPage, font, buf, 32, 100, TEXT_COLOR);
        sprintf(buf, "proj: %d ", (active_projectiles - non));
        textout(HiddenPage, font, buf, 170, 64, TEXT_COLOR);

	    sprintf(buf, "bgx = %d   bgy = %d", bgx, bgy);
	    textout(HiddenPage, font, buf, 32, 78, TEXT_COLOR);

	    sprintf(buf, "bgxoff = %d   bgyoff = %d", bgxoff, bgyoff);
	    textout(HiddenPage, font, buf, 32, 110, TEXT_COLOR);

	    sprintf(buf, "dist: %f", fixtof(distance(player.x, player.y, penx, peny)));
	    textout(HiddenPage, (FONT *) datafile[SMALL_FONT].dat, buf, 32, 140, TEXT_COLOR);

		sprintf(buf, "slide: %d,%d", bgxslide, bgyslide);
	    textout(HiddenPage, font, buf, 32, 130, TEXT_COLOR);

		sprintf(buf, "speed: %d", fixtoi(player.speed));
	    textout(HiddenPage, font, buf, 150, 130, TEXT_COLOR);

		sprintf(buf, "dir: %d", player.direction);
	    textout(HiddenPage, font, buf, 250, 130, TEXT_COLOR);

		textprintf(HiddenPage, font, 32, 170, TEXT_COLOR, "(%d,%d) - (%f,%f)", player.xoff, player.yoff, player.fxoff, player.fyoff);

		textprintf(HiddenPage, font, 400, 170, TEXT_COLOR, "input=%d", player.input);

//		sprintf(buf, "Units: %d/%d", active_units, NUM_UNITS);
//	    textout(HiddenPage, font, buf, 32, 160, TEXT_COLOR);

		// Count up the current total health in all of the main-bases
		i = 0;
		for (mnode = mbase; mnode != NULL; mnode = mnode->next)
			i += mnode->health;

		if (!initial_base_health)
			initial_base_health = 1;
		sprintf(buf, "Base health: %d/%d (%.1f%%)", i, initial_base_health, (float) i / (float) initial_base_health * 100.0);
	    textout(HiddenPage, font, buf, 32, 190, TEXT_COLOR);

		sprintf(buf, "Allocated: %d", heapused);
	    textout(HiddenPage, font, buf, 32, 220, TEXT_COLOR);
    }

    sprintf(buf, "%d fps", fps);
    textout(HiddenPage, font, buf, 32 + PhysicalWidth - text_length(font, buf), 32, TEXT_COLOR);

	if ((comm_mode != COMM_NONE) && (debug))
	{
		latency = ms_time - last_packet_time;
		if (latency > peak_latency)
			peak_latency = latency;
	    sprintf(buf, "silence: %dms", latency);
	    textout(HiddenPage, font, buf, 32 + PhysicalWidth - text_length(font, buf), 42, TEXT_COLOR);
	    sprintf(buf, "peak: %dms", peak_latency);
	    textout(HiddenPage, font, buf, 32 + PhysicalWidth - text_length(font, buf), 52, TEXT_COLOR);
	}
//    sprintf(buf, "Health: %d%%", player.health);
//    textout(HiddenPage, font, buf, 32, LogicalHeight - 42, TEXT_COLOR);

//    sprintf(buf, "Fuel: %d%%", (player.fuel / 100));
//    textout(HiddenPage, font, buf, 160, LogicalHeight - 42, TEXT_COLOR);

	if (game_state == GAME_STATE_MAP)
		rectfill(HiddenPage, 32, LogicalHeight - 57, 132, LogicalHeight - 33, 0);

	if (comm_mode != COMM_NONE)
	{
		for (i = 0; i < active_enemies; i++)
		{
			if (enemy[i].comm_state == CS_NONEXISTENT)
				continue;
			if (enemy[i].comm_state != CS_CONNECTED)
				enemy[i].health = 0;
			// Print the enemy name, kills, deaths, and ID number
			if (debug)
				sprintf(buf, "%s-%d [%d/%d] S%d R%d L%d PL%.1f (%dms)", enemy[i].name, enemy[i].id, enemy[i].kills, enemy[i].deaths, enemy[i].sentpackets, enemy[i].recvpackets, enemy[i].seqnum, 100.0 - (float) enemy[i].recvpackets / (float) (enemy[i].seqnum + 1) * 100.0, ms_time - enemy[i].latency);
			else
				sprintf(buf, "%s [%d]", enemy[i].name, enemy[i].kills - enemy[i].deaths);
			textprintf(HiddenPage, font, 134, LogicalHeight - (66 + (i * 9)), TEXT_COLOR, buf);
			// Fill in their health rect
		    rectfill(HiddenPage, 32, LogicalHeight - (66 + (i * 9)), 32 + enemy[i].health, LogicalHeight - (60 + (i * 9)), makecol(174, 165, 72));
			// Draw the bordering white rect
		    rect(HiddenPage, 32, LogicalHeight - (66 + (i * 9)), 132, LogicalHeight - (60 + (i * 9)), makecol(255, 255, 255));
		}
	}

	// Main base health meter
	if (player.mbase)
	    rectfill(HiddenPage, 32, LogicalHeight - 57, 32 + player.mbase->health / 100, LogicalHeight - 51, makecol(72, 145, 222));
    rect(HiddenPage, 32, LogicalHeight - 57, 132, LogicalHeight - 51, makecol(255, 255, 255));
	textprintf(HiddenPage, font, 134, LogicalHeight - 57, TEXT_COLOR, "$%d", player.money);

	// Player health meter
	if (debug)
		sprintf(buf, "%s-%d [%d/%d]", player.name, player.id, player.kills, player.deaths);
	else
		sprintf(buf, "%s [%d]", player.name, player.kills - player.deaths);
	textprintf(HiddenPage, font, 134, LogicalHeight - 48, TEXT_COLOR, buf);
    rectfill(HiddenPage, 32, LogicalHeight - 48, 32 + player.health, LogicalHeight - 42, makecol(190, 56, 64));
    rect(HiddenPage, 32, LogicalHeight - 48, 132, LogicalHeight - 42, makecol(255, 255, 255));

	// Player fuel meter
    rectfill(HiddenPage, 32, LogicalHeight - 39, 32 + (player.fuel / 100), LogicalHeight - 33, makecol(113, 141, 97));
    rect(HiddenPage, 32, LogicalHeight - 39, 132, LogicalHeight - 33, makecol(255, 255, 255));

//        sprintf(buf, "Player direction: %d", player.direction);
//        textout(HiddenPage, datafile[FRITZ].dat, buf, 32, 100, TEXT_COLOR);
//        sprintf(buf, "Milliseconds: %lu", ms_time);
//        textout(HiddenPage, datafile[FRITZ].dat, buf, 32, 120, TEXT_COLOR);

/*        fx = player.x;
        fy = player.y;
        fx = fx * fx;
        fy = fy * fy;
        fx = fx + fy;
        dist = sqrt(fx);
        x_printf(32, 250, HiddenPageOffs, TEXT_COLOR, "Dist from 0,0: %f", dist); */
////        x_printf(32, 280, HiddenPageOffs, TEXT_COLOR, "Health: %d %", player.health);   //health_display
//      x_printf(70+fix, LogicalHeight-15, HiddenPageOffs, TEXT_COLOR, "Score: %ld", score);
//      x_printf(135+fix, LogicalHeight-15, HiddenPageOffs, TEXT_COLOR, "Kills: %d", kills);

/*    x1 = (unit[0].x * 32) + (unit[0].xoff);
    y1 = (unit[0].y * 32) + (unit[0].yoff);
    x2 = (unit[1].x * 32) + (unit[1].xoff);
    y2 = (unit[1].y * 32) + (unit[1].yoff);
    sprintf(buf, "x1: %d x2: %d distance: %f", x1, x2, fixtof(distance(x1, y1, x2, y2)));
    sprintf(buf, "distance: %f", fixtof(distance(unit[0].x, unit[0].y, unit[1].x, unit[1].y)));
    textout(HiddenPage, datafile[FRITZ].dat, buf, 32, 80, TEXT_COLOR); */
}

/* draw_player(): ---------------------------------------------------------
 * Draws the player.  Takes a page, as the player may be needed on the
 * Visible -or- Hidden page..
 */
void draw_player(BITMAP *page)
{
	int x, y, i;
	BITMAP *pbmp, *ubmp;
	static BITMAP *rot = 0;
	static BITMAP *shadow = 0;
	static int shiptype = 666;

	pbmp = (BITMAP *) shipgfx[player.shiptype].dat;

	// Do shadow bitmap creation
#ifdef SHADOWS
	if (!rot)
		rot = create_bitmap(pbmp->w > pbmp->h ? pbmp->w : pbmp->h, pbmp->h > pbmp->w ? pbmp->h : pbmp->w);
	if ((!shadow) || (shiptype != player.shiptype))
	{
		if (shadow)
			destroy_bitmap(shadow);
		shiptype = player.shiptype;
		shadow = create_bitmap(pbmp->w, pbmp->h);
		blit(pbmp, shadow, 0, 0, 0, 0, pbmp->w, pbmp->h);
		for (y = 0; y < pbmp->h; y++)
			for (x = 0; x < pbmp->w; x++)
				if (getpixel(pbmp, x, y))
					putpixel(shadow, x, y, makecol(0, 0, 0));
	}
#endif

    if ((player.health > 0) || (simclock - death_start_time < 18*3))
	{
		if (player.unit_carried)
		{
			pbmp = (BITMAP *) shipgfx[player.shiptype].dat;
			ubmp = (BITMAP *) datafile[UNIT1_PCX + player.unit_carried->type].dat;
			x = 32 + centerx;
			y = centery;
			x += (pbmp->w - ubmp->w) / 2;
			y += (pbmp->h - ubmp->h) / 2;
			acquire_screen();
			rotate_sprite(page, (BITMAP *) datafile[UNIT1_PCX + player.unit_carried->type].dat, x, y, itofix(player.direction));
			release_screen();
		}

//        if (!player.unit_carried)
		acquire_screen();
#ifdef SHADOWS
		clear(rot);
		rotate_sprite(rot, shadow, 0, 0, itofix(player.direction));
		draw_trans_sprite(page, rot, 32 + centerx + 6, centery + 6);
#endif
		// Handle helicopter animation
		if (ms_time > player.rotortime)
		{
			player.rotortime = ms_time + 40;
			player.rotorframe++;
			if (player.rotorframe > 3)
				player.rotorframe = 0;
		}
		while (player.direction < 0)
			player.direction += 256;
		while (player.direction > 255)
			player.direction -= 256;
		i = player.direction >> 4;
		i += 16 * player.rotorframe;
		draw_sprite(page, (BITMAP *) choppergfx[i].dat, 32 + centerx - 16, centery - 5);
//		rotate_sprite(page, (BITMAP *) shipgfx[player.shiptype].dat, 32 + centerx, centery, itofix(player.direction));
		release_screen();
//        else
//            rotate_sprite(page, (BITMAP *) shipgfx[player.shiptype].dat, 32 + centerx, centery, itofix(player.direction));
	}
}

/* process_classic_player_input(): ----------------------------------------
 *   Processes player movement and turning, in the "classic" way.
 *   This means that whatever direction you press, that is the direction
 *   that your ship will try to move in.
 */
void process_classic_player_input(signed int *ldir)
{
    static signed int ndown = 0, sdown = 0, wdown = 0, edown = 0;
    signed int target;
    fixed speed;

// Handle modifying the ship direction based on which arrows are held down
        target = -1;
        if (ndown && edown)                 // Northeast
            target = 224;
        else
        if (ndown && wdown)                 // Northwest
            target = 160;
        else
        if (sdown && wdown)                 // Southwest
            target = 96;
        else
        if (sdown && edown)                 // Southeast
            target = 32;
        else
        if (edown)                          // East
            target = 0;
        else
        if (ndown)                          // North
            target = 192;
        else
        if (wdown)                          // West
            target = 128;
        else
        if (sdown)                          // South
            target = 64;
        else
            target = player.direction;

// *** Process player turning

// If they're not carrying anything..
        if (!player.unit_carried)
        {
            while (ms_turn > 30)        // And it's time for them to turn... (30)
            {
                ms_turn -= 30;
                if (target != -1)
                    modify_direction(&player.direction, target, 8);		// (8)
            }

        }
        else    // They are carrying a unit
        {
            while (ms_turn > 50)        // Increased interval of turning..
            {
                ms_turn -= 50;
                if (target != -1)
                    modify_direction(&player.direction, target, 8);
            }

        }

// Update the direction of motion if they're heading in the right direction
       if (player.direction == target)
          *ldir = player.direction;
       else
          modify_direction(ldir, player.direction, 2);

// Player input processing... moving
        // Increase acceleration if an arrow key has been held down
        // Or decrease it if a key was not held down...
        // But, to slow down general acceleration and deceleration,
        //  use the # of frames variable to only change these values
        //  every couple of frames.
        while (ms_accel > 40)
        {
            ms_accel -= 40;
            if (wdown || ndown || edown || sdown)
            {
                if (player.direction == target)
                {
                    if (player.speed == 0)
                        player.speed += ftofix(0.75);				// was 0.75
                    player.speed = player.speed + ftofix(0.25);		// was 0.25
                }
//                else
//                    player.speed = player.speed - ftofix(0.25);
            }
            else
            {
                if (player.direction == target)
                    player.speed = player.speed - ftofix(1.0);		// was 1.0
            }
            if (fixtof(player.speed) < 0.0)
                player.speed = ftofix(0.0);
        }

        // We can't go accelerating out of control! :o
        // Don't let the speed go over the player.maxspeed limit.
        // Double the speed if afterburners are on though. (shift)
        speed = player.maxspeed;
        if (afterburner)
            speed += speed;
        if (player.unit_carried)
            speed = fmul(speed, ftofix(0.6));
        if (player.speed > speed)
            player.speed = speed;

        // Reset the "key was held down last frame" flags
        wdown = edown = ndown = sdown = 0;

        // Set the "key was held down last frame" flags
		// Check keyboard input
        if (key[KEY_UP])
            ndown = 1;
        if (key[KEY_DOWN])
            sdown = 1;
        if (key[KEY_LEFT])
            wdown = 1;
        if (key[KEY_RIGHT])
            edown = 1;
		// Check joystick input (if joystick is enabled)
		if (joystick_on)
		{
			poll_joystick();
	        if (joy_up)
	            ndown = 1;
	        if (joy_down)
	            sdown = 1;
	        if (joy_left)
	            wdown = 1;
	        if (joy_right)
	            edown = 1;
		}
}

/* process_modern_player_input(): -----------------------------------------
 *   Processes player movement and turning, in the "modern" way.
 *   This means that when you press up, you go forwards.  When you press
 *   down, you go backwards.  Left and right actually turn the ship.
 */
void process_modern_player_input(signed int *ldir)
{
    fixed speed;
	char ndown = 0, sdown = 0, wdown = 0, edown = 0;

// Get input
		// Check keyboard input
        if (key[KEY_UP])
            ndown = 1;
        if (key[KEY_DOWN])
            sdown = 1;
        if (key[KEY_LEFT])
            wdown = 1;
        if (key[KEY_RIGHT])
            edown = 1;
		// Check joystick input (if joystick is enabled)
		if (joystick_on)
		{
			poll_joystick();
	        if (joy_up)
	            ndown = 1;
	        if (joy_down)
	            sdown = 1;
	        if (joy_left)
	            wdown = 1;
	        if (joy_right)
	            edown = 1;
		}

		if (ndown)
			player.input |= INPUT_FORWARD;
		else
			player.input &= ~INPUT_FORWARD;
		if (sdown)
			player.input |= INPUT_BACK;
		else
			player.input &= ~INPUT_BACK;
		if (wdown)
			player.input |= INPUT_LEFT;
		else
			player.input &= ~INPUT_LEFT;
		if (edown)
			player.input |= INPUT_RIGHT;
		else
			player.input &= ~INPUT_RIGHT;

// *** Process player turning

// If they're not carrying anything..
        if (!player.unit_carried)
        {
            while (ms_turn > 20)        // And it's time for them to turn... (40)
            {
                ms_turn -= 20;
                if (wdown)
                    player.direction -= 4;									// (8)
                else
                if (edown)
                    player.direction += 4;
            }

        }
        else    // They are carrying a unit
        {
            while (ms_turn > 50)        // Increased interval of turning..
            {
                ms_turn -= 50;
                if (wdown)
                    player.direction -= 8;
                else
                if (edown)
                    player.direction += 8;
            }

        }

// Player input processing... moving
        // Increase acceleration if an arrow key has been held down
        // Or decrease it if a key was not held down...
        // But, to slow down general acceleration and deceleration,
        //  use the # of frames variable to only change these values
        //  every couple of frames.
        while (ms_accel > 40)
        {
            ms_accel -= 40;
            if (ndown)    // If up is held down, accelerate
            {
                if (player.speed == 0)
                    player.speed += ftofix(0.50);				// was 0.75
                player.speed = player.speed + ftofix(0.15);		// was 0.25
            }
            else
            if (sdown)  // Go backwards..
            {
                if (player.speed == 0)
                    player.speed -= ftofix(0.75);
                player.speed = player.speed - ftofix(0.25);
            }
            else                // Otherwise, decelerate
            {
                if (fixtof(player.speed >= 0.0))
                {
                    player.speed = player.speed - ftofix(0.35);	// was 0.75
                    if (fixtof(player.speed) < 0.0)
                        player.speed = ftofix(0.0);
                }
                else
                {
                    player.speed = player.speed + ftofix(0.35); // was 0.75
                    if (fixtof(player.speed) > 0.0)
                        player.speed = ftofix(0.0);
                }
            }
        }

        // We can't go accelerating out of control! :o
        // Don't let the speed go over (or under) the player.maxspeed limit.
        // Double the speed if afterburners are on though. (shift)
        speed = player.maxspeed;
        if (afterburner)
            speed += speed;
        if (player.unit_carried)
            speed /= 2;
        if (player.speed > speed)
            player.speed = speed;
        else
        if (player.speed < -speed)
            player.speed = -speed;

        *ldir = player.direction;
}

/* render_world(): --------------------------------------------------------
 *   Renders the entire world (tiles, units, player) out to a huge WORLD.PCX.
 */
void render_world(void)
{
    BITMAP *bmp;
    PALETTE pal;
    int x, y, xoffset, yoffset;
    char buf[80];
	unit_t *u;

    get_palette(pal);
    bmp = create_bitmap((32 * MAP_WIDTH), (32 * MAP_HEIGHT));
	if (bmp == NULL)
	{
        sprintf(buf, "Couldn't allocate %d bytes!", 32 * MAP_WIDTH * 32 * MAP_HEIGHT);
		acquire_screen();
        textout(VisiblePage, font, buf, 32, 128, TEXT_COLOR);
		release_screen();
		rest(2000);
		return;
	}

    xoffset = yoffset = 0;

    text_mode(0);

    for (y = 0; y < (32 * MAP_HEIGHT); y += 32)
    {
        for (x = 0; x < (32 * MAP_WIDTH); x += 32)
        {
//            display_sprite(screen, datafile[(TILE01_PCX + mpos(map, yoffset, xoffset))].dat, 0, 0);
			acquire_screen();
            display_sprite(screen, terraingfx[mpos(map, yoffset, xoffset)].dat, 0, 0);
            blit(screen, bmp, 0, 0, x, y, 32, 32);
            sprintf(buf, "Processing tile: (%d,%d)  ", xoffset, yoffset);
            textout(VisiblePage, font, buf, 32, 128, TEXT_COLOR);
			release_screen();
            xoffset++;
        }
        xoffset = 0;
        yoffset++;
    }

    xoffset = yoffset = 0;

// Draw units
	for (u = units; u; u = u->next)
    {
        if (u != player.unit_carried)
            rotate_sprite(bmp, (BITMAP *) datafile[UNIT1_PCX + u->type].dat, (u->x * 32) + u->xoff, (u->y * 32) + u->yoff, itofix(u->direction));
    }

// Draw the player
    if (!player.unit_carried)
        rotate_sprite(bmp, (BITMAP *) shipgfx[player.shiptype].dat, (player.x * 32), (player.y * 32), itofix(player.direction));
    else
        rotate_sprite(bmp, (BITMAP *) shipgfx[player.shiptype].dat, (player.x * 32), (player.y * 32), itofix(player.direction));

    sprintf(buf, "Saving bitmap to disk...  ");
	acquire_screen();
    textout(VisiblePage, font, buf, 32, 128, TEXT_COLOR);
	release_screen();

    save_bitmap("world.pcx", bmp, pal);
    destroy_bitmap(bmp);

    text_mode(-1);
}

/* play_sound_vol(): ------------------------------------------------------
 *   Plays a sound in the world at a volume based on the distance between
 *   the player and the given coordinates (the sound's world location).
 *   Also takes into account panning now (woo!)
 */
int play_sound_vol(SAMPLE *spl, int x, int y, int vol, int pan, int freq, int loop)
{
	float dist;

	dist = fixtof(distance(player.x, player.y, x, y));

	if (dist < HEARING_THRESHOLD)// Don't play sounds too far away from player
	{
		vol -= (int) (dist / HEARING_THRESHOLD * (float) vol);
		if (vol < 0)
			vol = 0;

		pan = (int) (player.x < x ?
			(1.0 + (x - player.x) / HEARING_THRESHOLD) * 128 :
			(1.0 - (player.x - x) / HEARING_THRESHOLD) * 128);
		if (pan < 0)
			pan = 0;
		else
			if (pan > 255)
				pan = 255;
		play_sample(spl, vol, pan, freq, loop);
	}

	return 0;
}

/* replace_color():
 *   Takes a bitmap and replaces all occurances of one color with another
 */
void replace_color(BITMAP *bmp, int oldcolor, int newcolor)
{
	int x, y;

	for (y = 0; y < bmp->h; y++)
		for (x = 0; x < bmp->w; x++)
			if (getpixel(bmp, x, y) == oldcolor)
				putpixel(bmp, x, y, newcolor);
}

/* recompile_bitmap():
 *   Takes a mode-x compiled bitmap and returns a linear compiled bitmap.
 */
COMPILED_SPRITE *recompile_bitmap(COMPILED_SPRITE *cbitmap)
{
	BITMAP *bmp;
	COMPILED_SPRITE *spr;

    bmp = create_bitmap_ex(depth, cbitmap->w, cbitmap->h);
	clear_to_color(bmp, bitmap_mask_color(bmp));
    rectfill(screen, 0, 0, cbitmap->w, cbitmap->h, bitmap_mask_color(screen));
    draw_compiled_sprite(screen, cbitmap, 0, 0);
    blit(screen, bmp, 0, 0, 0, 0, cbitmap->w, cbitmap->h);
	replace_color(bmp, 0, bitmap_mask_color(bmp));
	spr = get_compiled_sprite(bmp, FALSE);
	destroy_bitmap(bmp);
	return spr;
}

/* convert_to_linear():
 *   Converts all compiled bitmaps used in a given datafile to linear bitmaps.
 */
void convert_to_linear(DATAFILE *datafile)
{
	int i;
    COMPILED_SPRITE *bmp;
	BITMAP *bitmap;

//	for (i = TILE01_PCX; i < (TILE01_PCX + NUM_TILES); i++)
	for (i = 0; i < ZZZ_MARKER; i++)
	{
		if (datafile[i].type == DAT_BITMAP)
		{
			bitmap = create_bitmap_ex(depth, ((BITMAP *) datafile[i].dat)->w, ((BITMAP *) datafile[i].dat)->h);
			clear_to_color(bitmap, bitmap_mask_color(bitmap));
			blit((BITMAP *) datafile[i].dat, bitmap, 0, 0, 0, 0, bitmap->w, bitmap->h);
			replace_color(bitmap, 0, bitmap_mask_color(bitmap));
//			draw_sprite(bitmap, (BITMAP *) datafile[i].dat, 0, 0);
			destroy_bitmap((BITMAP *) datafile[i].dat);
            datafile[i].dat = bitmap;
		}
		else
        if (datafile[i].type == DAT_XC_SPRITE)
		{
            bmp = recompile_bitmap((COMPILED_SPRITE *) datafile[i].dat);

			destroy_compiled_sprite((COMPILED_SPRITE *) datafile[i].dat);
			datafile[i].dat = bmp;

			// Now for the half size gfx..
/*			destroy_compiled_sprite((COMPILED_SPRITE *) halfdatafile[i].dat);
			src = create_bitmap(bmp->w, bmp->h);
			bitmap = create_bitmap((bmp->w) / 2, (bmp->h) / 2);
			draw_compiled_sprite(src, (COMPILED_SPRITE *) datafile[i].dat, 0, 0);

			u = v = 0;
            for (y = 0; y < bmp->h; y += 2)
			{
				for (x = 0; x < bmp->w; x += 2)
					putpixel(bitmap, u++, v, getpixel(src, x, y));
				u = 0;
				v++;
			}
			halfdatafile[i].dat = (COMPILED_SPRITE *) get_compiled_sprite(bitmap, FALSE);
			destroy_bitmap(bitmap);
			destroy_bitmap(src); */
		}
	}

/*	for (i = 0; i < TERRAIN_TILES; i++)
	{
        if (terraingfx[i].type == DAT_XC_SPRITE)
		{
            bmp = recompile_bitmap((COMPILED_SPRITE *) terraingfx[i].dat);

			destroy_compiled_sprite((COMPILED_SPRITE *) terraingfx[i].dat);
			terraingfx[i].dat = bmp;
		}
	} */
}

/* display_sprite():
 *   Wrapper function to draw a sprite or tile..
 */
void display_sprite(BITMAP *bmp, void *sprite, int x, int y)
{
	COMPILED_SPRITE *spr = (COMPILED_SPRITE *) sprite;

//#ifndef WIN32
	if ((x >= 0) && (y >= 0) && (x <= VIRTUAL_W - spr->w + 64) && (y <= VIRTUAL_H - spr->h))
		draw_sprite(bmp, (BITMAP *) sprite, x, y);
//		draw_compiled_sprite(bmp, (COMPILED_SPRITE *) sprite, x, y);
//#endif
}

/* zero_timers():
 *   Resets all of the timers (for initialization only I hope)
 */
void zero_timers(void)
{
	ms_time = 0;  ms_unit = 0;   ms_proj = 0;
	ms_fly = 0;   ms_accel = 0;  ms_turn = 0;
	ms_regen = 0; ms_send = 0;
}

/* toggle_timers():
 *   Toggles whether or not the timers are active (by saving/restoring them)
 *   This is for when the main game loop gets frozen by some event...
 */
void toggle_timers(void)
{
	static char paused = 0;
	static unsigned long bk_time, bk_unit, bk_proj, bk_fly, bk_accel, bk_turn, bk_regen, bk_send;

	if (!paused)
	{
		// Save all of the timer states..
		bk_time = ms_time;   bk_unit = ms_unit;    bk_proj = ms_proj;
		bk_fly = ms_fly;     bk_accel = ms_accel;  bk_turn = ms_turn;
		bk_regen = ms_regen; bk_send = ms_send;
		paused = 1;
	}
	else
	{
		// Restore all of the timer states
		ms_time = bk_time;   ms_unit = bk_unit;    ms_proj = bk_proj;
		ms_fly = bk_fly;     ms_accel = bk_accel;  ms_turn = bk_turn;
		ms_regen = bk_regen; ms_send = bk_send;
		paused = 0;
	}
}

/* do_game_processing():
 *   The part of the main game loop that doesn't draw to the screen..
 */
void do_game_processing(void)
{
	if (talking)
		process_talking();	// Process talking (typing) input
	else
	    process_input();    // Process input (but not player directions)
    process_player();       // Process player state and direction input
    process_units();        // Process the unit states (incl moving them)
    process_projectiles();  // Process the projectiles (move, hit, etc)
    process_explosions();   // Process the explosions (advance frame..)
	process_bases();		// Process the bases (like destroy em)
	process_animations();	// Process the animations (like flags)
	process_messages();		// Process the messages in the message buffer
}

/* do_game_drawing():
 *   The part of the main game loop that draws to the screen..
 */
void do_game_drawing(void)
{
	int y;

    tiled_background();     // Display the tiled background

    draw_sprites();         // Draw units and then projectiles

    draw_player(HiddenPage);           // Draw the player

/*    if (mpos(map, player.y, player.x) == 11) // Island damage to player.. woo
    {
        create_explosion(SMALL_EXPLOSION, player.x, player.y, player.xoff + ((rand() % 16) - 8), player.yoff + ((rand() % 16) - 8));
		if (!invincible)
	        player.health -= 1;
    } */
    draw_explosions();      // Draw the explosions

	process_gui();			// Process and draw the GUI

	draw_messages();		// Draw the messages in the message buffer

    if (!(total_frames % 10)) // FIXME: Horrible timing system.. bleh!
        rotate_palette(240, 244);   // Do some color cycling animation
/*	if (!use_big_map)
	    draw_locator();         // Draw the little position indicator
	else
		display_big_map(); */
}

// +=====================================================================+
// |=*=-=*=-=*=-=*=-=/|( M A I N   G A M E   L O O P )|\=*=-=*=-=*=-=*=-=|
// +=====================================================================+
void do_game(void)
{

	while (1)
    {
		switch (game_state)
		{
			case GAME_STATE_PLAY:
			{
				do_game_processing();
				do_game_drawing();
				display_info();         // Display framerate and other info

				if (game_state != GAME_STATE_PLAY)
				{
					clear(HiddenPage);
					acquire_screen();
					clear(VisiblePage);
					release_screen();
				}
				break;
			}
            case GAME_STATE_HELP:
            {
				do_game_processing();
				display_help();
				break;
            }
			case GAME_STATE_MAP:
			{
				do_game_processing();
                display_big_map();
		        display_info();         // Display framerate and other info
				break;
			}
            case GAME_STATE_INITCONN:
			{
				if (keypressed())
					if ((readkey() >> 8) == 1)
						game_state = GAME_STATE_EXIT;
                clear(HiddenPage);
				draw_messages();
				process_messages();		// Process the messages in the message buffer
				display_big_map();
		        display_info();         // Display framerate and other info
				break;
			}
			case GAME_STATE_EXIT:
			{
                return;
			}
		}

        page_flip();            // Make the working page visible!
        frame_count++;          // Increase frame counter (to find fps)
		total_frames++;			// Increase total frame counter
		do_communications();	// Process communications
    }
}

/* send_signal():
 *   Sends a single-byte packet that consists only of a packet type.
 */
void send_signal(char packet_type)
{
	char buf;
	int i;

	if ((comm_mode != COMM_NONE) && (conn_type == CONN_WSOCK))
	{
		if (active_enemies)
		{
			if (comm_mode == COMM_CLIENT)
			{
				buf = packet_type;
				Send(&buf, 1, enemy[0].addr);
			}
			else if (comm_mode == COMM_SERVER)
			{
				buf = packet_type;
				for (i = 0; i < active_enemies; i++)
					if (enemy[i].comm_state == CS_CONNECTED)
	                    Send(&buf, 1, enemy[i].addr);
			}
		}
	}
}

/////// M A I N ///////

int main(int argc, char *argv[])
{
    int i;
    int main_loop_delay = 0;

    if (argc > 1)   // Handle arguments
    {
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "/?", 2) || !strnicmp(argv[i], "?", 1) || !strnicmp(argv[i], "-?", 2))
            {
                execution_help();
                return 0;
            }
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-q", 2))
                quick_start = 1;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-delay", 8))
                main_loop_delay = atoi(argv[(i + 1)]);
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-depth", 6))
                depth = atoi(argv[(i + 1)]);
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-alt", 4))
                alt_input = 1;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-classic", 8))
                alt_input = 1;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-nosound", 8))
                sound_on = 0;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-joy", 4))
                joystick_on = 1;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-slide", 6))
                slide_enabled = 1;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-locator", 8))
                use_big_map = 0;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-buffer", 7))
                gfx_method = DOUBLE_BUFFER;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-map", 4))
                strcpy(terrainfile, argv[i + 1]);
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-units", 6))
                num_initial_units = atoi(argv[(i + 1)]);
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-bases", 6))
                num_initial_mainbases = atoi(argv[(i + 1)]);
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-mainbases", 10))
                num_initial_mainbases = atoi(argv[(i + 1)]);
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-minibases", 10))
                num_initial_minibases = atoi(argv[(i + 1)]);

// Communications parameters
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-wsock", 6))
                conn_type = CONN_WSOCK;
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-listen", 7) || !strnicmp(argv[i], "-server", 7))
			{
				if (conn_type == CONN_NONE)
					conn_type = CONN_WSOCK;
                comm_mode = COMM_SERVER;
			}
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-connect", 8))
			{
				if (conn_type == CONN_NONE)
					conn_type = CONN_WSOCK;
                comm_mode = COMM_CLIENT;
				strcpy(server_addr, argv[(i + 1)]);
			}
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-sendfreq", 9))
				send_frequency = atoi(argv[(i + 1)]);
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-port", 5))
				WSOCK_PORT = atoi(argv[(i + 1)]);
        for (i = 1; i < argc; i++)
            if (!strnicmp(argv[i], "-win", 4) || !strnicmp(argv[i], "-windows", 8))
				i_love_bill = 1;
    }

// Run normal system initialization
    if (init_system())
        return 1;

	do_game();

    shutdown_system();          // Perform system shutdown functions...

    return 0;                   // Normal program termination
}

// WHEW!  That's the end!!  For now...
