// map.c: Map related functions for Bapple2.  Created on September 2nd, 1997.
// This includes the GUI :)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <math.h>

#include <allegro.h>
#include "bapple2.h"
#include "bap2dat.h"
#include "globals.h"
#include "msg.h"
#include "map.h"

char scatter_windows = 0;		// Heh, this is fun.

// Auto-focusing for windows on mouse rollover (X-style)
char auto_focus = 0;

// Memory tile
int memtile = 0;
int memtilew = 1;
int memtileh = 1;
int memtileperiodicity = 0;
short int memtiles[100][100];
int memterrain[100][100];

// Tile source window dimensions (in tiles)
int tilesrcw = 4, tilesrch = 4;
// total tile area size (in tiles)
int tileareaw = 8, tileareah = 8;
// Periodicity (how many tiles per row of the master tile bitmap)
int tileperiodicity = 8;
// Tile source window x and y offsets (in tiles)
int tilexoff = 0, tileyoff = 0;

// Flag for whether or not the user is selecting tiles from the world (map)
char selecting_world = 0;

// Pointer to a bitmap for use as a mouse cursor
BITMAP *mouse_cursor;

// Map functions

int save_map(char *filename);
int load_map(char *filename);
int load_vista(char *filename, int sealevel, int smoothing);
void smooth_vista(BITMAP *bmp);
void fix_coastline(void);
void process_gui(void);
void init_gui(void);
void init_editor_gui(void);
void destroy_windows(void);
void destroy_window(wnd_t *wnd);
void process_mouse_input(void);
wnd_t *create_wnd(int x, int y, int w, int h, char style, char *title, void (*draw)(wnd_t *w), int (*input)(wnd_t *w), void *data);
void bring_to_front(wnd_t *w);
void draw_window_frame(wnd_t *wnd);
void paste_memtile(int tilex, int tiley);
void fill_map_with_clipboard(void);
void random_paste(int num);
void draw_button(int x1, int y1, int x2, int y2, unsigned char style);
void draw_text_button(int x1, int y1, char *str);
char findunit(int x, int y);
void load_terrain(char *filename);
void load_frill_gfx(char *filename);

// Window handlers
void draw_tilesrc(wnd_t *w);
void draw_tilewnd(wnd_t *w);
void draw_palette(wnd_t *w);
void draw_memgraph(wnd_t *w);
void draw_projgraph(wnd_t *w);
void draw_minimap(wnd_t *w);
void draw_unitinfo(wnd_t *w);
void draw_orderwnd(wnd_t *w);
int input_tilesrc(wnd_t *w);
int input_tilewnd(wnd_t *w);
int input_plainresize(wnd_t *w);
int input_easymove(wnd_t *w);
int input_orderwnd(wnd_t *w);
int input_minimap(wnd_t *w);

// From msg.cc:

void add_msg(char type, char *text, ...);

// save_map(): Saves the current map using the given filename
int save_map(char *filename)
{
	PACKFILE *fp;
	map_header_t mh;
#ifndef WIN32
	struct date d;
	struct time t;
#endif

	// Clear out the map header struct to all zeroes
	memset(&mh, 0, sizeof(mh));

	// Fill the map header struct
	mh.magic = 314159;
	mh.w = MAP_WIDTH;
	mh.h = MAP_HEIGHT;
#ifndef WIN32
	getdate(&d);
	gettime(&t);
	sprintf(mh.name, "%dx%d map saved from Bapple2 on %d/%d/%d at %d:%02d%cM", mh.w, mh.h, d.da_mon, d.da_day, d.da_year, (t.ti_hour > 12) ? t.ti_hour - 12 : t.ti_hour, t.ti_min, (t.ti_hour > 12) ? 'P' : 'A');
#else
	sprintf(mh.name, "%dx%d map saved from Bapple2", mh.w, mh.h);
#endif
	strcpy(mh.author, player.name);
	strcpy(mh.tilefile, tilefile);

	// Add an extension to the filename if one has not been given
	if (!strstr(filename, "."))
		strcat(filename, ".map");

    // Try to open the output map file
	fp = pack_fopen(filename, F_WRITE_PACKED);

	// If there's a problem opening the file, tell the user
	if (!fp)
	{
		add_msg(MSG_COMPUTER, "Unable to open %s for writing!", filename);
		return 1;
	}

	// Write the map header
	pack_fwrite(&mh, sizeof(map_header_t), fp);

	// Write the terrain type data
	pack_fwrite(terrain, mh.w * mh.h, fp);
	// Write the tile index data (note - short ints for every tile, NOT CHARS, hence the 2)
	pack_fwrite(map, mh.w * mh.h * 2, fp);

	// Close the map file, we're done with it
	pack_fclose(fp);

	// Give the user a confirmation message that the map has been saved
	add_msg(MSG_COMPUTER, mh.name);

	return 0;
}

// load_map(): Loads a map using the given filename
int load_map(char *filename)
{
	PACKFILE *fp;
	map_header_t mh;
	int i, j, x, y;
	PALETTE oldpal, thepal;

	// Add an extension to the filename if one has not been given
	if (!strstr(filename, "."))
		strcat(filename, ".map");

    // Try to open the input map file
	fp = pack_fopen(filename, F_READ_PACKED);

	// If there's a problem opening 7the file, tell the user
	if (!fp)
	{
		add_msg(MSG_COMPUTER, "Unable to open %s for reading!", filename);
		return 1;
	}

	// Read the map header
	if (pack_fread(&mh, sizeof(map_header_t), fp) == -1)
	{
		add_msg(MSG_COMPUTER, "Opened %s for reading, but reading failed!", filename);
		return 1;
	}

	if (mh.magic != 314159)
	{
		add_msg(MSG_COMPUTER, "Error: %s is not a valid Bapple2 map file.", filename);
		return 1;
	}

	// Get info from the map header struct
	MAP_WIDTH = mh.w;
	MAP_HEIGHT = mh.h;
	strcpy(tilefile, mh.tilefile);

	// Reallocate memory for the new map
	// Free memory...
	if (terrain_bmp)
		free(terrain_bmp);
	if (mapobj_bmp)
		free(mapobj_bmp);
	if (vismap_bmp)
		free(vismap_bmp);
	if (map)
		free(map);
	if (frill)
		free(frill);
	if (terrain)
		free(terrain);
	if (vis)
		free(vis);
	if (unit_at)
		free(unit_at);
    // Reallocate memory...
    terrain_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
	mapobj_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
	vismap_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
	map = (short *) calloc(1, MAP_WIDTH * MAP_HEIGHT * 2);
	frill = (short *) calloc(1, MAP_WIDTH * MAP_HEIGHT * 2);
	terrain = (signed char *) calloc(1, MAP_WIDTH * MAP_HEIGHT);
	vis = (char *) calloc(1, MAP_WIDTH * MAP_HEIGHT);
	unit_at = (thing_t **) calloc(1, MAP_WIDTH * MAP_HEIGHT * sizeof(thing_t *));

	// Clear the newly allocated bitmaps
	clear(terrain_bmp);
	clear(mapobj_bmp);
	clear(vismap_bmp);

	// Read the terrain type data
	if (pack_fread(terrain, mh.w * mh.h, fp) == -1)
	{
		add_msg(MSG_COMPUTER, "Opened %s for reading, failed while reading terrain data!", filename);
		return 1;
	}
	// Read the tile index data (these are SHORTS, not CHARS, hence the 2)
	if (pack_fread(map, mh.w * mh.h * 2, fp) == -1)
	{
		add_msg(MSG_COMPUTER, "Opened %s for reading, failed while reading tile index data!", filename);
		return 1;
	}
/*	for (y = 0; y < MAP_HEIGHT; y++)	// this code reads old char-based tilemaps
		for (x = 0; x < MAP_WIDTH; x++)
			mpos(map, y, x) = fgetc(fp); */

	// Close the map file, we're done with it
	pack_fclose(fp);

	// Give the user a confirmation message that the map has been loaded
	if (!tilefile[0])
		add_msg(MSG_COMPUTER, "Loaded %s by %s (using internal tileset)", mh.name, mh.author);
	else
	{
		load_terrain(tilefile);
	}

	// Save the map info struct in a global struct
	memcpy(&current_map, &mh, sizeof(current_map));
	strcpy(current_map_filename, filename);

	// Invalidate the cached terrain bitmap for the mini-map, since we loaded a new map
	redraw_terrain_bmp = 1;

	return 0;
}

// smooth_vista(): Smooths out a height map (like a filter.. or anti-aliasing)
void smooth_vista(BITMAP *bmp)
{
	BITMAP *tmp;
	int x, y, avg;

	tmp = create_bitmap(bmp->w, bmp->h);

	for (y = 0; y < bmp->h; y++)
	{
		for (x = 0; x < bmp->w; x++)
		{
            avg = 	getpixel(bmp, x - 1, y - 1) +		// NW
					getpixel(bmp, x + 0, y - 1) +		// N
					getpixel(bmp, x + 1, y - 1) +		// NE
					getpixel(bmp, x - 1, y + 0) +		// W
					getpixel(bmp, x + 0, y + 0) +		// Center
					getpixel(bmp, x + 1, y + 0) +		// E
					getpixel(bmp, x - 1, y + 1) +		// SW
					getpixel(bmp, x + 0, y + 1) +		// S
					getpixel(bmp, x + 1, y + 1); 		// SE
			avg /= 9;
			if (avg < 0)
				avg = 0;
			putpixel(tmp, x, y, avg);
		}
	}

	blit(tmp, bmp, 0, 0, 0, 0, bmp->w, bmp->h);

	destroy_bitmap(tmp);
}

// load_vista(): Loads a height map from an image file (probably generated by Vistapro)
int load_vista(char *filename, int sealevel, int smoothing)
{
	BITMAP *bmp;
	RGB pal[256];
	int x, y, i, j, min, max, mid;

	// Load the bitmap
	bmp = load_bitmap(filename, pal);

	if (!bmp)
	{
		add_msg(MSG_COMPUTER, "Unable to load %s!", filename);
		return 1;
	}

	// Apply smoothing
	for (i = 0; i < smoothing; i++)
		smooth_vista(bmp);

	// Redefine map dimensions
	MAP_WIDTH = bmp->w;
	MAP_HEIGHT = bmp->h;
	redraw_terrain_bmp = 1;

	// Free memory...
	if (terrain_bmp)
		free(terrain_bmp);
	if (mapobj_bmp)
		free(mapobj_bmp);
	if (vismap_bmp)
		free(vismap_bmp);
	if (map)
		free(map);
	if (frill)
		free(frill);
	if (terrain)
		free(terrain);
	if (vis)
		free(vis);
	if (unit_at)
		free(unit_at);
    // Reallocate memory...
    terrain_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
	mapobj_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
	vismap_bmp = create_bitmap(MAP_WIDTH, MAP_HEIGHT);
	map = (short *) calloc(1, MAP_WIDTH * MAP_HEIGHT * 2);
	frill = (short *) calloc(1, MAP_WIDTH * MAP_HEIGHT * 2);
	terrain = (signed char *) calloc(1, MAP_WIDTH * MAP_HEIGHT);
	vis = (char *) calloc(1, MAP_WIDTH * MAP_HEIGHT);
	unit_at = (thing_t **) calloc(1, MAP_WIDTH * MAP_HEIGHT * sizeof(thing_t *));

	// Clear the newly allocated bitmaps
	clear(terrain_bmp);
	clear(mapobj_bmp);
	clear(vismap_bmp);

/*	if (sealevel < 0)
	{
		// Find min and max points in the bitmap
		min = 256;
		max = 0;
		for (y = 0; y < bmp->h; y++)
			for (x = 0; x < bmp->w; x++)
			{
				i = getpixel(bmp, x, y);
				if ((i > max) && (i != 255))
					max = i;
				if (i < min)
					min = i;
			}
		mid = (min + max) / 2;
		add_msg(MSG_COMPUTER, "Loading %s (min = %d, max = %d)", filename, min, max);
	}
	else */
		mid = sealevel;


	if (sealevel >= 0)
	{
		// Construct a map from the bitmap data
		for (y = 0; y < MAP_HEIGHT; y++)
			for (x = 0; x < MAP_WIDTH; x++)
			{
				i = getpixel(bmp, x, y);
				if (i > mid)    // If it's land...
				{
					mpos(map, y, x) = 26;
					mpos(terrain, y, x) = TERRAIN_LAND;
				}
				else			// otherwise it's water
				{
					mpos(map, y, x) = (rand() % 2) ? 0 : 1;
					mpos(terrain, y, x) = TERRAIN_WATER;
				}
			}
	}
	else	// construct a map using the negatized sealevel as the sea index!
	{
		mid = sealevel * -1;
		// Construct a map from the bitmap data
		for (y = 0; y < MAP_HEIGHT; y++)
			for (x = 0; x < MAP_WIDTH; x++)
			{
				i = getpixel(bmp, x, y);
				if (i != mid)    // If it's land...
				{
					mpos(map, y, x) = 26;
					mpos(terrain, y, x) = TERRAIN_LAND;
				}
				else			// otherwise it's water
				{
					mpos(map, y, x) = (rand() % 2) ? 0 : 1;
					mpos(terrain, y, x) = TERRAIN_WATER;
				}
			}
	}

	// Draw the walls around the map
    for (y = 0; y < MAP_HEIGHT; y++)
    {
		mpos(terrain, y, 0) = TERRAIN_BLOCKED;
        mpos(terrain, y, (MAP_WIDTH - 1)) = TERRAIN_BLOCKED;
    }
    for (x = 0; x < MAP_WIDTH; x++)
    {
		mpos(terrain, 0, x) = TERRAIN_BLOCKED;
        mpos(terrain, (MAP_WIDTH - 1), x) = TERRAIN_BLOCKED;
    }

	fix_coastline();

	return 0;
}

// Tile number definitions
#define TN_WATER	0
#define TN_LAND     26

#define WT(x, y)   ((mpos(terrain, y, x)) == TERRAIN_WATER)		// water tile
#define CT(x, y)   ((mpos(terrain, y, x)) == TERRAIN_COAST)		// coast tile
#define LT(x, y)   ((mpos(terrain, y, x) != TERRAIN_WATER) && (mpos(terrain, y, x) != TERRAIN_COAST))  // land tile
#define WC(x, y)   ((mpos(terrain, y, x) == TERRAIN_WATER) || (mpos(terrain, y, x) == TERRAIN_COAST))  // water or coast tile
#define LC(x, y)   !WT(x, y)	// land or coast tile

#define LC_NW	14
#define LC_NE	19
#define LC_SW	12
#define LC_SE	16

#define LC_N	25
#define LC_S	24
#define LC_W	23
#define LC_E    22

#define WC_NW   17
#define WC_NE	13
#define WC_SW	18
#define WC_SE	15

void fix_coastline(void)
{
	int x, y, i, j, done;
	int pass;

	for (pass = 0; pass < 5; pass++)
	{
		done = 1;

		for (y = 1; y < MAP_HEIGHT - 1; y++)
			for (x = 1; x < MAP_WIDTH - 1; x++)
			{
				// *** Check for totally isolated bits of land or water
				j = 0;
				if (WT(x - 1, y - 1)) j++;
				if (WT(x + 0, y - 1)) j++;
				if (WT(x + 1, y - 1)) j++;
				if (WT(x - 1, y + 0)) j++;
				if (WT(x + 1, y + 0)) j++;
				if (WT(x - 1, y + 1)) j++;
				if (WT(x + 0, y + 1)) j++;
				if (WT(x + 1, y + 1)) j++;
				if ((j == 8) && (!WT(x, y)))	// island surrounded by water
				{
					mpos(terrain, y, x) = TERRAIN_WATER;
					mpos(map, y, x) = TN_WATER;
					done = 0;
				}
				else
				if ((!j) && (WT(x, y)))			// water surrounded by land
				{
					mpos(terrain, y, x) = TERRAIN_LAND;
					mpos(map, y, x) = TN_LAND;
					done = 0;
				}
				else	// NW land corner
				if (LT(x, y) && /*WT(x - 1, y - 1) &&*/ WT(x - 1, y) && WT(x, y - 1))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = LC_NW;
				}
				else	// NE land corner
				if (LT(x, y) && /*WT(x + 1, y - 1) &&*/ WT(x + 1, y) && WT(x, y - 1))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = LC_NE;
				}
				else	// SE land corner
				if (LT(x, y) && /*WT(x + 1, y + 1) &&*/ WT(x + 1, y) && WT(x, y + 1))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = LC_SE;
				}
				else	// SW land corner
				if (LT(x, y) && /*WT(x - 1, y + 1) &&*/ WT(x - 1, y) && WT(x, y + 1))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = LC_SW;
				}
				else	// N land corner
				if (LT(x, y) /*&& WC(x + 1, y - 1)*/ && WT(x, y - 1) /*&& WC(x - 1, y - 1)*/
                              && LC(x - 1, y) && LC(x + 1, y))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = LC_N;
				}
				else	// S land corner
				if (LT(x, y) /*&& WC(x + 1, y + 1)*/ && WT(x, y + 1) /*&& WC(x - 1, y + 1)*/
                              && LC(x - 1, y) && LC(x + 1, y))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = LC_S;
				}
				else	// E land corner
				if (LT(x, y) /*&& WC(x + 1, y - 1)*/ && WT(x + 1, y) /*&& WC(x + 1, y + 1)*/
                              && LC(x, y - 1) && LC(x, y + 1))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = LC_E;
				}
				else	// W land corner
				if (LT(x, y) /*&& WC(x - 1, y - 1)*/ && WT(x - 1, y) /*&& WC(x - 1, y + 1)*/
                              && LC(x, y - 1) && LC(x, y + 1))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = LC_W;
				}
				else    // NW water corner
				if (LT(x, y) /*&& WC(x + 1, y) && WC(x, y + 1)*/ && WT(x + 1, y + 1)
                    && LC(x, y + 1) && LC(x + 1, y))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = WC_NW;
				}
				else	// NE water corner
				if (LT(x, y) /*&& WC(x - 1, y) && WC(x, y + 1)*/ && WT(x - 1, y + 1)
					&& LC(x - 1, y) && LC(x, y + 1))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = WC_NE;
				}
				else	// SW water corner
				if (LT(x, y) /*&& WC(x + 1, y) && WC(x, y - 1)*/ && WT(x + 1, y - 1)
					&& LC(x, y - 1) && LC(x + 1, y))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = WC_SW;
				}
				else	// SE water corner
				if (LT(x, y) /*&& WC(x - 1, y) && WC(x, y - 1)*/ && WT(x - 1, y - 1)
					&& LC(x - 1, y) && LC(x, y - 1))
				{
					mpos(terrain, y, x) = TERRAIN_COAST;
					mpos(map, y, x) = WC_SE;
				}
			}

	} while (!done);
}

void process_gui(void)
{
	wnd_t *w;
	int x, y, i;

//	if (!map_editor)	// If we're not in the map editor mode,
//		return;			// this function should not be called.
// This function is now called all the time, for the whole game's GUI

	// Set the mouse cursor to be a plain one- other routines will alter it if necessary
	mouse_cursor = (BITMAP *) datafile[CURSOR].dat;

	// Calculate the tile-under-mouse world coordinates
	mtilex = ((mouse_x + bgxoff) / 32) + bgx;
	mtiley = ((mouse_y + bgyoff) / 32) + bgy;

	if ((!selecting_world) && (map_editor))
	{
		x = ((mouse_x + bgxoff) - ((mouse_x + bgxoff) % 32)) - bgxoff;
		y = ((mouse_y + bgyoff) - ((mouse_y + bgyoff) % 32)) - bgyoff;
		rect(HiddenPage, x + 32, y + 32, x + 32 + (memtilew * 32), y + 32 + (memtileh * 32), 96 + (simclock % 16));
	}

	i = 0;
	// Draw the windows
	if (!hide_all_windows)
		for (w = wnd_list; w; w = w->next)
		{
			// Silly little window scattering effect - toggled by W
			if ((scatter_windows) && (!(rand() % 4)))
			{
				w->x += (rand() % 3) - 1;
				w->y += (rand() % 3) - 1;
				w->intx = w->x + BORDER_WIDTH;
				if (w->style != WND_NOTITLE)
					w->inty = w->y + TOP_HEIGHT;
				else
					w->inty = w->y + BORDER_WIDTH;
			}
			// Draw the window frame (borders, title bar)
			draw_window_frame(w);
			// Draw the interior of the window
			if (w && w->draw)
				w->draw(w);
			// Display some debug info
			i++;
			if (debug)
				textprintf(HiddenPage, font, 234, 64 + (12 * i), TEXT_COLOR, "Window %d (%p) (%d,%d)", i, w, w->x, w->y);
		}

	// Process mouse input
	process_mouse_input();

	// Draw the mouse cursor  (quick hack to modify hotspot for *resize* cursor)
	if (mouse_cursor == datafile[CURSOR_RESIZE].dat)
		draw_sprite(HiddenPage, mouse_cursor, mouse_x + 28, mouse_y + 28);
    else
		draw_sprite(HiddenPage, mouse_cursor, mouse_x + 32, mouse_y + 32);
}

void draw_tilesrc(wnd_t *w)
{
	int i, x, y, tx, ty;
	DATAFILE *data;
	int tileoffset, tilenum, tilenumoffset;
	int interiorx, interiory;
	int scrollbarsize, scrollbaroffset;

	// Display some debug info
//	textprintf(HiddenPage, font, 134, LogicalHeight - 66, TEXT_COLOR, "Mouse: %d,%d  World: %d,%d", mouse_x, mouse_y, mtilex, mtiley);

	// Debugging - randomly change the x and y tile offsets
//	tilexoff = rand() % 5;
//	tileyoff = rand() % 5;

	// Save the initial x and y offsets in the tile source window..
	tx = tilexoff;
	ty = tileyoff;

	// Draw scrollbars, if they exist for this window
	if ((w->style & WND_HSCROLLBAR) || (w->style & WND_SCROLLBARS))
	{
		// Calculate width of scrollbar
		scrollbarsize = (int) ( ((float) w->intw / 32.0) / ((float) tileareaw) * (float) (w->intw - (SCROLLBAR_WIDTH * 2)) ) - 1;
		// Calculate horizontal offset (position) of scrollbar
		scrollbaroffset = (int) ( ((float) tilexoff) / ((float) tileareaw) * (float) (w->intw - (SCROLLBAR_WIDTH * 2)) );
		// Calculate starting coordinates for the scrollbar
	    x = 32 + w->intx + SCROLLBAR_WIDTH + scrollbaroffset + 1;
		y = 32 + w->y + w->h - SCROLLBAR_WIDTH - BORDER_WIDTH;
		// Draw the actual scrollbar!
		draw_button(x, y, x + scrollbarsize - 2, y + SCROLLBAR_WIDTH, 0);
	}

	if ((w->style & WND_VSCROLLBAR) || (w->style & WND_SCROLLBARS))
	{
		// Calculate height of scrollbar
		scrollbarsize = (int) ( ((float) w->inth / 32.0) / ((float) tileareah) * (float) (w->inth - (SCROLLBAR_WIDTH * 2)) ) - 1;
		// Calculate vertical offset (position) of scrollbar
		scrollbaroffset = (int) ( ((float) tileyoff) / ((float) tileareah) * (float) (w->inth - (SCROLLBAR_WIDTH * 2)) );
		// Calculate starting coordinates for the scrollbar
	    x = 32 + w->intx + w->intw;
		y = 32 + w->inty + SCROLLBAR_WIDTH + scrollbaroffset + 1;
		// Draw the actual scrollbar!
		draw_button(x, y, x + SCROLLBAR_WIDTH, y + scrollbarsize - 2, 0);
	}

	// Internal or external tilesets?
	if (terraingfx)
	{
		data = terraingfx;
		tilenum = 0;
	}
/*	else
	{
		data = datafile;
		tilenum = TILE01_PCX;
	} */

	// We need to know the actual starting value of tilenum for later
	tilenumoffset = tilenum;

	// Display some debug info
	if (debug)
	{
		textprintf(HiddenPage, font, 134, LogicalHeight - 80, TEXT_COLOR, "Size: %dx%d  int: %dx%d  tot: %dx%d", w->intw / 32, w->inth / 32, w->intw, w->inth, w->w, w->h);
		textprintf(HiddenPage, font, 134, LogicalHeight - 60, TEXT_COLOR, "Tilearea: %dx%d  tileoffsets: %dx%d", tileareaw, tileareah, tilexoff, tileyoff);
	}

	// If tile periodicity is enabled, use it to calculate the offset.  Otherwise,
	// use the source tile window tile width.  Heheh.  "Source tile window tile width."
	if (tileperiodicity)		// terraingfx used to be a requirement here...
		tilenum += (tileperiodicity * tileyoff) + tilexoff;
//	else
//		tilenum += ((w->intw / 32) * tileyoff) + tilexoff;

	// Draw the source tile box
	for (y = 0; y <= w->inth - 32; y += 32)
	{
		if ((y + w->inty > SCREEN_H) || (y + w->inty < -32))
			continue;
		for (x = 0; x <= w->intw - 32; x += 32)
		{
			if ((x + w->intx > SCREEN_W) || (x + w->intx < -32))
				continue;
			// Find the correct index # to use for the tile
			i = tilenum + (x / 32);
			// If this tile # does not exist, draw an empty white rect
			if (i >= NUM_TILES)
			{
				rectfill(HiddenPage, w->intx + x + 32, w->inty + y + 32, w->intx + x + 63, w->inty + y + 63, makecol(255, 255, 255));
				textout_centre(HiddenPage, font, "Oy", w->intx + x + 48, w->inty + y + 48, makecol(0, 0, 0));
				continue;
			}
//			if ((datafile[i].type) && ((datafile[i].type == DAT_XC_SPRITE) || (datafile[i].type == DAT_C_SPRITE)))
			// If the tile is within the tile area, draw it..
			if (i - tilenumoffset < tileareaw * tileareah)
			{
				if (data[i].type == DAT_XC_SPRITE)
					draw_compiled_sprite(HiddenPage, (COMPILED_SPRITE *) data[i].dat, w->intx + x + 32, w->inty + y + 32);
				else
					draw_sprite(HiddenPage, (BITMAP *) data[i].dat, w->intx + x + 32, w->inty + y + 32);
//				else
//					rectfill(HiddenPage, w->intx + x + 32, w->inty + y + 32, w->intx + x + 64, w->inty + y + 64, 0);
			}
		}
		// If tile periodicity is enabled (not 0), use it.. otherwise inc tilenum linearly
		if (tileperiodicity)				// terraingfx used to be a requirement here...
			tilenum += tileperiodicity;
//		else
//			tilenum += (w->intw / 32);
	}
}

void draw_tilewnd(wnd_t *w)
{
	int i, x, y, tx, ty;
	DATAFILE *data;
	int tileoffset, tilenum;
	int interiorx, interiory;

	// Internal or external graphics...
	if (terraingfx)
	{
		data = terraingfx;
		tileoffset = 0;
	}
/*	else
	{
		data = datafile;
		tileoffset = TILE01_PCX;
	} */

	// If tile periodicity is enabled, use it to calculate the offset.  Otherwise,
	// use the source tile window tile width.  Heheh.  "Source tile window tile width."
/*	if ((terraingfx) && (memtileperiodicity))
		tilenum += (memtileperiodicity * memtileyoff) + tilexoff;
	else
		tilenum += ((w->intw / 32) * tileyoff) + tilexoff; */
	tilenum = memtile;

	// Size the window according to the selection
	w->intw = memtilew * 32;
	w->inth = memtileh * 32;
	w->w = (BORDER_WIDTH * 2) + w->intw;			// - 1
	w->h = BORDER_WIDTH + TOP_HEIGHT + w->inth;		// - 1

	// Draw the selection tile box
	for (y = 0; y <= w->inth - 32; y += 32)
	{
		if ((y + w->inty > SCREEN_H) || (y + w->inty < -32))
			continue;
		for (x = 0; x <= w->intw - 32; x += 32)
		{
			if ((x + w->intx > SCREEN_W) || (x + w->intx < -32))
				continue;
			i = tilenum + (x / 32);
//			if (data[((memtiles[(y / 32)][(x / 32)]) + tileoffset)].dat)
//			if ((datafile[i].type) && ((datafile[i].type == DAT_XC_SPRITE) || (datafile[i].type == DAT_C_SPRITE)))
			// If this tile # does not exist, draw an empty white rect
			if (i >= NUM_TILES)
			{
				rectfill(HiddenPage, w->intx + x + 32, w->inty + y + 32, w->intx + x + 63, w->inty + y + 63, makecol(255, 255, 255));
				textout_centre(HiddenPage, font, "Oy", w->intx + x + 48, w->inty + y + 48, makecol(0, 0, 0));
				continue;
			}

			if (data[(i + tileoffset)].type == DAT_XC_SPRITE)
				draw_compiled_sprite(HiddenPage, (COMPILED_SPRITE *) data[(i + tileoffset)].dat, w->intx + x + 32, w->inty + y + 32);
			else
				draw_sprite(HiddenPage, (BITMAP *) data[(i + tileoffset)].dat, w->intx + x + 32, w->inty + y + 32);

//			draw_compiled_sprite(HiddenPage, (COMPILED_SPRITE *) data[((memtiles[(y / 32)][(x / 32)]) + tileoffset)].dat, w->intx + x + 32, w->inty + y + 32);
//			else
//				rectfill(HiddenPage, w->intx + x + 32, w->inty + y + 32, w->intx + x + 64, w->inty + y + 64, 0);
		}
		// If tile periodicity is enabled (not 0), use it.. otherwise inc tilenum linearly
//		if ((terraingfx) && (memtileperiodicity))
			tilenum += memtileperiodicity;
//		else
//			tilenum += (w->intw / 32);
	}

/*	if ((w->intx < 0) || (w->intx > SCREEN_W + 31))
		return;
	if ((w->inty < 0) || (w->inty > SCREEN_H + 31))
		return;
	draw_compiled_sprite(HiddenPage, (COMPILED_SPRITE *) data[(tileoffset + memtile)].dat, w->intx + 32, w->inty + 32); */
}

void draw_palette(wnd_t *w)
{
	int i, x, y;

	i = 0;
	for (y = 0; y < w->inth; y += 4)
		for (x = 0; x < w->intw; x += 4)
			rectfill(HiddenPage, w->intx + x + 32, w->inty + y + 32, w->intx + x + 36, w->inty + y + 36, i++);
}

void draw_orderwnd(wnd_t *w)
{
	char buf[80];
	int bw, bh;		// unit bitmap width and height
	int x, y, i, mw, mh;
	BITMAP *ubmp;
	float pct;

	// Change the window title with a % complete for an order in progress
	if (order_state >= 0)
	{
		i = order_state;
		pct = ((float) (ms_time - order_time) / (float) unit_order_time[i]) * 100.0;
		if (pct < 100.0)
		{
			// Change the window title
			if (w->title)
				free(w->title);
			sprintf(buf, "%s Order: %.0f%%", unit_name[i], pct);
			w->title = (char *) malloc(strlen(buf) + 1);
			strcpy(w->title, buf);
		}
	}

	// Draw the empty white background
	rectfill(HiddenPage, w->intx + 32, w->inty + 32, w->intx + 32 + w->intw, w->inty + 32 + w->inth, makecol(255, 255, 255));

	// Starting coordinates for drawing
	x = w->intx + 32 + 14;
	y = w->inty + 32 + 4;

	// Find the width and height of the widest and tallest units
	mw = mh = 0;
	for (i = 0; i < NUM_UNITS; i++)
	{
		if ((((BITMAP *) datafile[UNIT1_PCX + i].dat)->w) > mw)
			mw = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->w;
		if ((((BITMAP *) datafile[UNIT1_PCX + i].dat)->h) > mh)
			mh = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->h;
	}

	// Draw the unit bitmaps
	for (i = 0; i < NUM_UNITS; i++)
	{
		// Grab the width and height of the unit bitmap
		bw = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->w;
		bh = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->h;
		// Draw the unit in the upper-left corner of the window
		ubmp = (BITMAP *) datafile[UNIT1_PCX + i].dat;
		draw_sprite(HiddenPage, ubmp, x, y + (mh - bh) / 2);
		// Display the unit's price below it
		textprintf_centre(HiddenPage, font, x + ubmp->w / 2, y + mh + 2, makecol(0, 0, 0), "$%d", unit_cost[i]);
		// Increment our horizontal (x) drawing location by the unit width + margin
//		x += bw + 4;
		x += mw + 4;
	}
}

int input_orderwnd(wnd_t *w)
{
	static wnd_t *moving = 0;
	static int xorigin, yorigin;
	static wnd_t *destroy = 0;
	int x, y, bh, bw, i, mw;
	char txtbuf[100];

	if ((mouse_b & 1) || (mouse_b & 2))
	{
		// WINDOW CLOSING BEHAVIOR
		if ((!moving) && (!(w->style & WND_NOTITLE)))
		{
			if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 7, w->x + w->w - BORDER_WIDTH + 2))
			 && (BETWEEN(mouse_y, w->y + BORDER_WIDTH - 1, w->y + BORDER_WIDTH + TITLEBAR_HEIGHT - 4)))
			{
				destroy = w;
				return 1;
			}
			else
				if (destroy == w)
					destroy = 0;
		}

		// WINDOW MOVING BEHAVIOR
		if (!moving)
		{
			// Check for window movement (anywhere in the window, easymove, woo!)
			if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH, w->x + w->w - BORDER_WIDTH))
	         && (BETWEEN(mouse_y, w->y + BORDER_WIDTH, w->y + TOP_HEIGHT)))
//			if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH, w->x + w->w - BORDER_WIDTH))
//	         && (BETWEEN(mouse_y, w->y + BORDER_WIDTH, w->y + w->h - BORDER_WIDTH)))
			{
				moving = w;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of moving a window.
			if (w == moving)	// Make sure they're moving THIS window though!
			{
				w->x += mouse_x - xorigin;
				w->intx += mouse_x - xorigin;
				xorigin = mouse_x;
				w->y += mouse_y - yorigin;
				w->inty += mouse_y - yorigin;
				yorigin = mouse_y;
				return 1;
			}

		// UNIT SELECTION BEHAVIOR
		if ((BETWEEN(mouse_x, w->intx, w->intx + w->intw)) && (BETWEEN(mouse_y, w->inty, w->inty + w->inth)))
		{
			// If we're not already ordering something...
			if (order_state == -1)
			{
				// Check to see which unit was clicked on
				x = w->intx + 14;
				y = w->inty + 4;
				// Find the width of the widest unit
				mw = 0;
				for (i = 0; i < NUM_UNITS; i++)
					if ((((BITMAP *) datafile[UNIT1_PCX + i].dat)->w) > mw)
						mw = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->w;
				// Loop through the units checking for input
				for (i = 0; i < NUM_UNITS; i++)
				{
					// Grab the width and height of the unit bitmap
					bw = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->w;
					bh = ((BITMAP *) datafile[UNIT1_PCX + i].dat)->h;
					// Check the coordinates
					if ((BETWEEN(mouse_x, x, x + bw)) && (BETWEEN(mouse_y, y, y + bh)))
					{
						// See if the player has enough money to buy this unit
						if (player.money >= unit_cost[i])
						{
							// Take their money
							player.money -= unit_cost[i];
							// Set up the order state and start time
	                        order_state = i;
							order_time = ms_time;
							// Change the window title
							if (w->title)
								free(w->title);
							sprintf(txtbuf, "%s Order: 0%%", unit_name[i]);
							w->title = (char *) malloc(strlen(txtbuf) + 1);
							strcpy(w->title, txtbuf);
						}
						else
							add_msg(MSG_COMPUTER, "Insufficient funds to purchase a %s!", unit_name[i]);
						break;
					}
					// Increment our horizontal (x) location by the unit width + margin
//					x += bw + 4;
					x += mw + 4;
				}
				return 1;
			}
		}

		// ADD NEW WINDOW INPUT HANDLING FUNCTIONALITY HERE

		// Handle clicks on the border that would not normally be acknowledged
		if ((BETWEEN(mouse_x, w->x - 2, w->x + w->w + 2) && (BETWEEN(mouse_y, w->y - 2, w->y + w->h + 2))))
			return 1;
	}
	else
	{
		// If they -were- moving the window, stop it cause they're not anymore. :)
		if (moving == w)
			moving = 0;
		// If window is marked for destruction, destroy it
		if (destroy == w)
		{
			destroy_window(w);
			destroy = 0;
			return 1;
		}
	}

	return 0;
}

void draw_unitinfo(wnd_t *w)
{
	char buf[80];
	unit_t *u;
	int bw, bh;		// unit bitmap width and height
	int x, y;
	#define HH 8    // health meter height

	// Draw the empty white background
	rectfill(HiddenPage, w->intx + 32, w->inty + 32, w->intx + 32 + w->intw, w->inty + 32 + w->inth, makecol(255, 255, 255));
	// If the unit is dead, destroy the window and return
	if (!w->data)
	{
		destroy_window(w);
		return;
	}
	else
		u = (unit_t *) w->data;
//		memcpy(&u, w->data, sizeof(unit_t));

	// Draw the unit in the upper-left corner of the window
	draw_sprite(HiddenPage, (BITMAP *) datafile[UNIT1_PCX + u->type].dat, 34 + w->intx, 34 + w->inty);
	// Grab the width and height of the unit bitmap
	bw = ((BITMAP *) datafile[UNIT1_PCX + u->type].dat)->w;
	bh = ((BITMAP *) datafile[UNIT1_PCX + u->type].dat)->h;
	// Draw the actual health meter
	if (u->health >= 0)
		rectfill(HiddenPage, w->intx + 37 + bw, w->inty + 34 + (bh / 2) - (HH / 2), w->intx + 37 + bw + (int) (((float) (u->health * (w->intx + 30 + w->intw - (w->intx + 37 + bw)))) / (float) u->maxhealth), w->inty + (34 + HH) + (bh / 2) - (HH / 2), makecol(127, 0, 0));
	// Draw the black rectangular outline around (over) the health meter
	rect(HiddenPage, w->intx + 37 + bw, w->inty + 34 + (bh / 2) - (HH / 2), w->intx + 30 + w->intw, w->inty + (34 + HH) + (bh / 2) - (HH / 2), 0);
//	textprintf(HiddenPage, font, w->intx + 34, w->inty + 40, TEXT_COLOR, "%d/%d", u.health, u.maxhealth);
	switch (u->orders)
	{
		case ORDER_GUARD: strcpy(buf, "Guarding"); break;
		case ORDER_RANDOM: strcpy(buf, "SEIZURE!"); break;
		case ORDER_GOTO: strcpy(buf, "Moving"); break;
		case ORDER_SEEK_PLAYER: strcpy(buf, "Hunting Player"); break;
		case ORDER_WAIT: strcpy(buf, "Waiting"); break;
		case ORDER_GOTO_CLOSEST_MBASE: strcpy(buf, "Attacking Base"); break;
		default: strcpy(buf, "Unknown order!!!");
	}
	// Display the current state of the unit
	textout_centre(HiddenPage, font, buf, 32 + w->intx + w->intw / 2, w->inty + 36 + bh, 0);

	// Draw some buttons
	x = 32 + w->intx + 2;
	y = w->inty + 36 + bh + text_height(font) + 2;

//	draw_text_button(x, y, "Guard");
//	draw_text_button(32 + w->intx + w->intw - text_length(font, "Guard") - 18, y, "Attack");

	// Some debug info...
	textprintf(HiddenPage, font, x, y, 0, "Dir: %d", u->direction);
//	textprintf(HiddenPage, font, x, y + text_height(font), 0, "Target: %p", u->target);
//	textprintf(HiddenPage, font, x, y + text_height(font), 0, "Tdir: %d", u->targetdir);
	textprintf(HiddenPage, font, x, y + text_height(font), 0, "Speed: %d", fixtoi(u->speed));
}

#define GRAPH_RESOLUTION	100
#define GRAPH_FREQUENCY		500

void draw_memgraph(wnd_t *w)
{
	int i, x, y;
	static int memsample[GRAPH_RESOLUTION];
	static int longtermpeak = 0;
	int min, max;
	static char init = 1;
	static unsigned long lastupdate = 0;
	float scalefactor;
	char txtbuf[80];

	if (init)
	{
		for (i = 0; i < GRAPH_RESOLUTION; i++)
			memsample[i] = 0;
		init = 0;
	}


	if (ms_time - lastupdate > GRAPH_FREQUENCY)
	{
        for (i = 0; i < GRAPH_RESOLUTION - 1; i++)
			memsample[i] = memsample[i + 1];
		memsample[GRAPH_RESOLUTION - 1] = heapused;
		lastupdate = ms_time;
//		add_msg(MSG_COMPUTER, "Heap usage: %d", heapused);
		if (w->title)
			free(w->title);
		sprintf(txtbuf, "Mem: %d", heapused);
		w->title = (char *) malloc(strlen(txtbuf) + 1);
		strcpy(w->title, txtbuf);
	}

	min = 1000000000;
	max = 0;
	for (i = 0; i < GRAPH_RESOLUTION; i++)
	{
		if (memsample[i] < min)
			min = memsample[i];
		if (memsample[i] > max)
			max = memsample[i];
	}

	min = 0;   // Uncomment this to get rid of the floating min graph...

	if (max > longtermpeak)
		longtermpeak = max;
	else
		max = longtermpeak;

	if ((max - min) == 0)	// Protect from division by zero!!!!
		scalefactor = 0;
	else
		scalefactor = (float) w->inth / ((float) max - (float) min);

//	add_msg(MSG_COMPUTER, "min = %d  max = %d  Scalefactor = %f", min, max, scalefactor);

	rectfill(HiddenPage, w->intx + 32, w->inty + 32, w->intx + w->intw + 32, w->inty + w->inth + 32, makecol(255, 255, 255));

	for (i = 0; i < GRAPH_RESOLUTION - 1; i++)
	{
		line(HiddenPage,
			 32 + w->intx + (i * w->intw / (GRAPH_RESOLUTION - 1)),
			 32 + w->inty + w->inth - (int) (((float) (memsample[i] - min)) * scalefactor),
			 32 + w->intx + ((i + 1) * w->intw / (GRAPH_RESOLUTION - 1)),
			 32 + w->inty + w->inth - (int) (((float) (memsample[i + 1] - min)) * scalefactor),
			 0);
	}
}

void draw_projgraph(wnd_t *w)
{
	int i, x, y;
	static int memsample[GRAPH_RESOLUTION];
	static int longtermpeak = 0;
	int min, max;
	static char init = 1;
	static unsigned long lastupdate = 0;
	float scalefactor;
	char txtbuf[80];

	if (init)
	{
		for (i = 0; i < GRAPH_RESOLUTION; i++)
			memsample[i] = 0;
		init = 0;
	}


	if (ms_time - lastupdate > GRAPH_FREQUENCY)
	{
        for (i = 0; i < GRAPH_RESOLUTION - 1; i++)
			memsample[i] = memsample[i + 1];
		memsample[GRAPH_RESOLUTION - 1] = active_projectiles - non;
		lastupdate = ms_time;
//		add_msg(MSG_COMPUTER, "Heap usage: %d", heapused);
		if (w->title)
			free(w->title);
		sprintf(txtbuf, "Proj: %d", active_projectiles - non);
		w->title = (char *) malloc(strlen(txtbuf) + 1);
		strcpy(w->title, txtbuf);
	}

	min = 1000000000;
	max = 0;
	for (i = 0; i < GRAPH_RESOLUTION; i++)
	{
		if (memsample[i] < min)
			min = memsample[i];
		if (memsample[i] > max)
			max = memsample[i];
	}

	min = 0;   // Uncomment this to get rid of the floating min graph...

	if (max > longtermpeak)
		longtermpeak = max;
	else
		max = longtermpeak;

	if ((max - min) == 0)	// Protect from division by zero!!!!
		scalefactor = 0;
	else
		scalefactor = (float) w->inth / ((float) max - (float) min);

//	add_msg(MSG_COMPUTER, "min = %d  max = %d  Scalefactor = %f", min, max, scalefactor);

	rectfill(HiddenPage, w->intx + 32, w->inty + 32, w->intx + w->intw + 32, w->inty + w->inth + 32, makecol(255, 255, 255));

	for (i = 0; i < GRAPH_RESOLUTION - 1; i++)
	{
		line(HiddenPage,
			 32 + w->intx + (i * w->intw / (GRAPH_RESOLUTION - 1)),
			 32 + w->inty + w->inth - (int) (((float) (memsample[i] - min)) * scalefactor),
			 32 + w->intx + ((i + 1) * w->intw / (GRAPH_RESOLUTION - 1)),
			 32 + w->inty + w->inth - (int) (((float) (memsample[i + 1] - min)) * scalefactor),
			 0);
	}
}

void draw_button(int x1, int y1, int x2, int y2, unsigned char style)
{
	static int white, gray, ltgray, dkgray, black, blue;
	static int refresh_colors = 0;
	int x, y;

	// Do some color calculations, only every so often (like every 200 buttons drawn)
	if (!refresh_colors)
	{
		white = makecol(255, 255, 255);
		gray = makecol(191, 191, 191);
		ltgray = makecol(223, 223, 223);
		dkgray = makecol(127, 127, 127);
		black = makecol(0, 0, 0);
		blue = makecol(0, 0, 127);
		refresh_colors = 400;
	}
	else
		refresh_colors--;

	// Draw top border
	hline(HiddenPage, x1, y1, x2 - 1, ltgray);
	hline(HiddenPage, x1 + 1, y1 + 1, x2 - 2, white);

	// Draw bottom border
	hline(HiddenPage, x1, y2, x2, black);
	hline(HiddenPage, x1 + 1, y2 - 1, x2 - 1, dkgray);

	// Draw left border
	vline(HiddenPage, x1, y1 + 1, y2 - 1, ltgray);
	vline(HiddenPage, x1 + 1, y1 + 1, y2 - 2, white);

	// Draw right border
	vline(HiddenPage, x2, y1, y2 - 1, black);
	vline(HiddenPage, x2 - 1, y1 + 1, y2 - 2, dkgray);

	// Fill in the center
	rectfill(HiddenPage, x1 + 2, y1 + 2, x2 - 2, y2 - 2, gray);

	switch (style)
	{
		case BUTTON_DOWN:
		{
			x = (x1 + x2) / 2;
			y = (y1 + y2) / 2 - 1;
			hline(HiddenPage, x - 3, y + 0, x + 3, black);
			hline(HiddenPage, x - 2, y + 1, x + 2, black);
			hline(HiddenPage, x - 1, y + 2, x + 1, black);
			putpixel(HiddenPage, x, y + 3, black);
			break;
		}
		case BUTTON_UP:
		{
			x = (x1 + x2) / 2;
			y = (y1 + y2) / 2 - 2;
			putpixel(HiddenPage, x, y, black);
			hline(HiddenPage, x - 1, y + 1, x + 1, black);
			hline(HiddenPage, x - 2, y + 2, x + 2, black);
			hline(HiddenPage, x - 3, y + 3, x + 3, black);
			break;
		}
		case BUTTON_LEFT:
		{
			x = (x1 + x2) / 2 - 2;
			y = (y1 + y2) / 2;
			putpixel(HiddenPage, x, y, black);
			vline(HiddenPage, x + 1, y - 1, y + 1, black);
			vline(HiddenPage, x + 2, y - 2, y + 2, black);
			vline(HiddenPage, x + 3, y - 3, y + 3, black);
			break;
		}
		case BUTTON_RIGHT:
		{
			x = (x1 + x2) / 2 + 2;
			y = (y1 + y2) / 2;
			putpixel(HiddenPage, x, y, black);
			vline(HiddenPage, x - 1, y - 1, y + 1, black);
			vline(HiddenPage, x - 2, y - 2, y + 2, black);
			vline(HiddenPage, x - 3, y - 3, y + 3, black);
			break;
		}
		case BUTTON_X:
		{
/*			line(HiddenPage, x1 + 2, y1 + 2, x2 - 2, y2 - 2, black);	// center ul-lr
			line(HiddenPage, x1 + 2, y2 - 2, x2 - 2, y1 + 2, black);	// center ll-ur

			line(HiddenPage, x1 + 2, y1 + 3, x2 - 3, y2 - 2, black);
			line(HiddenPage, x1 + 3, y1 + 2, x2 - 2, y2 - 3, black);

			line(HiddenPage, x1 + 2, y2 - 3, x2 - 3, y1 + 2, black);
			line(HiddenPage, x1 + 3, y2 - 2, x2 - 2, y1 + 3, black); */

			line(HiddenPage, x1 + 3, y1 + 3, x2 - 3, y2 - 3, black);	// center ul-lr
			line(HiddenPage, x1 + 3, y2 - 3, x2 - 3, y1 + 3, black);	// center ll-ur

			break;
		}
	}

}

#define TEXT_MARGIN		4

void draw_text_button(int x1, int y1, char *str)
{
	static int white, gray, ltgray, dkgray, black, blue;
	static int refresh_colors = 0;
	int x, y;
	int x2, y2;

	// Do some color calculations, only every so often (like every 200 buttons drawn)
	if (!refresh_colors)
	{
		white = makecol(255, 255, 255);
		gray = makecol(191, 191, 191);
		ltgray = makecol(223, 223, 223);
		dkgray = makecol(127, 127, 127);
		black = makecol(0, 0, 0);
		blue = makecol(0, 0, 127);
		refresh_colors = 400;
	}
	else
		refresh_colors--;

	// Calculate the dimensions of the button based on the text dimensions
	x2 = x1 + (TEXT_MARGIN * 2) + text_length(font, str);
	y2 = y1 + (TEXT_MARGIN * 2) + text_height(font);

	// Draw top border
	hline(HiddenPage, x1, y1, x2 - 1, ltgray);
	hline(HiddenPage, x1 + 1, y1 + 1, x2 - 2, white);

	// Draw bottom border
	hline(HiddenPage, x1, y2, x2, black);
	hline(HiddenPage, x1 + 1, y2 - 1, x2 - 1, dkgray);

	// Draw left border
	vline(HiddenPage, x1, y1 + 1, y2 - 1, ltgray);
	vline(HiddenPage, x1 + 1, y1 + 1, y2 - 2, white);

	// Draw right border
	vline(HiddenPage, x2, y1, y2 - 1, black);
	vline(HiddenPage, x2 - 1, y1 + 1, y2 - 2, dkgray);

	// Fill in the center
	rectfill(HiddenPage, x1 + 2, y1 + 2, x2 - 2, y2 - 2, gray);

	// Draw the text
    textout(HiddenPage, font, str, x1 + TEXT_MARGIN, y1 + TEXT_MARGIN, black);
}

void draw_window_frame(wnd_t *wnd)
{
	int tw, th;			// total window width, total window height
	static int white, gray, ltgray, dkgray, black, blue;
	static int refresh_colors = 0;
	int x, y, w, h;
	int tbh = TITLEBAR_HEIGHT;	// Title bar height

	x = wnd->x;
	y = wnd->y;
//	w = wnd->intw;
//	h = wnd->inth;
	w = wnd->w;
	h = wnd->h;

	// Display some debug info
//	textprintf(HiddenPage, font, 134, LogicalHeight - 100, TEXT_COLOR, "x = %d, y = %d, w = %d, h = %d, title = %s", x, y, w, h, title);

	// Do some color calculations, only every so often (like every 100 windows drawn)
	if (!refresh_colors)
	{
		white = makecol(255, 255, 255);
		gray = makecol(191, 191, 191);
		ltgray = makecol(223, 223, 223);
		dkgray = makecol(127, 127, 127);
		black = makecol(0, 0, 0);
		blue = makecol(0, 0, 127);
		refresh_colors = 100;
	}
	else
		refresh_colors--;

	// Calculate the total window width and height (border + interior)
//	tw = (BORDER_WIDTH * 2) + w - 1;
//	th = BORDER_WIDTH + TOP_HEIGHT + h - 1;

	// Um, no.
	tw = w;
	th = h;

	// Adjust the origin because of the virtual screen alley of 32 pixels...
	x += 32;
	y += 32;

	// Draw the window title bar (top)
	if (!(wnd->style & WND_NOTITLE))
	{
		hline(HiddenPage, x, y, x + tw - 1, ltgray);
		hline(HiddenPage, x + 1, y + 1, x + tw - 2, white);
		hline(HiddenPage, x + 2, y + 2, x + tw - 2, gray);
		hline(HiddenPage, x + 2, y + 3, x + tw - 2, gray);
		rectfill(HiddenPage, x + 4, y + 4, x + tw - 4, y + 4 + tbh, wnd->next ? dkgray : blue);
		hline(HiddenPage, x + 4, y + tbh + 4, x + tw - 2, gray);
		hline(HiddenPage, x + 4, y + tbh + 5, x + tw - 3, dkgray);
		hline(HiddenPage, x + 5, y + tbh + 6, x + tw - 4, black);
		// Draw the window title text on top of the title bar
	    textout_centre(HiddenPage, font, wnd->title, x + ((tw - 12) / 2) + 2, y + (tbh / 2), wnd->next ? ltgray : white);
	}
	else
	{
		tbh = 0;	// If no title bar, set the title bar height to 0
		// Draw the window border (top)
		hline(HiddenPage, x, y, x + tw - 1, ltgray);
		hline(HiddenPage, x + 1, y + 1, x + tw - 2, white);
		hline(HiddenPage, x + 2, y + 2, x + tw - 2, gray);
		hline(HiddenPage, x + 2, y + 3, x + tw - 2, gray);
		hline(HiddenPage, x + 4, y + 4, x + tw - 2, gray);
		hline(HiddenPage, x + 4, y + 5, x + tw - 3, dkgray);
	}
	// Draw the window border (left)
	vline(HiddenPage, x, y + 1, y + th - 1, ltgray);
	vline(HiddenPage, x + 1, y + 2, y + th - 2, white);
	vline(HiddenPage, x + 2, y + 4, y + th - 2, gray);
	vline(HiddenPage, x + 3, y + 4, y + th - 2, gray);
	vline(HiddenPage, x + 4, y + 4 + tbh + 2, y + th - 5, dkgray);
	vline(HiddenPage, x + 5, y + 4 + tbh + 3, y + th - 6, black);

	// Draw the window border (right)
	vline(HiddenPage, x + tw, y, y + th, black);
	vline(HiddenPage, x + tw - 1, y + 1, y + th - 1, dkgray);
	vline(HiddenPage, x + tw - 2, y + 4, y + th - 2, gray);
	vline(HiddenPage, x + tw - 3, y + 4, y + th - 2, gray);
	vline(HiddenPage, x + tw - 4, y + 4 + tbh + 2, y + th - 5, white);
	vline(HiddenPage, x + tw - 5, y + 4 + tbh + 3, y + th - 6, ltgray);

	// Draw the window border (bottom)
	hline(HiddenPage, x, y + th, x + tw, black);
	hline(HiddenPage, x + 1, y + th - 1, x + tw - 1, dkgray);
	hline(HiddenPage, x + 2, y + th - 2, x + tw - 2, gray);
	hline(HiddenPage, x + 2, y + th - 3, x + tw - 2, gray);
	hline(HiddenPage, x + 4, y + th - 4, x + tw - 4, white);
	hline(HiddenPage, x + 5, y + th - 5, x + tw - 5, ltgray);

	// If horizontal scrollbar, draw its buttons..
	if ((wnd->style & WND_HSCROLLBAR) || (wnd->style & WND_SCROLLBARS))
	{
		// Draw the left button
	    x = 32 + wnd->intx;
		y = 32 + wnd->y + wnd->h - SCROLLBAR_WIDTH - BORDER_WIDTH;              // <----
		draw_button(x, y, x + SCROLLBAR_WIDTH, y + SCROLLBAR_WIDTH, BUTTON_LEFT);
		// Draw the right button
		x = 32 + wnd->intx + wnd->intw - SCROLLBAR_WIDTH - 1;
//		y = 32 + wnd->y + wnd->h - SCROLLBAR_WIDTH - BORDER_WIDTH; // redundant (look up)
		draw_button(x, y, x + SCROLLBAR_WIDTH, y + SCROLLBAR_WIDTH, BUTTON_RIGHT);

		// Draw the empty space behind the actual scrollbar
		rectfill(HiddenPage, 32 + wnd->intx + SCROLLBAR_WIDTH + 1, y, 32 + wnd->intx + wnd->intw - SCROLLBAR_WIDTH - 2, y + SCROLLBAR_WIDTH, ltgray);
	}

	// If vertical scrollbar, draw its buttons..
	if ((wnd->style & WND_VSCROLLBAR) || (wnd->style & WND_SCROLLBARS))
	{
		// Draw the top button
	    x = 32 + wnd->intx + wnd->intw;
		y = 32 + wnd->inty;
		draw_button(x, y, x + SCROLLBAR_WIDTH, y + SCROLLBAR_WIDTH, BUTTON_UP);
		// Draw the bottom button
//	    x = 32 + wnd->intx + wnd->intw;  // redundant (see a few lines up)
		y = 32 + wnd->inty + wnd->inth - SCROLLBAR_WIDTH - 1;
		draw_button(x, y, x + SCROLLBAR_WIDTH, y + SCROLLBAR_WIDTH, BUTTON_DOWN);

		// Draw the empty space behind the actual scrollbar
		rectfill(HiddenPage, 32 + wnd->intx + wnd->intw, 32 + wnd->inty + SCROLLBAR_WIDTH + 1, 32 + wnd->intx + wnd->intw + SCROLLBAR_WIDTH, 32 + wnd->inty + wnd->inth - SCROLLBAR_WIDTH - 2, ltgray);
	}

	// If both scrollbars, fill in that lower-left corner
    if (wnd->style & WND_SCROLLBARS)
		rectfill(HiddenPage, 32 + wnd->intx + wnd->intw, 32 + wnd->inty + wnd->inth, 32 + wnd->intx + wnd->intw + SCROLLBAR_WIDTH, 32 + wnd->inty + wnd->inth + SCROLLBAR_WIDTH, ltgray);

	// Now we need an X button to terminate windows
	if (!(wnd->style & WND_NOTITLE))
		draw_button(32 + wnd->x + wnd->w - BORDER_WIDTH - 7, 32 + wnd->y + BORDER_WIDTH - 1, 32 + wnd->x + wnd->w - BORDER_WIDTH + 2, 32 + wnd->y + BORDER_WIDTH + TITLEBAR_HEIGHT - 4, BUTTON_X);
}

// Searches for a unit at the specified coordinates and creates unit info windows
// for any units found there.  Returns the # of units found.
char findunit(int x, int y)
{
	char j = 0;
	char buf[80];
	unit_t *u;

	for (u = units; u; u = u->next)
		if ((u->x == x) && (u->y == y))
		{
			sprintf(buf, "Unit #%p", u);
			create_wnd(32 + rand() % (SCREEN_W - 200), 32 + rand() % (SCREEN_H - 200), 120, 50, WND_PLAIN, buf, draw_unitinfo, input_easymove, u);
			j++;
		}
	return j;
}

void process_mouse_input(void)
{
	int res = 0, x, y, i, j, numwindows, tileoffset;
	wnd_t *w;
	wnd_t *all[256];
	static int xorigin, yorigin, mtilexorigin, mtileyorigin;
	static int leftlock = 0, rightlock = 0;		// for mouse button locks
	char buf[80];

	// Draw a nice little rect around the current tile-under-mouse, or the selection
	if ((selecting_world) && (map_editor))
	{
		x = ((xorigin + bgxoff) - ((xorigin + bgxoff) % 32)) - bgxoff;
		y = ((yorigin + bgyoff) - ((yorigin + bgyoff) % 32)) - bgyoff;
//		rect(HiddenPage, xorigin + 32, yorigin + 32, xorigin + 32 + (memtilew * 32), yorigin + 32 + (memtileh * 32), 96 + (simclock % 16));
		rect(HiddenPage, x + 32, y + 32, x + 32 + (memtilew * 32), y + 32 + (memtileh * 32), 96 + (simclock % 16));
	}

	// Go through the windows, putting pointers to them in an array (so we can then
	// call the input functions in backwards order, against the list)
	if ((!selecting_world) && (!hide_all_windows))
	{
		numwindows = 0;
		for (w = wnd_list; w; w = w->next)
			if (w->input)
				all[numwindows++] = w;
	
		res = 0;
		for (i = numwindows - 1; i >= 0; i--)
			if (all[i]->input(all[i]))
			{
				// Ok, a window processed some input, put it on top (give it "pseudo-focus")
				bring_to_front(all[i]);
				res = 1;
				break;
			}
	}

	// If none of the window input handlers processed any mouse action...
	// Assume that the user is clicking on the world itself.
	if (!res)
	{
		if (map_editor)
		{
			// Right mouse button grabs a tile from the map and stores it in memory
			if (mouse_b & 2)
			{
				if (!selecting_world)
				{
					selecting_world = 1;
					xorigin = mouse_x;
					yorigin = mouse_y;
					mtilexorigin = mtilex;
					mtileyorigin = mtiley;
	//				memtile = mpos(map, mtiley, mtilex);
					memtiles[0][0] = mpos(map, mtiley, mtilex);
					memterrain[0][0] = mpos(terrain, mtiley, mtilex);
					memtilew = memtileh = 1;
				}
				else
				{
					mtilex = mtilexorigin = ((xorigin + bgxoff) / 32) + bgx;
					mtiley = mtileyorigin = ((yorigin + bgyoff) / 32) + bgy;
	
					if (mouse_x < xorigin)
					{
						xorigin = mouse_x;
						mtilexorigin = mtilex;
					}
					if (mouse_y < yorigin)
					{
						yorigin = mouse_y;
						mtileyorigin = mtiley;
					}
					memtilew = (mouse_x - xorigin) / 32 + 1;
					memtileh = (mouse_y - yorigin) / 32 + 1;
	                for (y = 0; y < memtileh; y++)
						for (x = 0; x < memtilew; x++)
						{
							memtiles[y][x] = mpos(map, (mtileyorigin + y), (mtilexorigin + x));
							memterrain[y][x] = mpos(terrain, (mtileyorigin + y), (mtilexorigin + x));
						}
				}
			}
			else
				if (selecting_world)
				{
					selecting_world = 0;
				}
	
			// Left mouse button puts the tile in memory onto the map
			if (mouse_b & 1)
				paste_memtile(mtilex, mtiley);
		}
		else	// clicking on the world, but not in map editor mode
		{
			if (mouse_b & 1)	// left click - could be selecting a unit
			{
				if (!leftlock)
				{
					leftlock = 1;

					// Search for a unit where the player clicked
					if (!findunit(mtilex, mtiley))
						if (!findunit(mtilex - 1, mtiley - 1))
							if (!findunit(mtilex, mtiley - 1))
								if (!findunit(mtilex + 1, mtiley - 1))
									if (!findunit(mtilex - 1, mtiley))
										if (!findunit(mtilex + 1, mtiley))
											if (!findunit(mtilex - 1, mtiley + 1))
												if (!findunit(mtilex, mtiley + 1))
													findunit(mtilex + 1, mtiley + 1);
				}
			}
			else
			if (mouse_b & 2)	// right click - (unused at the moment)
			{
				if (!rightlock)
				{
					rightlock = 1;
				}
			}
			else
			{
				if (leftlock)	// left mouse button has been released
				{
					leftlock = 0;
				}
				else
				if (rightlock)	// right mouse button has been released
				{
					rightlock = 0;
				}
			}
		}
	}
}

int input_tilesrc(wnd_t *w)
{
	static wnd_t *moving = 0;
	static wnd_t *resizing = 0;
	static wnd_t *selecting = 0;
	static int xorigin, yorigin;
	static int lmemtilew, lmemtileh;	// local memtilew and memtileh
	static long scrolltime = 0;			// last time when the view was scrolled
	static wnd_t *destroy = 0;
	int x, y, i, j;

	// Change cursor to a resize cursor when it's over the lower-right corner of a window
	if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 4, w->x + w->w + BORDER_WIDTH - 4))
     && (BETWEEN(mouse_y, w->y + w->h - BORDER_WIDTH - 4, w->y + w->h + BORDER_WIDTH - 4)))
		mouse_cursor = (BITMAP *) datafile[CURSOR_RESIZE].dat;

	if ((mouse_b & 1) || (mouse_b & 2))
	{
		// WINDOW CLOSING BEHAVIOR
		if ((!resizing) && (!moving) && (!selecting))
		{
			if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 7, w->x + w->w - BORDER_WIDTH + 2))
			 && (BETWEEN(mouse_y, w->y + BORDER_WIDTH - 1, w->y + BORDER_WIDTH + TITLEBAR_HEIGHT - 4)))
			{
				destroy = w;
				return 1;
			}
			else
				if (destroy == w)
					destroy = 0;
		}

		// WINDOW MOVING BEHAVIOR
		if ((!moving) && (!resizing) && (!selecting))
		{
			// Check for window movement
			if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH, w->x + w->w - BORDER_WIDTH))
	         && (BETWEEN(mouse_y, w->y + BORDER_WIDTH, w->y + TOP_HEIGHT)))
			{
				moving = w;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of moving a window.
			if (w == moving)	// Make sure they're moving THIS window though!
			{
				w->x += mouse_x - xorigin;
				w->y += mouse_y - yorigin;
				w->intx = w->x + BORDER_WIDTH;
				w->inty = w->y + TOP_HEIGHT;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}

		// WINDOW RESIZING BEHAVIOR
		if ((!resizing) && (!moving) && (!selecting))
		{
			// Check for window resizing
			if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 4, w->x + w->w + BORDER_WIDTH - 4))
			 && (BETWEEN(mouse_y, w->y + w->h - BORDER_WIDTH - 4, w->y + w->h + BORDER_WIDTH - 4)))
			{
				resizing = w;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of resizing a window.
			if (w == resizing)	// Make sure they're resizing THIS window though!
			{
				if (abs(mouse_x - xorigin) >= 32)
				{
					// If they moved the mouse left by at least 32 pixels, size down
					if (mouse_x - xorigin < 0)
					{
						w->intw -= 32;
						position_mouse(xorigin - 32, mouse_y);
					}
					else	// Otherwise they moved right
					{
						w->intw += 32;
						position_mouse(xorigin + 32, mouse_y);
					}

					// Don't allow the window to get smaller than 2x2 tiles
					if (w->intw < 64)
					{
						position_mouse(w->intx + w->intw, mouse_y);
						w->intw = 64;
					}

					// If the window is being sized down, keep the scrollbar in bounds
					if (mouse_x - xorigin < 0)
					{
						if (tilexoff >= (tileareaw - (w->intw / 32)))
							tilexoff = tileareaw - (w->intw / 32) - 1;
						if (tileyoff >= (tileareah - (w->inth / 32)))
							tileyoff = tileareah - (w->inth / 32) - 1;
					}

					if (tileperiodicity)	// terraingfx used to be a requirement here...
						if (w->intw > tileperiodicity * 32)
						{
							w->intw = tileperiodicity * 32;
							position_mouse(w->intx + w->intw, mouse_y);
						}
					w->w = (BORDER_WIDTH * 2) + w->intw;	// - 1
	
					// If this window has a horizontal scrollbar, add onto the total window width..
					if ((w->style & WND_SCROLLBARS) || (w->style & WND_HSCROLLBAR))
						w->w += SCROLLBAR_WIDTH;
				
					xorigin = mouse_x;
				}
				if (abs(mouse_y - yorigin) >= 32)
				{
					// If they moved the mouse up by at least 32 pixels, size down
					if (mouse_y - yorigin < 0)
					{
						w->inth -= 32;
						position_mouse(mouse_x, yorigin - 32);
					}
					else	// Otherwise they moved down
					{
						w->inth += 32;
						position_mouse(mouse_x, yorigin + 32);
					}

					// Don't allow the window to get smaller than 2x2 tiles
					if (w->inth < 64)
					{
						position_mouse(mouse_x, w->inty + w->inth);
						w->inth = 64;
					}
					w->h = BORDER_WIDTH + TOP_HEIGHT + w->inth - 1;
					// If this window has a vertical scrollbar, add onto the total window height..
					if ((w->style & WND_SCROLLBARS) || (w->style & WND_VSCROLLBAR))
						w->h += SCROLLBAR_WIDTH;
					yorigin = mouse_y;
				}

				// Decide whether or not scrollbars are necessary

				// Recalculate the window's total width and height
				w->w = (BORDER_WIDTH * 2) + w->intw;
				w->h = BORDER_WIDTH + TOP_HEIGHT + w->inth;

                // Assume no scrollbars as the default
				w->style = WND_PLAIN;

				// Check for horizontal scrollbar
				if (w->intw / 32 < tileareaw)
				{
					if (w->inth / 32 < tileareah)
					{
						w->style = WND_SCROLLBARS;
						w->w += SCROLLBAR_WIDTH;
						w->h += SCROLLBAR_WIDTH;
					}
					else
					{
						w->style = WND_HSCROLLBAR;
						w->h += SCROLLBAR_WIDTH;	// make vertical room for the hbar
					}
				}
				else	// Check for vertical scrollbar
					if (w->inth / 32 < tileareah)
					{
						w->style = WND_VSCROLLBAR;
						w->w += SCROLLBAR_WIDTH;	// make horizontal room for the vbar
					}

				return 1;
			}

		// Check for the left scroll button
		if ((w->style & WND_HSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx, w->intx + SCROLLBAR_WIDTH) &&
			    BETWEEN(mouse_y, w->y + w->h - SCROLLBAR_WIDTH - BORDER_WIDTH, w->y + w->h - BORDER_WIDTH))
				if (tilexoff > 0)
					if (ms_time - scrolltime > 100)
					{
						scrolltime = ms_time;
						tilexoff--;
					}

		// Check for the right scroll button
		if ((w->style & WND_HSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + w->intw - SCROLLBAR_WIDTH - 1, w->intx + w->intw - 1) &&
			    BETWEEN(mouse_y, w->y + w->h - SCROLLBAR_WIDTH - BORDER_WIDTH, w->y + w->h - BORDER_WIDTH))
				if (tilexoff < (tileareaw - (w->intw / 32)))
					if (ms_time - scrolltime > 100)
					{
						scrolltime = ms_time;
						tilexoff++;
					}

		// Check for the top scroll button
		if ((w->style & WND_VSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + w->intw, w->intx + w->intw + SCROLLBAR_WIDTH) &&
			    BETWEEN(mouse_y, w->inty, w->inty + SCROLLBAR_WIDTH))
				if (tileyoff > 0)
					if (ms_time - scrolltime > 100)
					{
						scrolltime = ms_time;
						tileyoff--;
					}

		// Check for the bottom scroll button
		if ((w->style & WND_VSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + w->intw, w->intx + w->intw + SCROLLBAR_WIDTH) &&
			    BETWEEN(mouse_y, w->inty + w->inth - SCROLLBAR_WIDTH - 1, w->inty + w->inth - 1))
				if (tileyoff < (tileareah - (w->inth / 32)))
					if (ms_time - scrolltime > 100)
					{
						scrolltime = ms_time;
						tileyoff++;
					}

		// Check for dragging on the horizontal scrollbar
		if ((w->style & WND_HSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + SCROLLBAR_WIDTH, w->intx + w->intw - SCROLLBAR_WIDTH - 1) &&
			    BETWEEN(mouse_y, w->y + w->h - SCROLLBAR_WIDTH - BORDER_WIDTH, w->y + w->h - BORDER_WIDTH))
			{
				// Calculate width of total possible scrollbar area
				i = (w->intx + w->intw - SCROLLBAR_WIDTH - 1) - (w->intx + SCROLLBAR_WIDTH) - 1;
				// Calculate where in this area the mouse is..
                j = mouse_x - (w->intx + SCROLLBAR_WIDTH);

				tilexoff = tileareaw * j / i - (w->intw / 64);

				if (tilexoff < 0)
					tilexoff = 0;
				if (tilexoff > (tileareaw - (w->intw / 32)))
					tilexoff = (tileareaw - (w->intw / 32));
			}

		// Check for dragging on the vertical scrollbar
		if ((w->style & WND_VSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + w->intw, w->intx + w->intw + SCROLLBAR_WIDTH) &&
			    BETWEEN(mouse_y, w->inty + SCROLLBAR_WIDTH, w->inty + w->inth - SCROLLBAR_WIDTH - 1))
			{
				// Calculate width of total possible scrollbar area
				i = (w->inty + w->inth - SCROLLBAR_WIDTH - 1) - (w->inty + SCROLLBAR_WIDTH) - 1;
				// Calculate where in this area the mouse is..
                j = mouse_y - (w->inty + SCROLLBAR_WIDTH);

				tileyoff = tileareah * j / i - (w->inth / 64);

				if (tileyoff < 0)
					tileyoff = 0;
				if (tileyoff > (tileareah - (w->inth / 32)))
					tileyoff = (tileareah - (w->inth / 32));
			}

		// ADD NEW WINDOW INPUT HANDLING FUNCTIONALITY HERE

		// CLICKING INSIDE WINDOW BEHAVIOR - grab a tile

		if ((!selecting) && (!resizing) && (!moving))
		{
			// Check for selecting tiles inside window
			if ((BETWEEN(mouse_x, w->intx, w->intx + w->intw))
			 && (BETWEEN(mouse_y, w->inty, w->inty + w->inth)))
			{
				selecting = w;
				x = mouse_x - (w->x + BORDER_WIDTH);
				y = mouse_y - (w->y + TOP_HEIGHT);
				i = w->intx + ((x / 32) * 32) + 32;
				j = w->inty + ((y / 32) * 32) + 32;
				rect(HiddenPage, i, j, i + 31, j + 31, makecol(255, 255, 255));
				textprintf(HiddenPage, font, 134, LogicalHeight - 66, TEXT_COLOR, "IntMouse: %d,%d  Tiles: %d,%d", x, y, x / 32, y / 32);
				// Record the x and y tile coords where that initial click was..
				xorigin = x / 32;
				yorigin = y / 32;
				// Set the memtile width and height to 1
				lmemtilew = lmemtileh = 1;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of selecting some tiles.
			if (w == selecting)	// Make sure they're doing that in this window though.
			{
				x = mouse_x - (w->x + BORDER_WIDTH);
				y = mouse_y - (w->y + TOP_HEIGHT);
				lmemtilew = (x / 32) - xorigin;
				lmemtileh = (y / 32) - yorigin;
				if (lmemtilew < 0)
				{
					xorigin += lmemtilew;
					lmemtilew = 1;
				}
				if (lmemtileh < 0)
				{
					yorigin += lmemtileh;
					lmemtileh = 1;
				}
				i = w->intx + (xorigin * 32) + 32;
				j = w->inty + (yorigin * 32) + 32;
				rect(HiddenPage, i, j, i + (lmemtilew * 32) + 31, j + (lmemtileh * 32) + 31, makecol(255, 255, 255));
				textprintf(HiddenPage, font, 134, LogicalHeight - 100, TEXT_COLOR, "lmemtilew = %d, lmemtileh = %d, xorigin = %d, yorigin = %d", lmemtilew, lmemtileh, xorigin, yorigin);
				return 1;
			}
		// Handle clicks on the border that would not normally be acknowledged
		if ((BETWEEN(mouse_x, w->x - 2, w->x + w->w + 2) && (BETWEEN(mouse_y, w->y - 2, w->y + w->h + 2))))
			return 1;
	}
	else
	{
		// If they -were- moving the window, stop it cause they're not anymore. :)
		if (moving == w)
			moving = 0;
		// If they -were- resizing the window, stop it cause they're not anymore. :)
		if (resizing == w)
			resizing = 0;
		// If they -were- selecting tiles, stop it cause they're not anymore. :)
		if (selecting == w)
		{
			selecting = 0;
			// Update the actual selected tile area width and height
			memtileh = lmemtileh + 1;
			memtilew = lmemtilew + 1;
			// Save some other window-specific values needed for tile index calculations
			if (tileperiodicity)	// terraingfx used to be a requirement here...
	            memtileperiodicity = tileperiodicity;
//			else
//	            memtileperiodicity = w->intw / 32;
			// Actually go and set that first tile index in the upper left of the selection
			memtile = (memtileperiodicity * (tileyoff + yorigin)) + (tilexoff + xorigin);

			// Fill the tile clipboard with the selected tile #'s
			i = memtile;
            for (y = 0; y < memtileh; y++)
			{
				for (x = 0; x < memtilew; x++)
				{
					memtiles[y][x] = i++;
					memterrain[y][x] = TERRAIN_LAND;
				}
				i += memtileperiodicity - memtilew;
			}
		}
		// If this window has been marked for destruction, destroy it
		if (destroy == w)
		{
			destroy_window(w);
			destroy = 0;
			return 1;
		}
	}

	return 0;
}

int input_tilewnd(wnd_t *w)
{
	static wnd_t *moving = 0;
	static int xorigin, yorigin;
	static wnd_t *destroy = 0;

	if ((mouse_b & 1) || (mouse_b & 2))
	{
		// WINDOW CLOSING BEHAVIOR
		if (!moving)
		{
			if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 7, w->x + w->w - BORDER_WIDTH + 2))
			 && (BETWEEN(mouse_y, w->y + BORDER_WIDTH - 1, w->y + BORDER_WIDTH + TITLEBAR_HEIGHT - 4)))
			{
				destroy = w;
				return 1;
			}
			else
				if (destroy == w)
					destroy = 0;
		}

		// WINDOW MOVING BEHAVIOR
		if (!moving)
		{
			// Check for window movement
			if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH, w->x + w->w - BORDER_WIDTH))
	         && (BETWEEN(mouse_y, w->y + BORDER_WIDTH, w->y + TOP_HEIGHT)))
			{
				moving = w;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of moving a window.
			if (w == moving)	// Make sure they're moving THIS window though!
			{
				w->x += mouse_x - xorigin;
				w->intx = w->x + BORDER_WIDTH;
				xorigin = mouse_x;
				w->y += mouse_y - yorigin;
				w->inty = w->y + TOP_HEIGHT;
				yorigin = mouse_y;
				return 1;
			}

		// ADD NEW WINDOW INPUT HANDLING FUNCTIONALITY HERE

		// CLICKING INSIDE WINDOW BEHAVIOR - grab a tile
		if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH - 1, w->x + w->w - BORDER_WIDTH))
		 && (BETWEEN(mouse_y, w->y + TOP_HEIGHT - 1, w->y + w->h - BORDER_WIDTH)))
			return 1;

		// Handle clicks on the border that would not normally be acknowledged
		if ((BETWEEN(mouse_x, w->x - 2, w->x + w->w + 2) && (BETWEEN(mouse_y, w->y - 2, w->y + w->h + 2))))
			return 1;
	}
	else
	{
		// If they -were- moving the window, stop it cause they're not anymore. :)
		if (moving == w)
			moving = 0;
		// If window is marked for destruction, destroy it
		if (destroy == w)
		{
			destroy_window(w);
			destroy = 0;
			return 1;
		}
	}

	return 0;
}

int input_plainresize(wnd_t *w)
{
	static wnd_t *moving = 0;
	static wnd_t *resizing = 0;
	static wnd_t *selecting = 0;
	int x, y, i, j;
	static int xorigin, yorigin;
	unsigned long scrolltime = 0;
	static wnd_t *destroy = 0;

	// Change cursor to a resize cursor when it's over the lower-right corner of a window
	if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 4, w->x + w->w + BORDER_WIDTH - 4))
     && (BETWEEN(mouse_y, w->y + w->h - BORDER_WIDTH - 4, w->y + w->h + BORDER_WIDTH - 4)))
		mouse_cursor = (BITMAP *) datafile[CURSOR_RESIZE].dat;

	if ((mouse_b & 1) || (mouse_b & 2))
	{
		// WINDOW CLOSING BEHAVIOR
		if ((!resizing) && (!moving) && (!selecting))
		{
			if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 7, w->x + w->w - BORDER_WIDTH + 2))
			 && (BETWEEN(mouse_y, w->y + BORDER_WIDTH - 1, w->y + BORDER_WIDTH + TITLEBAR_HEIGHT - 4)))
			{
				destroy = w;
				return 1;
			}
			else
				if (destroy == w)
					destroy = 0;
		}

		// WINDOW MOVING BEHAVIOR
		if ((!moving) && (!resizing) && (!selecting))
		{
			// Check for window movement
			if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH, w->x + w->w - BORDER_WIDTH))
	         && (BETWEEN(mouse_y, w->y + BORDER_WIDTH, w->y + TOP_HEIGHT)))
			{
				moving = w;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of moving a window.
			if (w == moving)	// Make sure they're moving THIS window though!
			{
				w->x += mouse_x - xorigin;
				w->y += mouse_y - yorigin;
				w->intx = w->x + BORDER_WIDTH;
				w->inty = w->y + TOP_HEIGHT;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}

		// WINDOW RESIZING BEHAVIOR
		if ((!resizing) && (!moving) && (!selecting))
		{
			// Check for window resizing
			if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 4, w->x + w->w + BORDER_WIDTH - 4))
			 && (BETWEEN(mouse_y, w->y + w->h - BORDER_WIDTH - 4, w->y + w->h + BORDER_WIDTH - 4)))
			{
				resizing = w;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of resizing a window.
			if (w == resizing)	// Make sure they're resizing THIS window though!
			{
//				if (abs(mouse_x - xorigin) >= 32)
//				{
					// If they moved the mouse left by at least 32 pixels, size down
/*					if (mouse_x - xorigin < 0)
					{
						w->intw -= (mouse_x - xorigin);
						position_mouse(xorigin - (mouse_x - xorigin), mouse_y);
					}
					else */	// Otherwise they moved right
					{
						w->intw += (mouse_x - xorigin);
						position_mouse(xorigin + (mouse_x - xorigin), mouse_y);
					}

					// Don't allow the window to get smaller than 100x100
					if (w->intw < 100)
					{
						position_mouse(w->intx + w->intw, mouse_y);
						w->intw = 100;
					}

					w->w = (BORDER_WIDTH * 2) + w->intw;	// - 1
	
					// If this window has a horizontal scrollbar, add onto the total window width..
					if ((w->style & WND_SCROLLBARS) || (w->style & WND_HSCROLLBAR))
						w->w += SCROLLBAR_WIDTH;
				
					xorigin = mouse_x;
//				}
//				if (abs(mouse_y - yorigin) >= 32)
//				{
					// If they moved the mouse up by at least 32 pixels, size down
					/* if (mouse_y - yorigin < 0)
					{
						w->inth -= (mouse_y - yorigin);
						position_mouse(mouse_x, yorigin - (mouse_y - yorigin));
					}
					else */	// Otherwise they moved down
					{
						w->inth += (mouse_y - yorigin);
						position_mouse(mouse_x, yorigin + (mouse_y - yorigin));
					}

					// Don't allow the window to get smaller than 100x100
					if (w->inth < 100)
					{
						position_mouse(mouse_x, w->inty + w->inth);
						w->inth = 100;
					}
					w->h = BORDER_WIDTH + TOP_HEIGHT + w->inth - 1;
					// If this window has a vertical scrollbar, add onto the total window height..
					if ((w->style & WND_SCROLLBARS) || (w->style & WND_VSCROLLBAR))
						w->h += SCROLLBAR_WIDTH;
					yorigin = mouse_y;
//				}

				// Decide whether or not scrollbars are necessary

				// Recalculate the window's total width and height
				w->w = (BORDER_WIDTH * 2) + w->intw;
				w->h = BORDER_WIDTH + TOP_HEIGHT + w->inth;

                // Assume no scrollbars as the default
				w->style = WND_PLAIN;

				return 1;
			}

		// Check for the left scroll button
		if ((w->style & WND_HSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx, w->intx + SCROLLBAR_WIDTH) &&
			    BETWEEN(mouse_y, w->y + w->h - SCROLLBAR_WIDTH - BORDER_WIDTH, w->y + w->h - BORDER_WIDTH))
				if (tilexoff > 0)
					if (ms_time - scrolltime > 100)
					{
						scrolltime = ms_time;
						tilexoff--;
					}

		// Check for the right scroll button
		if ((w->style & WND_HSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + w->intw - SCROLLBAR_WIDTH - 1, w->intx + w->intw - 1) &&
			    BETWEEN(mouse_y, w->y + w->h - SCROLLBAR_WIDTH - BORDER_WIDTH, w->y + w->h - BORDER_WIDTH))
				if (tilexoff < (tileareaw - (w->intw / 32)))
					if (ms_time - scrolltime > 100)
					{
						scrolltime = ms_time;
						tilexoff++;
					}

		// Check for the top scroll button
		if ((w->style & WND_VSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + w->intw, w->intx + w->intw + SCROLLBAR_WIDTH) &&
			    BETWEEN(mouse_y, w->inty, w->inty + SCROLLBAR_WIDTH))
				if (tileyoff > 0)
					if (ms_time - scrolltime > 100)
					{
						scrolltime = ms_time;
						tileyoff--;
					}

		// Check for the bottom scroll button
		if ((w->style & WND_VSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + w->intw, w->intx + w->intw + SCROLLBAR_WIDTH) &&
			    BETWEEN(mouse_y, w->inty + w->inth - SCROLLBAR_WIDTH - 1, w->inty + w->inth - 1))
				if (tileyoff < (tileareah - (w->inth / 32)))
					if (ms_time - scrolltime > 100)
					{
						scrolltime = ms_time;
						tileyoff++;
					}

		// Check for dragging on the horizontal scrollbar
		if ((w->style & WND_HSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + SCROLLBAR_WIDTH, w->intx + w->intw - SCROLLBAR_WIDTH - 1) &&
			    BETWEEN(mouse_y, w->y + w->h - SCROLLBAR_WIDTH - BORDER_WIDTH, w->y + w->h - BORDER_WIDTH))
			{
				// Calculate width of total possible scrollbar area
				i = (w->intx + w->intw - SCROLLBAR_WIDTH - 1) - (w->intx + SCROLLBAR_WIDTH) - 1;
				// Calculate where in this area the mouse is..
                j = mouse_x - (w->intx + SCROLLBAR_WIDTH);

				tilexoff = tileareaw * j / i - (w->intw / 64);

				if (tilexoff < 0)
					tilexoff = 0;
				if (tilexoff > (tileareaw - (w->intw / 32)))
					tilexoff = (tileareaw - (w->intw / 32));
			}

		// Check for dragging on the vertical scrollbar
		if ((w->style & WND_VSCROLLBAR) || (w->style & WND_SCROLLBARS))
			if (BETWEEN(mouse_x, w->intx + w->intw, w->intx + w->intw + SCROLLBAR_WIDTH) &&
			    BETWEEN(mouse_y, w->inty + SCROLLBAR_WIDTH, w->inty + w->inth - SCROLLBAR_WIDTH - 1))
			{
				// Calculate width of total possible scrollbar area
				i = (w->inty + w->inth - SCROLLBAR_WIDTH - 1) - (w->inty + SCROLLBAR_WIDTH) - 1;
				// Calculate where in this area the mouse is..
                j = mouse_y - (w->inty + SCROLLBAR_WIDTH);

				tileyoff = tileareah * j / i - (w->inth / 64);

				if (tileyoff < 0)
					tileyoff = 0;
				if (tileyoff > (tileareah - (w->inth / 32)))
					tileyoff = (tileareah - (w->inth / 32));
			}

		// ADD NEW WINDOW INPUT HANDLING FUNCTIONALITY HERE

		// Handle clicks on the border that would not normally be acknowledged
		if ((BETWEEN(mouse_x, w->x - 2, w->x + w->w + 2) && (BETWEEN(mouse_y, w->y - 2, w->y + w->h + 2))))
			return 1;
	}
	else
	{
		// If they -were- moving the window, stop it cause they're not anymore. :)
		if (moving == w)
			moving = 0;
		// If they -were- resizing the window, stop it cause they're not anymore. :)
		if (resizing == w)
			resizing = 0;
		// If the window is marked for destruction, destroy it
		if (destroy == w)
		{
			destroy_window(w);
			destroy = 0;
			return 1;
		}
	}

	return 0;
}

int input_easymove(wnd_t *w)
{
	static wnd_t *moving = 0;
	static int xorigin, yorigin;
	static wnd_t *destroy = 0;

	if ((mouse_b & 1) || (mouse_b & 2))
	{
		// WINDOW CLOSING BEHAVIOR
		if ((!moving) && (!(w->style & WND_NOTITLE)))
		{
			if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 7, w->x + w->w - BORDER_WIDTH + 2))
			 && (BETWEEN(mouse_y, w->y + BORDER_WIDTH - 1, w->y + BORDER_WIDTH + TITLEBAR_HEIGHT - 4)))
			{
				destroy = w;
				return 1;
			}
			else
				if (destroy == w)
					destroy = 0;
		}

		// WINDOW MOVING BEHAVIOR
		if (!moving)
		{
			// Check for window movement (anywhere in the window, easymove, woo!)
//			if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH, w->x + w->w - BORDER_WIDTH))
//	         && (BETWEEN(mouse_y, w->y + BORDER_WIDTH, w->y + TOP_HEIGHT)))
			if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH, w->x + w->w - BORDER_WIDTH))
	         && (BETWEEN(mouse_y, w->y + BORDER_WIDTH, w->y + w->h - BORDER_WIDTH)))
			{
				moving = w;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of moving a window.
			if (w == moving)	// Make sure they're moving THIS window though!
			{
				w->x += mouse_x - xorigin;
				w->intx += mouse_x - xorigin;
				xorigin = mouse_x;
				w->y += mouse_y - yorigin;
				w->inty += mouse_y - yorigin;
				yorigin = mouse_y;
				return 1;
			}

		// ADD NEW WINDOW INPUT HANDLING FUNCTIONALITY HERE

		// Handle clicks on the border that would not normally be acknowledged
		if ((BETWEEN(mouse_x, w->x - 2, w->x + w->w + 2) && (BETWEEN(mouse_y, w->y - 2, w->y + w->h + 2))))
			return 1;
	}
	else
	{
		// If they -were- moving the window, stop it cause they're not anymore. :)
		if (moving == w)
			moving = 0;
		// If window is marked for destruction, destroy it
		if (destroy == w)
		{
			destroy_window(w);
			destroy = 0;
			return 1;
		}
	}

	return 0;
}

int input_minimap(wnd_t *w)
{
	static wnd_t *moving = 0;
	static int xorigin, yorigin;
	static wnd_t *destroy = 0;

	if ((mouse_b & 1) || (mouse_b & 2))
	{
		// WINDOW CLOSING BEHAVIOR
		if ((!moving) && (!(w->style & WND_NOTITLE)))
		{
			if ((BETWEEN(mouse_x, w->x + w->w - BORDER_WIDTH - 7, w->x + w->w - BORDER_WIDTH + 2))
			 && (BETWEEN(mouse_y, w->y + BORDER_WIDTH - 1, w->y + BORDER_WIDTH + TITLEBAR_HEIGHT - 4)))
			{
				destroy = w;
				return 1;
			}
			else
				if (destroy == w)
					destroy = 0;
		}

		// WINDOW MOVING BEHAVIOR
		if (!moving)
		{
			// Check for window movement (anywhere in the window, easymove, woo!)
			// Only move the window if the player is not carrying a unit with orders
			// to travel somewhere (ORDER_GOTO), so that destination may be specified
			// by clicking on the mini-map.
			if ((BETWEEN(mouse_x, w->x + BORDER_WIDTH, w->x + w->w - BORDER_WIDTH))
	         && (BETWEEN(mouse_y, w->y + BORDER_WIDTH, w->y + w->h - BORDER_WIDTH))
			 && !(player.unit_carried && player.unit_carried->orders == ORDER_GOTO))
			{
				moving = w;
				xorigin = mouse_x;
				yorigin = mouse_y;
				return 1;
			}
		}
		else	// Otherwise they might be in the process of moving a window.
			if (w == moving)	// Make sure they're moving THIS window though!
			{
				w->x += mouse_x - xorigin;
				w->intx += mouse_x - xorigin;
				xorigin = mouse_x;
				w->y += mouse_y - yorigin;
				w->inty += mouse_y - yorigin;
				yorigin = mouse_y;
				return 1;
			}

		// ADD NEW WINDOW INPUT HANDLING FUNCTIONALITY HERE
		if ((BETWEEN(mouse_x, w->intx, w->intx + w->intw))
         && (BETWEEN(mouse_y, w->inty, w->inty + w->inth))
		 && (player.unit_carried) && (player.unit_carried->orders == ORDER_GOTO))
		{

			player.unit_carried->dx = mouse_x - w->intx;
			player.unit_carried->dy = mouse_y - w->inty;
			if (debug)
				add_msg(MSG_COMPUTER, "Dest for unit #%d set to (%d,%d)", player.unit_carried->id, player.unit_carried->dx, player.unit_carried->dy);
		}

		// Handle clicks on the border that would not normally be acknowledged
		if ((BETWEEN(mouse_x, w->x - 2, w->x + w->w + 2) && (BETWEEN(mouse_y, w->y - 2, w->y + w->h + 2))))
			return 1;
	}
	else
	{
		// If they -were- moving the window, stop it cause they're not anymore. :)
		if (moving == w)
			moving = 0;
		// If window is marked for destruction, destroy it
		if (destroy == w)
		{
			destroy_window(w);
			destroy = 0;
			return 1;
		}
	}

	return 0;
}

void init_gui(void)
{
	static char firstinit = 1;

	if (firstinit)
	{
		mouse_cursor = (BITMAP *) datafile[CURSOR].dat;
		wnd_list = NULL;
		firstinit = 0;
		atexit(destroy_windows);
	}
	else
		add_msg(MSG_COMPUTER, "Warning: init_gui() called again!");

	create_wnd(SCREEN_W - 112, SCREEN_H - 112, 99, 99, WND_NOTITLE, "Mini-Map", draw_minimap, input_minimap, NULL);
}

void init_editor_gui(void)
{
	create_wnd(120, 120, 4 * 32, 4 * 32, WND_SCROLLBARS, "Internal Tileset", draw_tilesrc, input_tilesrc, NULL);

	create_wnd(40, 40, 32, 32, WND_PLAIN, "Tile", draw_tilewnd, input_tilewnd, NULL);

	create_wnd(80, 80, 64, 64, WND_PLAIN, "Palette", draw_palette, input_tilewnd, NULL);

//	create_wnd(120, 32, 100, 100, WND_PLAIN, "Memgraph", draw_memgraph, input_plainresize, NULL);

//	create_wnd(160, 64, 100, 100, WND_PLAIN, "Projgraph", draw_projgraph, input_plainresize, NULL);
}

wnd_t *create_wnd(int x, int y, int w, int h, char style, char *title, void (*draw)(wnd_t *w), int (*input)(wnd_t *w), void *data)
{
	wnd_t *wnd;

	wnd = (wnd_t *) malloc(sizeof(wnd_t));
	wnd->x = x;
	wnd->y = y;
	wnd->intx = x + BORDER_WIDTH;
	wnd->inty = y + TOP_HEIGHT;
	wnd->w = (BORDER_WIDTH * 2) + w;
	wnd->h = BORDER_WIDTH + TOP_HEIGHT + h;
	if (style & WND_NOTITLE)	// Handle windows without titles
	{
		wnd->h = BORDER_WIDTH + BORDER_WIDTH + h;
		wnd->inty = y + BORDER_WIDTH;
	}
	wnd->intw = w;
	wnd->inth = h;
	wnd->style = style;
	wnd->title = (char *) malloc(strlen(title) + 1);
	strcpy(wnd->title, title);
	wnd->draw = draw;
	wnd->input = input;

	// Take care of window-specific data
/*	if ((data) && (datasize))
	{
		wnd->data = (void *) malloc(datasize);
		memcpy(wnd->data, data, datasize);
	}
	else
		wnd->data = NULL; */
	wnd->data = data;

	// Modify the window's total width and height values to compensate for scrollbars

	// If this window has a horizontal scrollbar, add onto the total window width..
	if ((wnd->style & WND_SCROLLBARS) || (wnd->style & WND_HSCROLLBAR))
		wnd->w += SCROLLBAR_WIDTH;

	// If this window has a vertical scrollbar, add onto the total window height..
	if ((wnd->style & WND_SCROLLBARS) || (wnd->style & WND_VSCROLLBAR))
		wnd->h += SCROLLBAR_WIDTH;

	// Insert the new window at the beginning of the window [linked] list
	wnd->next = wnd_list;
	wnd_list = wnd;

	return wnd;
}

void destroy_windows(void)
{
	wnd_t *w, *tmp;

	w = wnd_list;
	while (w)
	{
		if (w->title)
			free(w->title);
		tmp = w->next;
		free(w);
		w = tmp;
	}
	wnd_list = NULL;
}

void destroy_window(wnd_t *wnd)
{
	wnd_t *w, *tmp;

	if (!wnd)		// ignore null pointers
		return;

	// If we're destroying the first window in the list...
	if (wnd_list == wnd)
	{
		if (wnd->next)
			wnd_list = wnd->next;
		else
			wnd_list = NULL;
		if (wnd->title)
		{
			heapused -= strlen(wnd->title);
			free(wnd->title);
		}
		heapused -= sizeof(wnd);
		free(wnd);
		wnd = NULL;
		return;
	}

	tmp = w = wnd_list;
	while (w)
	{
		if (w == wnd)	// found the window to destroy
		{
			if (w->title)
			{
				heapused -= strlen(w->title);
				free(w->title);
			}
			tmp->next = w->next;
			heapused -= sizeof(wnd);
			free(wnd);
			wnd = NULL;
			return;
		}
		tmp = w;
		w = w->next;
	}
}

void bring_to_front(wnd_t *w)
{
	wnd_t *tmp;

	// We need to put the given window on top, which means it is drawn last,
	// which means it must go to the end of the linked window list.

	if (!wnd_list)	// If there's no window list, this function should not be called!
		return;

	if (!w->next)	// If it's already in front, return
		return;

	if (!w)			// Ignore null pointers
		return;

	if (w == wnd_list)	// If the wnd_list pointer is the window.. special case
	{
		wnd_list = wnd_list->next;
		for (tmp = wnd_list; tmp->next; tmp = tmp->next);
		tmp->next = w;
		w->next = NULL;
		return;
	}

	// Let's start by going through the list looking for a pointer to the window.
	for (tmp = wnd_list; tmp->next != w; tmp = tmp->next);

	// Display some debug info
//	textprintf(HiddenPage, font, 134, LogicalHeight - 100, TEXT_COLOR, "Putting window %p on top", w);

	tmp->next = w->next;
	w->next = NULL;

	for (tmp = wnd_list; tmp->next; tmp = tmp->next);

	tmp->next = w;
}

// Pastes the currently selected tiles onto the map at a given point
void paste_memtile(int tilex, int tiley)
{
	int x, y;

	// Draw the selection tile box
	for (y = 0; y < memtileh; y++)
	{
		if ((y + tiley > MAP_HEIGHT - 1) || (y + tiley < 1))
			continue;
		for (x = 0; x < memtilew; x++)
		{
			if ((x + tilex > MAP_WIDTH - 1) || (x + tilex < 1))
				continue;

			mpos(map, tiley + y, tilex + x) = memtiles[y][x];
			mpos(terrain, tiley + y, tilex + x) = memterrain[y][x];
		}
	}
	redraw_terrain_bmp = 1;
}

// Fills (tiles) the map with the currently selected tiles
void fill_map_with_clipboard(void)
{
	int x, y;

	for (y = 0; y < MAP_HEIGHT; y += memtileh)
		for (x = 0; x < MAP_WIDTH; x += memtilew)
			paste_memtile(x, y);
}

// Pastes the contents of the clipboard in random places, *not overlapping each other!*
void random_paste(int num)
{
	char *dup;
	int i, ok, tx, ty, x, y, infloopcheck, success = 0;

	dup = (char *) malloc(MAP_WIDTH * MAP_HEIGHT);
	memset(dup, 0, 1);

	for (i = 0; i < num; i++)
	{
		infloopcheck = 0;
		while (infloopcheck++ < 2000)
		{
			tx = rand() % MAP_WIDTH;
			ty = rand() % MAP_HEIGHT;
			// Check for dupes
			ok = 1;
			for (x = 0; x < memtilew; x++)
				for (y = 0; y < memtileh; y++)
					if (((tx + x) >= 0) && ((tx + x) < MAP_WIDTH) && ((ty + y) >= 0) && ((ty + y) < MAP_HEIGHT))
						if (mpos(dup, ty + y, tx + x))
						{
							ok = 0;
							x = memtilew;
							break;
						}
			if (!ok)
				continue;
		}
		// Ok, we're all good, set the dupes matrix though
		for (x = 0; x < memtilew; x++)
			for (y = 0; y < memtileh; y++)
				if (((tx + x) >= 0) && ((tx + x) < MAP_WIDTH) && ((ty + y) >= 0) && ((ty + y) < MAP_HEIGHT))
					mpos(dup, ty + y, tx + x) = 1;
		// Actually go and put the tile selection on the map
		paste_memtile(tx, ty);
		success++;
	}

	free(dup);
	add_msg(MSG_COMPUTER, "Tossed %d out of %d requested.", success, num);
}

void draw_minimap(wnd_t *w)
{
	int i, b, x, y, sx, sy, xoff, yoff, offset, active;
	unsigned char c;
	unit_t *u;
	static unsigned long last_radar_sweep = 0;
	static unsigned long last_sweep_tick = 0;
	static int radar_radius = 0;
	static char drawing_radar = 0;
	static int radar_x, radar_y;
	radar_t *r;

	// Check to see if the map dimensions have changed.. if so, resize the window
	if ((w->intw != MAP_WIDTH) || (w->inth != MAP_HEIGHT))
	{
		w->intw = MAP_WIDTH;
		w->inth = MAP_HEIGHT;
		w->w = MAP_WIDTH + 2 * BORDER_WIDTH;
		w->h = MAP_HEIGHT + 2 * BORDER_WIDTH;
	}

	// Calculate starting coordinates
	sx = w->intx + 32;
	sy = w->inty + 32;

	// If the actual terrain itself has changed, redraw the [cached] terrain bitmap
	if (redraw_terrain_bmp)
	{
		redraw_terrain_bmp = 0;
		// Draw terrain
		offset = 0;
		for (y = 0; y < MAP_HEIGHT; y++)
		{
			for (x = 0; x < MAP_WIDTH; x++)
			{
				offset++;
/*				if (mpos(vis, y, x) == VIS_DARK)
					c = 0;
				else */
					switch (terrain[offset])
					{
						case TERRAIN_LAND:    c = 194; break;
						case TERRAIN_WATER:   c = 120; break;
						case TERRAIN_COAST:   c = 224; break;
						case TERRAIN_BLOCKED: c = 20;  break;
					}
	            putpixel(terrain_bmp, x, y, c);
			}
		}
	}

	// Draw the terrain map
//    blit(terrain_bmp, HiddenPage, 0, 0, sx, sy, MAP_WIDTH, MAP_HEIGHT);

	///////// Draw refreshable map objects to mapobj_bmp /////////

	// Clear the map objects bitmap
	clear(mapobj_bmp);

	// Draw units
	for (u = units; u; u = u->next)
		if (BETWEEN(u->x, 0, MAP_WIDTH) && BETWEEN(u->y, 0, MAP_HEIGHT))
//			if (mpos(vis, u->y, u->x) == VIS_LIGHT)
				putpixel(mapobj_bmp, u->x, u->y, 105);

	// Draw projectiles
	for (i = 0; i < active_projectiles; i++)
		if (proj[i].type != PROJ_NONEXISTENT)
			if (BETWEEN(proj[i].x, 0, MAP_WIDTH) && BETWEEN(proj[i].y, 0, MAP_HEIGHT))
//				if (mpos(vis, proj[i].y, proj[i].x) == VIS_LIGHT)
					putpixel(mapobj_bmp, proj[i].x, proj[i].y, 42);

	///////// RADAR PROCESSING /////////

	process_radar();

	///////// Draw the current visible mini-map (terrain + map objects) /////////

    blit(vismap_bmp, HiddenPage, 0, 0, sx, sy, MAP_WIDTH, MAP_HEIGHT);

	///////// Draw things that always refresh instantly (like player, bases) /////////

	// Draw player
	c = 36 + (simclock % 8);
	putpixel(HiddenPage, sx + player.x, sy + player.y, c);
	putpixel(HiddenPage, sx + player.x + 1, sy + player.y, c);
	putpixel(HiddenPage, sx + player.x - 1, sy + player.y, c);
	putpixel(HiddenPage, sx + player.x, sy + player.y + 1, c);
	putpixel(HiddenPage, sx + player.x, sy + player.y - 1, c);

	// Draw the player's main base (if they have one.. which they should!)
	if (player.mbase)
	{
		x = player.mbase->x + 2;
		y = player.mbase->y + 2;
		c = 80 + (simclock % 16);
		putpixel(HiddenPage, sx + x, sy + y, c);
		putpixel(HiddenPage, sx + x + 1, sy + y, c);
		putpixel(HiddenPage, sx + x - 1, sy + y, c);
		putpixel(HiddenPage, sx + x, sy + y + 1, c);
		putpixel(HiddenPage, sx + x, sy + y - 1, c);

		putpixel(HiddenPage, sx + x + 1, sy + y + 1, c);
		putpixel(HiddenPage, sx + x - 1, sy + y - 1, c);
		putpixel(HiddenPage, sx + x - 1, sy + y + 1, c);
		putpixel(HiddenPage, sx + x + 1, sy + y - 1, c);
	}

	// If we're in multiplayer...
	if (comm_mode != COMM_NONE)
	{
		// Draw enemy players
		c = 100 + (simclock % 8);
		for (i = 0; i < active_enemies; i++)
		{
			if (enemy[i].comm_state != CS_CONNECTED)
				continue;
			if (BETWEEN(enemy[i].x, 0, MAP_WIDTH) && BETWEEN(enemy[i].y, 0, MAP_HEIGHT))
//				if (mpos(vis, enemy[i].y, enemy[i].x) == VIS_LIGHT)
				{
					putpixel(HiddenPage, sx + enemy[i].x, sy + enemy[i].y, c);
					putpixel(HiddenPage, sx + enemy[i].x + 1, sy + enemy[i].y, c);
					putpixel(HiddenPage, sx + enemy[i].x - 1, sy + enemy[i].y, c);
					putpixel(HiddenPage, sx + enemy[i].x, sy + enemy[i].y + 1, c);
					putpixel(HiddenPage, sx + enemy[i].x, sy + enemy[i].y - 1, c);
				}
		}
	}

	///////// Draw any radar sweeps that are in progress /////////

	// Set the clipping settings to just this window area
	set_clip(HiddenPage, sx, sy, sx + w->intw - 1, sy + w->inth - 1);

	// Loop through the radar sweeps
	for (r = radar; r; r = r->next)
	{
		if (r->active)
		{
			if (r->u)
			{
				x = r->u->x;
				y = r->u->y;
			}
			else
			{
				x = r->x;
				y = r->y;
			}
			// Draw the radar sweep circle, radiating from the player
			b = 256 - (int) ((float) r->radius / (float) r->range * 256.0);
//			circle(HiddenPage, sx + x, sy + y, r->radius - 1, makecol(127, 127, 127));
			circle(HiddenPage, sx + x, sy + y, r->radius, makecol(b, b, b));
		}
	}

	// Restore the proper clipping settings
	set_clip(HiddenPage, 0, 0, LogicalWidth - 1, LogicalHeight - 1);
}

void load_terrain(char *filename)
{
    char buf[256];
	int i, j, k;
	char *chrptr;
	PALETTE oldpal;
	PALETTE thepal;
	PALETTE *tmppal;
	FILE *fp;
	DATAFILE *newterraingfx;

	strcpy(buf, filename);

	fp = fopen(buf, "rb");
	if (!fp)
	{
		strcat(buf, ".dat");
		fp = fopen(buf, "rb");
	}
	if (fp)
		fclose(fp);
	else
	{
		add_msg(MSG_COMPUTER, "Error: Unable to open tile datafile %s", buf);
		tilefile[0] = '\0';
		return;
	}

	// If we already have a terrain gfx datafile loaded, unload it
//	if (terraingfx)
//		unload_datafile(terraingfx);

	// If we're in a highcolor/truecolor mode, try to read the palette first
	if (bitmap_color_depth(screen) != 8)
	{
		newterraingfx = load_datafile_object(buf, "ZZZ_PALETTE");
		if (newterraingfx)
		{
			set_palette(newterraingfx[0].dat);
			unload_datafile_object(newterraingfx);
		}
		else
			add_msg(MSG_COMPUTER, "Error: Couldn't load palette in %s", buf);
	}

	// Load the terrain gfx datafile
	newterraingfx = load_datafile(buf);
	// If we couldn't load the specified data file, complain but return gracefully
	if (!newterraingfx)
	{
		add_msg(MSG_COMPUTER, "Error: Unable to load tile datafile %s", tilefile);
		tilefile[0] = '\0';
		return;
	}
	else
	{
		// We have successfully loaded the datafile - save its filename
		strcpy(tilefile, buf);
		// Search the datafile for a palette to load
		// Also count the number of tiles
		j = 0;
		NUM_TILES = 0;
		for (i = 0; newterraingfx[i].type != DAT_END; i++)
			if (newterraingfx[i].type == DAT_PALETTE)
			{
				get_palette(oldpal);
                set_palette((RGB *) newterraingfx[i].dat);
				create_rgb_table(&rgb_table, (RGB *) newterraingfx[i].dat, NULL);
				rgb_map = &rgb_table;
				get_palette(thepal);
//				TEXT_COLOR = makecol(223, 207, 239);
				j = 1;
				break;
			}
			else
				NUM_TILES++;

		if (!j)
		{
			add_msg(MSG_COMPUTER, "Error: Loaded tile datafile %s, but it didn't have a palette", buf);
//			tilefile[0] = '\0';
//			unload_datafile(terraingfx);
//			terraingfx = 0;
		}
		else  // Search the datafile for bitmaps to have their palettes remapped
		{
			j = 0;
			add_msg(MSG_COMPUTER, "Color depth is %dbpp", bitmap_color_depth(screen));
			if (bitmap_color_depth(screen) == 8)
			{
/*				if (datafile)
					unload_datafile(datafile);
				datafile = load_datafile("bapple2.dat"); */
				for (i = 0; datafile[i].type != DAT_END; i++)
					if (datafile[i].type == DAT_BITMAP)
					{
						remap_colors((BITMAP *) datafile[i].dat, oldpal, thepal);
						j++;
					}
				for (i = 0; shipgfx[i].type != DAT_END; i++)
					if (shipgfx[i].type == DAT_BITMAP)
					{
						remap_colors((BITMAP *) shipgfx[i].dat, oldpal, thepal);
						j++;
					}
				for (k = 0; choppergfx[k].type != DAT_PALETTE && choppergfx[k].type != DAT_END; k++);
				if (choppergfx[k].type == DAT_PALETTE)
					for (i = 0; choppergfx[i].type != DAT_END; i++)
						if (choppergfx[i].type == DAT_BITMAP)
						{
							remap_colors((BITMAP *) choppergfx[i].dat, (RGB *) choppergfx[k].dat, thepal);
							j++;
						}
			}
			add_msg(MSG_COMPUTER, "Loaded tile datafile %s, remapped %d bitmaps", buf, j);
		}

		// If there are a lot of tiles, change the dimensions of the source tile window
		if (NUM_TILES > 50)
		{
			tileperiodicity = 16;
			tileareaw = 16;
			tileareah = 28;
		}

		// Get rid of the old terraingfx and bring in the new one
		if (terraingfx)
			unload_datafile(terraingfx);
		terraingfx = newterraingfx;

		// Try to load animation data associated with this terrain datafile
        strcpy(get_extension(strcpy(buf, tilefile)), "ani");
		read_anim_data(buf, terraingfx);
	}
}

void load_frill_gfx(char *filename)
{
    char buf[256];
	int i, j;
	FILE *fp;
	DATAFILE *newfrillgfx;

	strcpy(buf, filename);

	fp = fopen(buf, "rb");
	if (!fp)
	{
		strcat(buf, ".dat");
		fp = fopen(buf, "rb");
	}
	if (fp)
		fclose(fp);
	else
	{
		add_msg(MSG_COMPUTER, "Error: Unable to open frill gfx datafile %s", buf);
		frillgfx = 0;
		return;
	}

	// If we're in a highcolor/truecolor mode, try to read the palette first
	if (bitmap_color_depth(screen) != 8)
	{
		newfrillgfx = load_datafile_object(buf, "ZZZ_PALETTE");
		if (newfrillgfx)
		{
			set_palette(newfrillgfx[0].dat);
			unload_datafile_object(newfrillgfx);
		}
		else
			add_msg(MSG_COMPUTER, "Error: Couldn't load palette in %s", buf);
	}

	// Load the frill gfx datafile
	newfrillgfx = load_datafile(buf);
	// If we couldn't load the specified data file, complain but return gracefully
	if (!newfrillgfx)
	{
		add_msg(MSG_COMPUTER, "Error: Unable to load frill gfx datafile %s", buf);
		frillgfx = 0;
		return;
	}
	else
	{
		// Count the number of tiles
		j = 0;
		NUM_FRILL_TILES = 0;
		for (i = 0; newfrillgfx[i].type != DAT_END; i++)
			if (newfrillgfx[i].type != DAT_PALETTE)
				NUM_FRILL_TILES++;

		j = 0;
		add_msg(MSG_COMPUTER, "Loaded frill gfx datafile %s", buf);

		// Get rid of the old terraingfx and bring in the new one
		if (frillgfx)
			unload_datafile(frillgfx);
		frillgfx = newfrillgfx;
	}
}
